/**
 *
 *   AnisotropicFluidEquations.h
 *
 *   Copyright (C) 2017 Radoslaw Ryblewski and Michael Strickland
 *
 *   This file is part of aHydro3p1.
 *
 *   aHydro3p1 is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   aHydro3p1 is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with aHydro3p1.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef _ANISOTROPIC_FLUID_EQUATIONS_H_
#define _ANISOTROPIC_FLUID_EQUATIONS_H_

#include "Global.h"

class AnisotropicFluidEquations : public Equations{
    
public:
    
    // constructor
    
    AnisotropicFluidEquations(Configurator *aConfigurator, Grid *aGrid);
    
    // destructor
    
    virtual ~AnisotropicFluidEquations();
    
    // functions
    
    virtual usi          GetNumberOfEquations(void);
    virtual string      GetName(usi aNumberOfEquation);
    virtual void        LoadRHS(const double aTau, Vector3D *** const aVArr, const usi aArg, const usi aTo = 2);
    virtual void        SetEquationOfState(EquationOfState *aEquationOfState);
    virtual void        PrintInfo(void);
    virtual double      FormatOutput(double aX, int aN);
    virtual double      FormatInput(double aX, int aN);
    virtual double      Geteeff(int i, int j, int k){return 1e6;}
    virtual double      GetTeff(double x, double y, double e){return 1e6;}
    virtual double      GetTeffpt(int i, int j, int k){ return 1e6;}
    virtual void        EnforceLandauMatching(Vector3D *** const aVArr) {}
    
private:
    
    void                  SetVariables(const double aTau, const usi i, const usi j, const usi k,
                                       Vector3D * const aVec0,
                                       Vector3D * const aVec1,
                                       Vector3D * const aVec2,
                                       Vector3D * const aVec3,
                                       Vector3D * const aVec4);
    double               GetRHSValue(const usi aEquation);
    inline double       Dx(Vector3D * const aVec, const usi i, const usi j, const usi k);
    inline double       Dy(Vector3D * const aVec, const usi i, const usi j, const usi k);
    inline double       Dz(Vector3D * const aVec, const usi i, const usi j, const usi k);
    
    
    inline void         CheckValues(       Vector3D * const tVec0,
                                    Vector3D * const tVec1,
                                    Vector3D * const tVec2,
                                    Vector3D * const tVec3,
                                    Vector3D * const tVec4, int i, int j, int k);
    
    inline double       r(double X);
    inline double       rt(double X);
    inline double       rl(double X);
    inline double       rp(double X);
    inline double       rtp(double X);
    inline double       rlp(double X);
    
    inline double       energy(const double aX);
    inline double       pressure(const double aX);
    inline double       energyp(const double aX);
    inline double       pressurep(const double aX);
    
    // member variables
    
    static const int mNumberOfEquations = 5;
    string *mNames;
    Vector3D ***mVArr;
    EquationOfState *mEquationOfState;
    Grid * const mGrid;
    Configurator *mConfigurator;
    int mEOSChoice;
    double norm;
    
    // optimization
    
    const usi mXP;
    const usi mYP;
    const usi mZP;
    
    const double mdX;
    const double mdY;
    const double mdZ;
    
    const double mdXi;
    const double mdYi;
    const double mdZi;
    
    const double m2dX;
    const double m2dY;
    const double m2dZ;
    
    const double m2dXi;
    const double m2dYi;
    const double m2dZi;
    
    const usi mXPm1;
    const usi mYPm1;
    const usi mZPm1;
    
    const double eb;
    
    double  t, e, Y;
    double  L, T, X, Ux, Uy;
    
    double  DLDe, DLDx, DLDy;
    double  DTDe, DTDx, DTDy;
    double  DUxDe, DUxDx, DUxDy;
    double  DUyDe, DUyDx, DUyDy;
    double  DXDe, DXDx, DXDy;
    
    double  R, RT, RL;
    double  Rp, RTp, RLp;
    
    double  En, P, Enp, Pp;
    double  EE, PT, PL;
    
    double  R1, R2, R3, R4, R5;
    
    double  Col, G;
    double  denLambdaUxUy, denTheta, denXi1, denXi2;
    
    double  Ux2, Ux3, Ux4;
    double  Uy2, Uy3, Uy4;
    double  U0, U02, U03, U04;
    double  Sh;
    double  Ch;
    double  X1;
    
    // coefficients standing at time derivatives in equations
    // CLt1 describes the coefficient: CLt1 * partial_tau Lambda in Equation 1
    
    double CLt1;
    double CXt1;
    double CUxt1;
    double CUyt1;
    double CTt1;
    double CLt2;
    double CXt2;
    double CUxt2;
    double CUyt2;
    double CTt2;
    double CLt3;
    double CXt3;
    double CTt3;
    double CLt4;
    double CXt4;
    double CUxt4;
    double CTt4;
    double CLt5;
    double CXt5;
    double CUyt5;
    double CTt5;
};
#endif // _ANISOTROPIC_FLUID_EQUATIONS_H_
