/**
 *
 *   AnisotropicFluidEquationsNew.h
 *
 *   2017 Mubarak Alqahtani, Mohammad Nopoush, Radoslaw Ryblewski, and Michael Strickland
 *
 *   This file is part of aHydro3p1.
 *
 *   aHydro3p1 is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   aHydro3p1 is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with aHydro3p1.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef _ANISOTROPIC_FLUID_EQUATIONS_NEW_H_
#define _ANISOTROPIC_FLUID_EQUATIONS_NEW_H_

#include "Global.h"
#include "Hfuncs.h"
#include "RelaxationTime.h"
#include "RealisticEquationOfState.h"
#include "ThermalMass.h"
#include "TransportCoefficient.h"
#include "LinearTransportCoefficient.h"

#define DERIVORDER 4 // spatial derivative order used in updates; should be 2 or 4

struct lambdaFunc_params
{
    double lambda, ax, ay, az;
    Hfuncs *hfuncs;
    ThermalMass *mass;
};

class AnisotropicFluidEquationsNew : public Equations {
    
public:
    
    // constructor
    
    AnisotropicFluidEquationsNew(Configurator *aConfigurator, Grid *aGrid);
    
    // destructor
    
    virtual ~AnisotropicFluidEquationsNew();
    
    // functions
    
    virtual usi         GetNumberOfEquations(void);
    virtual string      GetName(usi aNumberOfEquation);
    virtual void        LoadRHS(const double aTau, Vector3D *** const aVArr, const usi aArg, const usi aTo = 2);
    virtual void        LoadAB(double * A, double * B);
    virtual void        SetEquationOfState(EquationOfState *aEquationOfState);
    virtual void        PrintInfo(void);
    virtual double      FormatOutput(double aX, int aN);
    virtual double      FormatInput(double aX, int aN);
    virtual double      Geteeff(int i, int j, int k);
    virtual double      GetTeffpt(int i, int j, int k);
    virtual double      GetTeff(double x, double y, double e);
    virtual void        EnforceLandauMatching(Vector3D *** const aVArr);
        
private:
    
    void                  SetVariables(const double aTau, const usi i, const usi j, const usi k,
                                       Vector3D * const aVec0,
                                       Vector3D * const aVec1,
                                       Vector3D * const aVec2,
                                       Vector3D * const aVec3,
                                       Vector3D * const aVec4,
                                       Vector3D * const aVec5,
                                       Vector3D * const aVec6,
				       Vector3D * const aVec7);

    inline double       Dx(Vector3D * const aVec, const usi i, const usi j, const usi k);
    inline double       Dy(Vector3D * const aVec, const usi i, const usi j, const usi k);
    inline double       Dz(Vector3D * const aVec, const usi i, const usi j, const usi k);
    
    
    inline void         CheckValues(Vector3D * const tVec0,
                                    Vector3D * const tVec1,
                                    Vector3D * const tVec2,
                                    Vector3D * const tVec3,
                                    Vector3D * const tVec4,
                                    Vector3D * const tVec5,
                                    Vector3D * const tVec6,
				    Vector3D * const tVec7, int i, int j, int k);
    
    //inline double       temperature(const double aX);
    inline double       energy(const double aX);
    inline double       pressure(const double aX);
    //inline double       energyp(const double aX);
    //inline double       pressurep(const double aX);
    
    // member variables
    
    static const int mNumberOfEquations = 8;
    string *mNames;
    Vector3D ***mVArr;
    EquationOfState *mEquationOfState;
    Grid * const mGrid;
    Configurator *mConfigurator;
    int mEOSChoice;
    
    double norm;
    Hfuncs hfuncs;
    RelaxationTime tq;
    ThermalMass mass;
    LinearTransportCoefficient etabar;
    
    // optimization
    
    const usi mXP;
    const usi mYP;
    const usi mZP;
    
    const double mdX;
    const double mdY;
    const double mdZ;
    
    const double mdXi;
    const double mdYi;
    const double mdZi;
    
    const double m2dX;
    const double m2dY;
    const double m2dZ;
    
    const double m2dXi;
    const double m2dYi;
    const double m2dZi;
    
    const usi mXPm1;
    const usi mYPm1;
    const usi mZPm1;
    
    const double eb; // to go away
    
    double  tau, vs;
    double  ax, ay, az, ux, uy, vt, L, T, m;
    
    double  d2ax,d3ax,d4ax;
    double  d2ay,d3ay,d4ay;
    double  d2az,d3az,d4az;
    
    double  d2ux,d3ux,d4ux;
    double  d2uy,d3uy,d4uy;
    double  d2vt,d3vt,d4vt;
    
    double  d2L,d3L,d4L;
    double  d2T,d3T,d4T;
    
    double  H3, H3X, H3Y, H3L, H3B;
    double  d1H3,d2H3,d3H3,d4H3;
    double  d1H3X,d2H3X,d3H3X,d4H3X;
    double  d1H3Y,d2H3Y,d3H3Y,d4H3Y;
    double  d1H3L,d2H3L,d3H3L,d4H3L;
    double  H3eq,d1H3eq;
    
    double  Eeq, Peq;
    
    double  teq;
    double  gFac;
    
    double  Ch, Sh;
    
    // convenience variables
    
    double ax2,ax3,ax4;
    double ay2,ay3,ay4;
    double az2,az3,az4;
    
    double ux2,ux3,ux4;
    double uy2,uy3,uy4;
    
    double L2,L3,L4,L5,T2,T3,T4,T5;
    double m2,m3,m4;

    double K1, K2,K3,K4,K1eq,K2eq,K3eq,K4eq,dTB,dTm,mh,mheq;   // *******************
    
    double up,up2,up3,up4,up5;
    double u0,u02,u03,u04;
    
    static double lambdaFunc(double T, void *params);
    double findT(double lambda, double ax, double ay, double az, double tguess, double delta);
    double findT(double lambda, double ax, double ay, double az);
        
};
#endif // _ANISOTROPIC_FLUID_EQUATIONS_NEW_H_
