/**
*
*   Global.h
*
*   Copyright (C) 2017 Radoslaw Ryblewski and Michael Strickland
*
*   This file is part of aHydro3p1.
*
*   aHydro3p1 is free software: you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation, either version 3 of the License, or
*   (at your option) any later version.
*
*   aHydro3p1 is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with aHydro3p1.  If not, see <http://www.gnu.org/licenses/>.
*
*/

#ifndef _GLOBAL_H_
#define _GLOBAL_H_

// ISO C++ libraries

#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <complex>
#include <iomanip>          // setw(), manipulators
#include <algorithm>        // replace(),
#include <vector>
#include <bitset>

// C libraries (std namespace used)

#include <cmath>            // exp(), ceil(),isnan(),sqrt(),log(), mathematic functions
#include <cstdlib>          // exit()
#include <cstdio>
#include <ctime>            // difftime()
#include <cerrno>           // errno, contains error definitions which may be set by different standard c functions
#include <cfloat>           // for checking limits of double
#include <climits>          // for checking limits of int

// non-ISO libraries

#include <dirent.h>         // for directories cleaning
#include <unistd.h>         // sleep();
#include <sys/types.h>
#include <sys/stat.h>       // int mkdir (const char *filename, mode_t mode)

#include <gsl/gsl_linalg.h>

//
//#include <gsl/gsl_errno.h>
//#include <gsl/gsl_spline.h>  // no longer needed

// namespaces directives

using namespace std;

// type definitions

typedef unsigned short int usi;
typedef float BINnumber;

// classes header files

// these are repeated to let them be recognized by Parser

#include "Vector1D.h"
#include "Vector3D.h"
#include "Vector4D.h"

#include "Crc32.h"
#include "GlobalFunctions.h"
#include "SpecialFunctions.h"
#include "Configurator.h"
#include "Parser.h"
#include "Grid.h"

#include "EquationOfState.h"
#include "IdealEquationOfState.h"
#include "RealisticEquationOfState.h"

#include "Nucleus.h"
#include "OpticalGlauber.h"
#include "MCGlauber.h"

#include "Equations.h"
#include "PerfectFluidEquations.h"
#include "AnisotropicFluidEquations.h"
#include "AnisotropicFluidEquationsNew.h"

#include "Initializer.h"

#include "InitialSource.h"
#include "FlatInitialSource.h"
#include "GaussianInitialSource.h"
#include "GlauberInspiredInitialSource.h"
#include "GaussianBlobsInitialSource.h"
#include "MCGlauberInitialSource.h"

#include "InitialAnisotropy.h"
#include "FlatInitialAnisotropy.h"

#include "InitialFlowX.h"
#include "NoInitialFlowX.h"

#include "InitialFlowY.h"
#include "NoInitialFlowY.h"

#include "InitialRapidity.h"
#include "BjorkenInitialRapidity.h"

#include "Stepper.h"
#include "RungeKuttaStepper.h"
#include "EulerStepper.h"

#include "Smoother.h"
#include "WeightedLAXSmoother.h"

#include "Hfuncs.h"
#include "RelaxationTime.h"
#include "Pintegrals.h"
#include "ThermalMass.h"

#include "TransportCoefficient.h"
#include "LinearTransportCoefficient.h"

#include "Solver.h"

#include "Freezer.h"
#include "ParametrizationFreezer.h"
#include "Cornelius.h"
#include "HyperCube.h"
#include "TriangulationFreezer.h"

// version definition

#define _AHYDRO3P1_VERSION "0.1a"

// debug information defitions

#define PRINT_MESSAGE(_mes) std::cout << _mes << std::endl;

#if _DEBUG_LEVEL_==0
  #define PRINT_DEBUG_3(_mes)
  #define PRINT_DEBUG_2(_mes)
  #define PRINT_DEBUG_1(_mes)
#elif _DEBUG_LEVEL_==1
  #define PRINT_DEBUG_3(_mes)
  #define PRINT_DEBUG_2(_mes)
  #define PRINT_DEBUG_1(_mes) std::cerr << "DL1:" << _mes << std::endl;
#elif _DEBUG_LEVEL_==2
  #define PRINT_DEBUG_3(_mes)
  #define PRINT_DEBUG_2(_mes) std::cerr << "DL2:" << _mes << std::endl;
  #define PRINT_DEBUG_1(_mes) std::cerr << "DL1:" << _mes << std::endl;
#elif _DEBUG_LEVEL_==3
  #define PRINT_DEBUG_3(_mes) std::cerr << "DL3:" << _mes << std::endl;
  #define PRINT_DEBUG_2(_mes) std::cerr << "DL2:" << _mes << std::endl;
  #define PRINT_DEBUG_1(_mes) std::cerr << "DL1:" << _mes << std::endl;
#endif

// global mathematical constants definitions

const double Pi  = fabs(2.0 * asin(1.0));
const double Pi2 = pow(Pi,2);
const double Pi3 = pow(Pi,3);
const double Pi4 = pow(Pi,4);
const double E   = exp(1);
const double zeta3 = 1.20206;                       // Riemann zeta function for 3

// global physical constants definitions

const double NeutronMass = 0.93957;                 // neutron mass [GeV]
const double ProtonMass = 0.93827;                  // proton mass [GeV]
const double h = 4.13566751691e-15 * 1.0e-9;        // planck constant [eV*s] * 10^-9 = [GeV*s]
const double hbar = 0.5 * h / Pi;                   // reduced planck constant [GeV*s]
const double c = 2.99792458e8 * 1.0e+15;            // speed of light [m * s^-1] * 10^-15 = [fm / s]
const double hbarc = hbar * c;                      // [fm / s] * [GeV*s] = [GeV * fm]
const double hbarc2 = hbarc * hbarc;                // [fm / s] * [GeV*s] = [GeV * fm]
const double hbarc3 = hbarc * hbarc2;               // [fm / s] * [GeV*s] = [GeV * fm]
const double hbarc4 = hbarc * hbarc3;               // [fm / s] * [GeV*s] = [GeV * fm]
const double rad2deg = 180 / Pi;                    // transformation constant between radians and degrees

const double small = (double)1e-8; // value used in the specification of initial conditions; ux, uy
const double verysmall = (double)1e-4; // value used in the specification of initial conditions; mc glauber
const double tiny = (double)1e-16; // value used in the specification of initial conditions
const double UnitsFactor = 1; // value oof the conversion factor for output temperature/lambda, hbarc => GeV units, 1 => fm

// other variables

extern string sConfigurationFileDirectory;
extern string sGeneralConfigurationFileName;

#endif
