/** 
*
*   AnisotropicFluidEquations.cxx
*	
*   Copyright (C) 2017 Radoslaw Ryblewski and Michael Strickland
*
*   This file is part of aHydro3p1.
*
*   aHydro3p1 is free software: you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation, either version 3 of the License, or
*   (at your option) any later version.
*
*   aHydro3p1 is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with aHydro3p1.  If not, see <http://www.gnu.org/licenses/>.
*    
*/

#include "Global.h"

/*!
*
*  \brief Default constructor
*
*  \li aConfigurator = pointer to configuration
*  \li aGrid = pointer to grid
*
*/

AnisotropicFluidEquations::AnisotropicFluidEquations(Configurator *aConfigurator, Grid *aGrid)
: mNames(0), mGrid(aGrid), mConfigurator(aConfigurator), mEquationOfState(0), eb(mConfigurator->GetDouble("etabar")),
mXP(aGrid->GetXP()), mYP(aGrid->GetYP()), mZP(aGrid->GetZP()),
mdX(aGrid->GetdX()), mdY(aGrid->GetdY()), mdZ(aGrid->GetdZ()),
m2dX(2 * mdX), m2dY(2 * mdY), m2dZ(2 * mdZ),
mdXi(1 / mdX), mdYi(1 / mdY), mdZi(1 / mdZ),
m2dXi(1 / m2dX), m2dYi(1 / m2dY), m2dZi(1 / m2dZ),
mXPm1(mXP - 1), mYPm1(mYP - 1), mZPm1(mZP - 1){

    PRINT_DEBUG_3("<AnisotropicFluidEquations::AnisotropicFluidEquations>");

    // initialization of names of equations

    mNames = new string [mNumberOfEquations];

    mNames[0] = "lambda";
    mNames[1] = "rap";
    mNames[2] = "ux";
    mNames[3] = "uy";
    mNames[4] = "xi";

    mEOSChoice = mConfigurator->GetInt("EquationOfState");

}

/*!
*
*  \brief Default destructor
*
*/

AnisotropicFluidEquations::~AnisotropicFluidEquations(){

    PRINT_DEBUG_3("<AnisotropicFluidEquations::~AnisotropicFluidEquations>");

    if(mNames){
    delete [] mNames;
    }
    mNames = 0;

}

/*!
*
*  \brief Sets equation of state
*
*  \li aEquationOfState = pointer to equation of state
*
*/

void AnisotropicFluidEquations::SetEquationOfState(EquationOfState *aEquationOfState){

    PRINT_DEBUG_3("<AnisotropicFluidEquations::SetEquationOfState>");

    mEquationOfState = aEquationOfState;

}

/*!
*
*  \brief Returns number of equations
*
*/

usi AnisotropicFluidEquations::GetNumberOfEquations(void){

    PRINT_DEBUG_3("<AnisotropicFluidEquations::GetNumberOfEquations>");

    return mNumberOfEquations;

}

/*!
*
*  \brief Returns name of equation
*
*  \li aNumberOfEquation = number of equation
*
*/

string AnisotropicFluidEquations::GetName(usi aNumberOfEquation){

    PRINT_DEBUG_3("<AnisotropicFluidEquations::GetName>");

    return mNames[aNumberOfEquation];

}

/*!
*
*  \brief Loads RHS of the equation
*
*  \li aTau = time
*  \li aVArr = arrays of 3D vectors
*  \li aArg = number of array which contains values of the fields being arguments for RHS equation
*  \li aTo = number of array where RHS equation will be loaded in
*
*/

void AnisotropicFluidEquations::LoadRHS(const double aTau, Vector3D *** const aVArr, const usi aArg, const usi aTo){

    PRINT_DEBUG_3("<AnisotropicFluidEquations::LoadRHS>");

    register Vector3D * const tVec0A = aVArr[0][aArg];
    register Vector3D * const tVec1A = aVArr[1][aArg];
    register Vector3D * const tVec2A = aVArr[2][aArg];
    register Vector3D * const tVec3A = aVArr[3][aArg];
    register Vector3D * const tVec4A = aVArr[4][aArg];

    register Vector3D * const tVec0 = aVArr[0][aTo];
    register Vector3D * const tVec1 = aVArr[1][aTo];
    register Vector3D * const tVec2 = aVArr[2][aTo];
    register Vector3D * const tVec3 = aVArr[3][aTo];
    register Vector3D * const tVec4 = aVArr[4][aTo];

    for (usi i = 0; i < mXP; i++)
        for (usi j = 0; j < mYP; j++)
            for (usi k = 0; k < mZP; k++){

            SetVariables(aTau,i,j,k,tVec0A,tVec1A,tVec2A,tVec3A,tVec4A);

            (*tVec0)(i,j,k) = GetRHSValue(0);
            (*tVec1)(i,j,k) = GetRHSValue(1);
            (*tVec2)(i,j,k) = GetRHSValue(2);
            (*tVec3)(i,j,k) = GetRHSValue(3);
            (*tVec4)(i,j,k) = GetRHSValue(4);

            CheckValues(tVec0,tVec1,tVec2,tVec3,tVec4,i,j,k);
            }
}


inline void AnisotropicFluidEquations::CheckValues( Vector3D * const tVec0,
                                                Vector3D * const tVec1,
                                                Vector3D * const tVec2,
                                                Vector3D * const tVec3,
                                                Vector3D * const tVec4, int i, int j, int k){

    PRINT_DEBUG_3("<AnisotropicFluidEquations::CheckValues>");

        int flag = 0;
		if (isnan((*tVec0)(i,j,k))) {
            if(!flag) { cout << endl; PrintMessage("Error!                   ");}
			PrintRawMessage("Found nan in ", mNames[0]);
			flag=1;
		}
		if (isnan((*tVec1)(i,j,k))) {
            if(!flag) { cout << endl; PrintMessage("Error!                   ");}
			PrintRawMessage("Found nan in ", mNames[1]);
			flag=1;
		}
		if (isnan((*tVec2)(i,j,k))) {
            if(!flag) { cout << endl; PrintMessage("Error!                   ");}
			PrintRawMessage("Found nan in ", mNames[2]);
			flag=1;
		}
		if (isnan((*tVec3)(i,j,k))) {
            if(!flag) { cout << endl; PrintMessage("Error!                   ");}
			PrintRawMessage("Found nan in ", mNames[3]);
			flag=1;
		}
        if (isnan((*tVec4)(i,j,k))) {
            if(!flag) { cout << endl; PrintMessage("Error!                   ");}
			PrintRawMessage("Found nan in ", mNames[4]);
			flag=1;
		}
    /*
		if ((*tVec4)(i,j,k) < -1) {
            if(!flag) { cout << endl; PrintMessage("Error!                   ");}
			PrintRawMessage("-1 > ", mNames[4]);
			flag=1;
		}
     */
//		if (i == 0 && j ==0 && k == 0) {
//            if(!flag) { cout << endl; PrintMessage("Error!                   ");}
//			PrintRawMessage("At edge");
//			flag=1;
//		}
		if (flag) {
            PrintRawMessage("");
            PrintRawMessage("----------------");
            PrintRawMessage("OBTAINED RESULT:");
            PrintRawMessage("----------------");
            PrintRawMessage("");
            PrintRawMessage("L [GeV] " , hbarc * (*tVec0)(i,j,k));
			PrintRawMessage("T " , (*tVec1)(i,j,k));
			PrintRawMessage("Ux " , (*tVec2)(i,j,k));
			PrintRawMessage("Uy " , (*tVec3)(i,j,k));
			PrintRawMessage("Xi " , (*tVec4)(i,j,k));
            PrintRawMessage("");
            PrintRawMessage("------------");
            PrintRawMessage("VALUES USED:");
            PrintRawMessage("------------");
            PrintRawMessage("");
            PrintRawMessage("GRID POINT");
            PrintRawMessage("");
			PrintRawMessage("i " , i);
			PrintRawMessage("j " , j);
			PrintRawMessage("k " , k);
            PrintRawMessage("");
			PrintRawMessage("SPACE-TIME POINT");
            PrintRawMessage("");
			PrintRawMessage("tau [fm] " , t);
			PrintRawMessage("x [fm] " , mGrid->GetX(i));
			PrintRawMessage("y [fm] " , mGrid->GetY(j));
			PrintRawMessage("eta " , mGrid->GetZ(k));
            PrintRawMessage("");
			PrintRawMessage("FIELDS");
            PrintRawMessage("");
			PrintRawMessage("L [GeV] " , hbarc *L);
			PrintRawMessage("T " , T);
			PrintRawMessage("X " , X);
			PrintRawMessage("Ux " , Ux);
			PrintRawMessage("Uy " , Uy);
			PrintRawMessage("Y = e -T" , Y);
            PrintRawMessage("");
            PrintRawMessage("SPACE DERIVATIVES");
            PrintRawMessage("");
			PrintRawMessage("DLDe " , DLDe);
			PrintRawMessage("DLDx " , DLDx);
			PrintRawMessage("DLDy " , DLDy);
			PrintRawMessage("DTDe " , DTDe);
			PrintRawMessage("DTDx " , DTDx);
			PrintRawMessage("DTDy " , DTDy);
			PrintRawMessage("DXDe " , DXDe);
			PrintRawMessage("DXDx " , DXDx);
			PrintRawMessage("DXDy " , DXDy);
			PrintRawMessage("DUxDe " , DUxDe);
			PrintRawMessage("DUxDx " , DUxDx);
			PrintRawMessage("DUxDy " , DUxDy);
			PrintRawMessage("DUyDe " , DUyDe);
			PrintRawMessage("DUyDx " , DUyDx);
			PrintRawMessage("DUyDy " , DUyDy);
            PrintRawMessage("");
			PrintRawMessage("R FUNCTIONS");
            PrintRawMessage("");
			PrintRawMessage("R " , R);
			PrintRawMessage("RT " , RT);
			PrintRawMessage("RL " , RL);
			PrintRawMessage("Rp " , Rp);
			PrintRawMessage("RTp " , RTp);
			PrintRawMessage("RLp " , RLp);
            PrintRawMessage("");
			PrintRawMessage("EOS");
            PrintRawMessage("");
			PrintRawMessage("E [GeV^4]" , En * hbarc4);
			PrintRawMessage("P [GeV^4]" , P * hbarc4);
			PrintRawMessage("E' [GeV^3]" , Enp * hbarc3);
			PrintRawMessage("P' [GeV^3]" , Pp * hbarc3);
			PrintRawMessage("E(X,L) [GeV^4]" , EE * hbarc4);
			PrintRawMessage("PT(X,L) [GeV^4]" , PT * hbarc4);
			PrintRawMessage("PL(X,L) [GeV^4]" , PL * hbarc4);
            PrintRawMessage("");
			PrintRawMessage("OTHER VARIABLES");
            PrintRawMessage("");
			PrintRawMessage("Sinh(Y) " , Sh);
			PrintRawMessage("Cosh(Y) " , Ch);
			PrintRawMessage("Gamma " , G);
			PrintRawMessage("Colisional Kernel " , Col);
			PrintRawMessage("R1" , R1);
			PrintRawMessage("R2 " , R2);
			PrintRawMessage("R3 " , R3);
			PrintRawMessage("R4 " , R4);
			PrintRawMessage("R5 " , R5);
			PrintRawMessage("Denominator of Lambda/Ux/Uy " , denLambdaUxUy);
			PrintRawMessage("Denominator of Theta" ,denTheta);
            PrintRawMessage("Denominator of Xi No1" , denXi1);
            PrintRawMessage("Denominator of Xi No2" , denXi2);
            PrintRawMessage("Double range on this machine :" , d2s((long double)DBL_MIN) + ", " + d2s((long double)DBL_MAX));

            PrintValue(CLt1);
            PrintValue(CXt1 );
            PrintValue(CUxt1 );
            PrintValue(CUyt1 );
            PrintValue(CTt1 );

            PrintValue(CLt2 );
            PrintValue(CXt2 );
            PrintValue(CUxt2 );
            PrintValue(CUyt2 );
            PrintValue(CTt2 );

            PrintValue(CLt3 );
            PrintValue(CXt3 );
            PrintValue(CTt3 );

            PrintValue(CLt4 );
            PrintValue(CXt4 );
            PrintValue(CUxt4 );
            PrintValue(CTt4 );

            PrintValue(CLt5 );
            PrintValue(CXt5 );
            PrintValue(CUyt5 );
            PrintValue(CTt5 );

			exit(ProgramStatus(ERROR_INCORRECT_VALUE, "AnisotropicFluidEquations::CheckValues"));
		}
}

inline double AnisotropicFluidEquations::Dx(Vector3D * const aVec, const usi i, const usi j, const usi k) {

    PRINT_DEBUG_3("<AnisotropicFluidEquations::Dx>");

	if (i == 0) return ((*aVec)(1,j,k) - (*aVec)(0,j,k)) * mdXi;                             // right difference
	else if (i == mXPm1) return ((*aVec)(mXPm1,j,k) - (*aVec)(mXP - 2,j,k)) * mdXi;          // left difference
	else return ((*aVec)(i + 1,j,k) - (*aVec)(i - 1,j,k)) * m2dXi;                           // centered difference
}

inline double AnisotropicFluidEquations::Dy(Vector3D * const aVec, const usi i, const usi j, const usi k) {

    PRINT_DEBUG_3("<AnisotropicFluidEquations::Dy>");

	if (j == 0) return ((*aVec)(i,1,k) - (*aVec)(i,0,k)) * mdYi;                         // right difference
	else if (j == mYPm1) return ((*aVec)(i,mYPm1,k) - (*aVec)(i,mYP -2,k)) * mdYi;     // left difference
	else return ((*aVec)(i,j + 1,k) - (*aVec)(i,j - 1,k)) * m2dYi;                       // centered difference
}

inline double AnisotropicFluidEquations::Dz(Vector3D * const aVec, const usi i, const usi j, const usi k) {

    PRINT_DEBUG_3("<AnisotropicFluidEquations::Dz>");

	if (k == 0) return ((*aVec)(i,j,1) - (*aVec)(i,j,0)) * mdZi;                         // right difference
	else if (k == mZPm1) return ((*aVec)(i,j,mZPm1) - (*aVec)(i,j,mZP - 2)) * mdZi;    // left difference
	else return ((*aVec)(i,j,k + 1) - (*aVec)(i,j,k - 1)) * m2dZi;                       // centered difference
}

void AnisotropicFluidEquations::SetVariables(const double aTau, const usi i, const usi j, const usi k, Vector3D * const aVec0, Vector3D * const aVec1,Vector3D * const aVec2,Vector3D * const aVec3,Vector3D * const aVec4){

    PRINT_DEBUG_3("<AnisotropicFluidEquations::SetVariables>");

    // set functions values

    L   = (*aVec0)(i,j,k);  // Lambda
    T   = (*aVec1)(i,j,k);  // Theta
    Ux  = (*aVec2)(i,j,k);  // Ux
    Uy  = (*aVec3)(i,j,k);  // Uy
    X   = (*aVec4)(i,j,k);  // Xi

    // set eta, tau and Y

    t = aTau;
    e = mGrid->GetZ(k);
    Y   =   e - T;

    // set derivatives

    DLDx    = Dx(aVec0,i,j,k);
    DLDy    = Dy(aVec0,i,j,k);
    DLDe    = Dz(aVec0,i,j,k);

    DTDx    = Dx(aVec1,i,j,k);
    DTDy    = Dy(aVec1,i,j,k);
    DTDe    = Dz(aVec1,i,j,k);

    DUxDx   = Dx(aVec2,i,j,k);
    DUxDy   = Dy(aVec2,i,j,k);
    DUxDe   = Dz(aVec2,i,j,k);

    DUyDx   = Dx(aVec3,i,j,k);
    DUyDy   = Dy(aVec3,i,j,k);
    DUyDe   = Dz(aVec3,i,j,k);

    DXDx    = Dx(aVec4,i,j,k);
    DXDy    = Dy(aVec4,i,j,k);
    DXDe    = Dz(aVec4,i,j,k);

    // set EOS

    R   =   r(X);
    RT  =   rt(X);
    RL  =   rl(X);
    Rp  =   rp(X);
    RTp =   rtp(X);
    RLp =   rlp(X);

    En  =   energy(L);
    P   =   pressure(L);
    Enp =   energyp(L);
    Pp  =   pressurep(L);

    /*  Do not use this for now
///// scale EOS
//
    double sc =  1/En;
    En *= sc;
    P  *= sc;
    Enp*= sc;
    Pp *= sc;
//
/////
    */

    EE  =   En * R;
    PT  =   P * RT;
    PL  =   P * RL;

    /// OPTIMIZATION BEGIN

    Sh      =   sinh(Y);
    Ch      =   cosh(Y);

    // check if Cosh or Sinh it is infinity

    if(!isfinite(Sh)) Sh = signbit(Sh) ? -1e20 : 1e20;
    if(!isfinite(Ch)) Ch = 1e21;


    Ux2 =   Ux * Ux;
    Ux3 =   Ux * Ux2;
    Ux4 =   Ux * Ux3;

    Uy2 =   Uy * Uy;
    Uy3 =   Uy * Uy2;
    Uy4 =   Uy * Uy3;

    U0  =   sqrt(1 + Ux2 + Uy2);
    U02 =   U0 * U0;
    U03 =   U0 * U02;
    U04 =   U0 * U03;

    X1     =   X + 1;

    /// OPTIMIZATION END

    // coefficients standing by the time derivatives of fields in all equations
    // for instance: CLt1 describes the coefficient: CLt1 standing in front of: partial_tau Lambda in Equation 1 before inverting matrix

    CLt1 = (-6*Ch*U0)/L;

    CXt1 = (Ch*U0)/X1;

    CUxt1 = (-2*Ch*Ux)/U0;

    CUyt1 = (-2*Ch*Uy)/U0;

    CTt1 = 2*Sh*U0;


    CLt2 = Ch*Enp*R*U0;

    CXt2 = Ch*En*Rp*U0;

    CUxt2 = (Ch*(EE + PT)*Ux)/U0;

    CUyt2 = (Ch*(EE + PT)*Uy)/U0;

    CTt2 = -((EE + PL)*Sh*U0);


    CLt3 = -(Pp*RL*Sh);

    CXt3 = -(P*RLp*Sh);

//    CUxt3 = 0;

//    CUyt3 = 0;

    CTt3 = Ch*(EE*U02 + P*(RL + RT*(-1 + U02)));


    CLt4 = Ch*Pp*RT*U0*Ux;

    CXt4 = Ch*P*RTp*U0*Ux;

    CUxt4 = Ch*(EE + PT)*U0;

//    CUyt4 = 0;

    CTt4 = P*(RL - RT)*Sh*U0*Ux;


    CLt5 = Ch*Pp*RT*U0*Uy;

    CXt5 = Ch*P*RTp*U0*Uy;

//    CUxt5 = 0;

    CUyt5 = Ch*(EE + PT)*U0;

    CTt5 = P*(RL - RT)*Sh*U0*Uy;


    G   =   0.4 * pow(R, 0.25) * L / eb; // \Gamma = 2 L / (5 eb)

    Col =   2 * G * (pow(R, 0.75) * sqrt(X1) - 1); // collisional term

    R1 = -2*DUxDx - 2*DUyDy - (2*Ch*DTDe*U0)/t + (6*DLDe*Sh*U0)/(L*t) - (6*DLDx*Ux)/L + (2*DUxDe*Sh*Ux)/(t*U0) - (6*DLDy*Uy)/L + (2*DUyDe*Sh*Uy)/(t*U0) - (DXDe*Sh*U0)/(t*X1) + (DXDx*Ux)/X1 + (DXDy*Uy)/X1;

    R1 += Col;

    R2 = DUxDx*(EE + PT) + DUyDy*(EE + PT) + (Ch*DTDe*(EE + PL)*U0)/t - (DLDe*Enp*R*Sh*U0)/t - (DXDe*En*Rp*Sh*U0)/t + DLDx*Enp*R*Ux + DXDx*En*Rp*Ux - (DUxDe*(EE + PT)*Sh*Ux)/(t*U0) + DLDy*Enp*R*Uy + DXDy*En*Rp*Uy -
   (DUyDe*(EE + PT)*Sh*Uy)/(t*U0);

    R3 = (Ch*DLDe*Pp*RL)/t + (Ch*DXDe*P*RLp)/t - (DTDe*Sh*(P*(RL + RT*(-1 + U02)) + EE*U02))/t + DTDx*(EE + PT)*U0*Ux + DTDy*(EE + PT)*U0*Uy;

    R4 = -((DUxDe*(EE + PT)*Sh*U0)/t) + DUxDx*(EE + PT)*Ux + (Ch*DTDe*P*(-RL + RT)*U0*Ux)/t - (DLDe*Pp*RT*Sh*U0*Ux)/t - (DXDe*P*RTp*Sh*U0*Ux)/t + DLDx*Pp*RT*(1 + Ux2) + DXDx*P*RTp*(1 + Ux2) + DUxDy*(EE + PT)*Uy +
   DLDy*Pp*RT*Ux*Uy + DXDy*P*RTp*Ux*Uy;

    R5 = -((DUyDe*(EE + PT)*Sh*U0)/t) + DUyDx*(EE + PT)*Ux + DUyDy*(EE + PT)*Uy + (Ch*DTDe*P*(-RL + RT)*U0*Uy)/t - (DLDe*Pp*RT*Sh*U0*Uy)/t - (DXDe*P*RTp*Sh*U0*Uy)/t + DLDx*Pp*RT*Ux*Uy + DXDx*P*RTp*Ux*Uy +
   DLDy*Pp*RT*(1 + Uy2) + DXDy*P*RTp*(1 + Uy2);


/// normalize both sides of the equation by a large number in order to avoid underflow during evolution

    switch(mEOSChoice){
        case 1: // ideal
            norm = pow(2.,30);
            break;
        case 2: // realistic
            norm = (L > 1.0e-2 ? pow(2.,20) : pow(2.,100));
            break;
    }



    R1 *= norm;
    R2 *= norm;
    R3 *= norm;
    R4 *= norm;
    R5 *= norm;
    CLt1 *= norm;
    CXt1 *= norm;
    CUxt1 *= norm;
    CUyt1 *= norm;
    CTt1 *= norm;
    CLt2 *= norm;
    CXt2 *= norm;
    CUxt2 *= norm;
    CUyt2 *= norm;
    CTt2 *= norm;
    CLt3 *= norm;
    CXt3 *= norm;
    CTt3 *= norm;
    CLt4 *= norm;
    CXt4 *= norm;
    CUxt4 *= norm;
    CTt4 *= norm;
    CLt5 *= norm;
    CXt5 *= norm;
    CUyt5 *= norm;
    CTt5 *= norm;

/// end normalize

    denLambdaUxUy = (-(CLt4*CTt3*CUxt2*CUyt5*CXt1) + CLt2*CTt3*CUxt4*CUyt5*CXt1 + CLt4*CTt3*CUxt1*CUyt5*CXt2 - CLt1*CTt3*CUxt4*CUyt5*CXt2 - CLt4*CTt5*CUxt2*CUyt1*CXt3 + CLt2*CTt5*CUxt4*CUyt1*CXt3 + CLt4*CTt5*CUxt1*CUyt2*CXt3 -
     CLt1*CTt5*CUxt4*CUyt2*CXt3 - CLt4*CTt2*CUxt1*CUyt5*CXt3 + CLt2*CTt4*CUxt1*CUyt5*CXt3 + CLt4*CTt1*CUxt2*CUyt5*CXt3 - CLt1*CTt4*CUxt2*CUyt5*CXt3 - CLt2*CTt1*CUxt4*CUyt5*CXt3 + CLt1*CTt2*CUxt4*CUyt5*CXt3 -
     CLt2*CTt3*CUxt1*CUyt5*CXt4 + CLt1*CTt3*CUxt2*CUyt5*CXt4 + CLt5*((CTt4*CUxt2*CUyt1 - CTt2*CUxt4*CUyt1 - CTt4*CUxt1*CUyt2 + CTt1*CUxt4*CUyt2)*CXt3 +
        CTt3*(-(CUxt4*CUyt2*CXt1) + CUxt4*CUyt1*CXt2 - CUxt2*CUyt1*CXt4 + CUxt1*CUyt2*CXt4)) + CLt4*CTt3*CUxt2*CUyt1*CXt5 - CLt2*CTt3*CUxt4*CUyt1*CXt5 - CLt4*CTt3*CUxt1*CUyt2*CXt5 + CLt1*CTt3*CUxt4*CUyt2*CXt5 +
     CLt3*(-(CTt2*CUxt4*CUyt5*CXt1) + CTt1*CUxt4*CUyt5*CXt2 + CTt2*CUxt1*CUyt5*CXt4 - CTt1*CUxt2*CUyt5*CXt4 + CTt5*(CUxt4*CUyt2*CXt1 - CUxt4*CUyt1*CXt2 + CUxt2*CUyt1*CXt4 - CUxt1*CUyt2*CXt4) +
        CTt2*CUxt4*CUyt1*CXt5 - CTt1*CUxt4*CUyt2*CXt5 + CTt4*(CUxt2*CUyt5*CXt1 - CUxt1*CUyt5*CXt2 - CUxt2*CUyt1*CXt5 + CUxt1*CUyt2*CXt5)));

    denTheta = ((CLt5*CTt3*CUxt4*CUyt2 + CTt3*(CLt4*CUxt2 - CLt2*CUxt4)*CUyt5 - CLt3*(CTt5*CUxt4*CUyt2 + CTt4*CUxt2*CUyt5 - CTt2*CUxt4*CUyt5))*
        ((CLt5*CUxt4*CUyt1 + CLt4*CUxt1*CUyt5 - CLt1*CUxt4*CUyt5)*CXt3 + CLt3*(CUxt4*CUyt5*CXt1 - CUxt1*CUyt5*CXt4 - CUxt4*CUyt1*CXt5)) -
       (CLt5*CTt3*CUxt4*CUyt1 + CTt3*(CLt4*CUxt1 - CLt1*CUxt4)*CUyt5 - CLt3*(CTt5*CUxt4*CUyt1 + CTt4*CUxt1*CUyt5 - CTt1*CUxt4*CUyt5))*
        ((CLt5*CUxt4*CUyt2 + CLt4*CUxt2*CUyt5 - CLt2*CUxt4*CUyt5)*CXt3 + CLt3*(CUxt4*CUyt5*CXt2 - CUxt2*CUyt5*CXt4 - CUxt4*CUyt2*CXt5)));

    denXi1 =  ((CLt5*CTt3*CUxt4*CUyt2 + CTt3*(CLt4*CUxt2 - CLt2*CUxt4)*CUyt5 - CLt3*(CTt5*CUxt4*CUyt2 + CTt4*CUxt2*CUyt5 - CTt2*CUxt4*CUyt5))*
         ((CLt5*CUxt4*CUyt1 + CLt4*CUxt1*CUyt5 - CLt1*CUxt4*CUyt5)*CXt3 + CLt3*(CUxt4*CUyt5*CXt1 - CUxt1*CUyt5*CXt4 - CUxt4*CUyt1*CXt5)) -
        (CLt5*CTt3*CUxt4*CUyt1 + CTt3*(CLt4*CUxt1 - CLt1*CUxt4)*CUyt5 - CLt3*(CTt5*CUxt4*CUyt1 + CTt4*CUxt1*CUyt5 - CTt1*CUxt4*CUyt5))*
         ((CLt5*CUxt4*CUyt2 + CLt4*CUxt2*CUyt5 - CLt2*CUxt4*CUyt5)*CXt3 + CLt3*(CUxt4*CUyt5*CXt2 - CUxt2*CUyt5*CXt4 - CUxt4*CUyt2*CXt5)));

    denXi2 = ((CLt5*CUxt4*CUyt1 + CLt4*CUxt1*CUyt5 - CLt1*CUxt4*CUyt5)*CXt3 + CLt3*(CUxt4*CUyt5*CXt1 - CUxt1*CUyt5*CXt4 - CUxt4*CUyt1*CXt5));

    // if some denominator is 0 then set the lowest possible value that can be expressed on the machine's architecture

    if(denLambdaUxUy == 0)  denLambdaUxUy = tiny;
    if(denTheta == 0)       denTheta = tiny;
    if(denXi1 == 0)         denXi1 = tiny;
    if(denXi2 == 0)         denXi2 = tiny;

}

double AnisotropicFluidEquations::GetRHSValue(const usi aEquation){

    PRINT_DEBUG_3("<AnisotropicFluidEquations::GetRHSValue>");

// imported from Mathematica notebook and optimized

switch(aEquation){

    case 0: // \Lambda simplified with MATHEMATICA

    return (
        (-(CTt4*CUxt2*CUyt5*CXt3*R1) + CTt2*CUxt4*CUyt5*CXt3*R1 + CTt4*CUxt1*CUyt5*CXt3*R2 - CTt1*CUxt4*CUyt5*CXt3*R2 + CTt4*CUxt2*CUyt5*CXt1*R3 - CTt2*CUxt4*CUyt5*CXt1*R3 - CTt4*CUxt1*CUyt5*CXt2*R3 +
             CTt1*CUxt4*CUyt5*CXt2*R3 + CTt2*CUxt1*CUyt5*CXt4*R3 - CTt1*CUxt2*CUyt5*CXt4*R3 - CTt4*CUxt2*CUyt1*CXt5*R3 + CTt2*CUxt4*CUyt1*CXt5*R3 + CTt4*CUxt1*CUyt2*CXt5*R3 - CTt1*CUxt4*CUyt2*CXt5*R3 -
             CTt2*CUxt1*CUyt5*CXt3*R4 + CTt1*CUxt2*CUyt5*CXt3*R4 + CTt5*(CUxt4*(-(CUyt2*CXt3*R1) + CUyt1*CXt3*R2 + CUyt2*CXt1*R3 - CUyt1*CXt2*R3) + (CUxt2*CUyt1 - CUxt1*CUyt2)*(CXt4*R3 - CXt3*R4)) +
             CTt4*CUxt2*CUyt1*CXt3*R5 - CTt2*CUxt4*CUyt1*CXt3*R5 - CTt4*CUxt1*CUyt2*CXt3*R5 + CTt1*CUxt4*CUyt2*CXt3*R5 +
             CTt3*(CUxt4*(-(CUyt5*CXt2*R1) + CUyt2*CXt5*R1 + CUyt5*CXt1*R2 - CUyt1*CXt5*R2 - CUyt2*CXt1*R5 + CUyt1*CXt2*R5) + CUxt2*(CUyt5*CXt4*R1 - CUyt5*CXt1*R4 + CUyt1*CXt5*R4 - CUyt1*CXt4*R5) +
                CUxt1*(-(CUyt5*CXt4*R2) + CUyt5*CXt2*R4 - CUyt2*CXt5*R4 + CUyt2*CXt4*R5))) / (- denLambdaUxUy)
    );

    break;

    case 1: // \theta simplified with MATHEMATICA

    return (
        -(
          (-(((CLt5*CUxt4*CUyt2 + CLt4*CUxt2*CUyt5 - CLt2*CUxt4*CUyt5)*CXt3 + CLt3*(CUxt4*CUyt5*CXt2 - CUxt2*CUyt5*CXt4 - CUxt4*CUyt2*CXt5))*
                  ((CLt5*CUxt4*CUyt1 + CLt4*CUxt1*CUyt5 - CLt1*CUxt4*CUyt5)*R3 + CLt3*(CUxt4*CUyt5*R1 - CUxt1*CUyt5*R4 - CUxt4*CUyt1*R5))) +
               ((CLt5*CUxt4*CUyt1 + CLt4*CUxt1*CUyt5 - CLt1*CUxt4*CUyt5)*CXt3 + CLt3*(CUxt4*CUyt5*CXt1 - CUxt1*CUyt5*CXt4 - CUxt4*CUyt1*CXt5))*
                ((CLt5*CUxt4*CUyt2 + CLt4*CUxt2*CUyt5 - CLt2*CUxt4*CUyt5)*R3 + CLt3*(CUxt4*CUyt5*R2 - CUxt2*CUyt5*R4 - CUxt4*CUyt2*R5)))
          / denTheta
        )
    );

    break;

    case 2: // Ux simplified with MATHEMATICA

    return (
    (CLt5*CTt4*CUyt2*CXt3*R1 - CLt2*CTt4*CUyt5*CXt3*R1 - CLt5*CTt3*CUyt2*CXt4*R1 + CLt2*CTt3*CUyt5*CXt4*R1 - CLt5*CTt4*CUyt1*CXt3*R2 + CLt1*CTt4*CUyt5*CXt3*R2 + CLt5*CTt3*CUyt1*CXt4*R2 - CLt1*CTt3*CUyt5*CXt4*R2 -
         CLt5*CTt4*CUyt2*CXt1*R3 + CLt2*CTt4*CUyt5*CXt1*R3 + CLt5*CTt4*CUyt1*CXt2*R3 - CLt1*CTt4*CUyt5*CXt2*R3 - CLt5*CTt2*CUyt1*CXt4*R3 + CLt2*CTt5*CUyt1*CXt4*R3 + CLt5*CTt1*CUyt2*CXt4*R3 - CLt1*CTt5*CUyt2*CXt4*R3 -
         CLt2*CTt1*CUyt5*CXt4*R3 + CLt1*CTt2*CUyt5*CXt4*R3 - CLt2*CTt4*CUyt1*CXt5*R3 + CLt1*CTt4*CUyt2*CXt5*R3 + CLt5*CTt3*CUyt2*CXt1*R4 - CLt2*CTt3*CUyt5*CXt1*R4 - CLt5*CTt3*CUyt1*CXt2*R4 + CLt1*CTt3*CUyt5*CXt2*R4 +
         CLt5*CTt2*CUyt1*CXt3*R4 - CLt2*CTt5*CUyt1*CXt3*R4 - CLt5*CTt1*CUyt2*CXt3*R4 + CLt1*CTt5*CUyt2*CXt3*R4 + CLt2*CTt1*CUyt5*CXt3*R4 - CLt1*CTt2*CUyt5*CXt3*R4 + CLt2*CTt3*CUyt1*CXt5*R4 - CLt1*CTt3*CUyt2*CXt5*R4 +
         CLt2*CTt4*CUyt1*CXt3*R5 - CLt1*CTt4*CUyt2*CXt3*R5 - CLt2*CTt3*CUyt1*CXt4*R5 + CLt1*CTt3*CUyt2*CXt4*R5 +
         CLt3*(-(CTt2*CUyt5*CXt4*R1) + CTt1*CUyt5*CXt4*R2 + CTt2*CUyt5*CXt1*R4 - CTt1*CUyt5*CXt2*R4 - CTt2*CUyt1*CXt5*R4 + CTt1*CUyt2*CXt5*R4 + CTt5*(CUyt2*CXt4*R1 - CUyt1*CXt4*R2 - CUyt2*CXt1*R4 + CUyt1*CXt2*R4) +
            CTt2*CUyt1*CXt4*R5 - CTt1*CUyt2*CXt4*R5 + CTt4*(CUyt5*CXt2*R1 - CUyt2*CXt5*R1 - CUyt5*CXt1*R2 + CUyt1*CXt5*R2 + CUyt2*CXt1*R5 - CUyt1*CXt2*R5)) +
         CLt4*(CTt2*CUyt5*CXt3*R1 - CTt1*CUyt5*CXt3*R2 - CTt2*CUyt5*CXt1*R3 + CTt1*CUyt5*CXt2*R3 + CTt2*CUyt1*CXt5*R3 - CTt1*CUyt2*CXt5*R3 + CTt5*(-(CUyt2*CXt3*R1) + CUyt1*CXt3*R2 + CUyt2*CXt1*R3 - CUyt1*CXt2*R3) -
            CTt2*CUyt1*CXt3*R5 + CTt1*CUyt2*CXt3*R5 + CTt3*(-(CUyt5*CXt2*R1) + CUyt2*CXt5*R1 + CUyt5*CXt1*R2 - CUyt1*CXt5*R2 - CUyt2*CXt1*R5 + CUyt1*CXt2*R5))) / denLambdaUxUy
    );

    break;

    case 3: // Uy simplified with MATHEMATICA

    return (
        (CLt4*CTt5*CUxt2*CXt3*R1 - CLt2*CTt5*CUxt4*CXt3*R1 - CLt4*CTt3*CUxt2*CXt5*R1 + CLt2*CTt3*CUxt4*CXt5*R1 - CLt4*CTt5*CUxt1*CXt3*R2 + CLt1*CTt5*CUxt4*CXt3*R2 + CLt4*CTt3*CUxt1*CXt5*R2 - CLt1*CTt3*CUxt4*CXt5*R2 -
             CLt4*CTt5*CUxt2*CXt1*R3 + CLt2*CTt5*CUxt4*CXt1*R3 + CLt4*CTt5*CUxt1*CXt2*R3 - CLt1*CTt5*CUxt4*CXt2*R3 - CLt2*CTt5*CUxt1*CXt4*R3 + CLt1*CTt5*CUxt2*CXt4*R3 - CLt4*CTt2*CUxt1*CXt5*R3 + CLt2*CTt4*CUxt1*CXt5*R3 +
             CLt4*CTt1*CUxt2*CXt5*R3 - CLt1*CTt4*CUxt2*CXt5*R3 - CLt2*CTt1*CUxt4*CXt5*R3 + CLt1*CTt2*CUxt4*CXt5*R3 + CLt2*CTt5*CUxt1*CXt3*R4 - CLt1*CTt5*CUxt2*CXt3*R4 - CLt2*CTt3*CUxt1*CXt5*R4 + CLt1*CTt3*CUxt2*CXt5*R4 +
             CLt5*(CTt2*CUxt4*CXt3*R1 - CTt1*CUxt4*CXt3*R2 - CTt2*CUxt4*CXt1*R3 + CTt1*CUxt4*CXt2*R3 + CTt2*CUxt1*CXt4*R3 - CTt1*CUxt2*CXt4*R3 + CTt4*(-(CUxt2*CXt3*R1) + CUxt1*CXt3*R2 + CUxt2*CXt1*R3 - CUxt1*CXt2*R3) -
                CTt2*CUxt1*CXt3*R4 + CTt1*CUxt2*CXt3*R4 + CTt3*(-(CUxt4*CXt2*R1) + CUxt2*CXt4*R1 + CUxt4*CXt1*R2 - CUxt1*CXt4*R2 - CUxt2*CXt1*R4 + CUxt1*CXt2*R4)) + CLt4*CTt3*CUxt2*CXt1*R5 - CLt2*CTt3*CUxt4*CXt1*R5 -
             CLt4*CTt3*CUxt1*CXt2*R5 + CLt1*CTt3*CUxt4*CXt2*R5 + CLt4*CTt2*CUxt1*CXt3*R5 - CLt2*CTt4*CUxt1*CXt3*R5 - CLt4*CTt1*CUxt2*CXt3*R5 + CLt1*CTt4*CUxt2*CXt3*R5 + CLt2*CTt1*CUxt4*CXt3*R5 - CLt1*CTt2*CUxt4*CXt3*R5 +
             CLt2*CTt3*CUxt1*CXt4*R5 - CLt1*CTt3*CUxt2*CXt4*R5 + CLt3*(-(CTt2*CUxt4*CXt5*R1) + CTt1*CUxt4*CXt5*R2 + CTt2*CUxt1*CXt5*R4 - CTt1*CUxt2*CXt5*R4 +
                CTt5*(CUxt4*CXt2*R1 - CUxt2*CXt4*R1 - CUxt4*CXt1*R2 + CUxt1*CXt4*R2 + CUxt2*CXt1*R4 - CUxt1*CXt2*R4) + CTt2*CUxt4*CXt1*R5 - CTt1*CUxt4*CXt2*R5 - CTt2*CUxt1*CXt4*R5 + CTt1*CUxt2*CXt4*R5 +
                CTt4*(CUxt2*CXt5*R1 - CUxt1*CXt5*R2 - CUxt2*CXt1*R5 + CUxt1*CXt2*R5))) / denLambdaUxUy
    );

    break;

    case 4: // \xi simplified with MATHEMATICA

        return (
                (
                 -((CLt5*CUxt4*CUyt1 + CLt4*CUxt1*CUyt5 - CLt1*CUxt4*CUyt5)*R3) + CLt3*(-(CUxt4*CUyt5*R1) + CUxt1*CUyt5*R4 + CUxt4*CUyt1*R5)

                 +

                 ((CLt5*CTt3*CUxt4*CUyt1 + CTt3*(CLt4*CUxt1 - CLt1*CUxt4)*CUyt5 - CLt3*(CTt5*CUxt4*CUyt1 + CTt4*CUxt1*CUyt5 - CTt1*CUxt4*CUyt5))*
                  (-(((CLt5*CUxt4*CUyt2 + CLt4*CUxt2*CUyt5 - CLt2*CUxt4*CUyt5)*CXt3 + CLt3*(CUxt4*CUyt5*CXt2 - CUxt2*CUyt5*CXt4 - CUxt4*CUyt2*CXt5))*
                     ((CLt5*CUxt4*CUyt1 + CLt4*CUxt1*CUyt5 - CLt1*CUxt4*CUyt5)*R3 + CLt3*(CUxt4*CUyt5*R1 - CUxt1*CUyt5*R4 - CUxt4*CUyt1*R5))) +
                   ((CLt5*CUxt4*CUyt1 + CLt4*CUxt1*CUyt5 - CLt1*CUxt4*CUyt5)*CXt3 + CLt3*(CUxt4*CUyt5*CXt1 - CUxt1*CUyt5*CXt4 - CUxt4*CUyt1*CXt5))*
                   ((CLt5*CUxt4*CUyt2 + CLt4*CUxt2*CUyt5 - CLt2*CUxt4*CUyt5)*R3 + CLt3*(CUxt4*CUyt5*R2 - CUxt2*CUyt5*R4 - CUxt4*CUyt2*R5))))
                 / denXi1
                 )/ denXi2
                );
    break;

    default:

    return -666;

    break;
}
}

inline double AnisotropicFluidEquations::energy(const double aX){

    PRINT_DEBUG_3("<AnisotropicFluidEquations::energy>");

    return mEquationOfState->ET(aX);

}

inline double AnisotropicFluidEquations::pressure(const double aX){

    PRINT_DEBUG_3("<AnisotropicFluidEquations::pressure>");

    return mEquationOfState->PT(aX);

}

inline double AnisotropicFluidEquations::pressurep(const double aX){

    PRINT_DEBUG_3("<AnisotropicFluidEquations::pressurep>");

    return mEquationOfState->ST(aX);

}

inline double AnisotropicFluidEquations::energyp(const double aX){

    PRINT_DEBUG_3("<AnisotropicFluidEquations::energyp>");

    return mEquationOfState->EpT(aX);

}

double AnisotropicFluidEquations::r(double x) {

    PRINT_DEBUG_3("<AnisotropicFluidEquations::r>");

	if (x==0) return 1;
	if (x>0) return (1/(1 + x) + atan(sqrt(x))/sqrt(x))/2.;
	if (x<0) return (1/(1 - fabs(x)) + atanh(sqrt(fabs(x)))/sqrt(fabs(x)))/2.;
    return 1;
}

double AnisotropicFluidEquations::rt(double x) {

    PRINT_DEBUG_3("<AnisotropicFluidEquations::rt>");

	if (x==0) return 1;
	if (x>0) return (3*(sqrt(x) + (-1 + x)*atan(sqrt(x))))/(4.*pow(x,1.5));
	if (x<0) return -3/(4.*fabs(x)) + (3*(1 + fabs(x))*atanh(sqrt(fabs(x))))/(4.*pow(fabs(x),1.5));
    return 1;
}

double AnisotropicFluidEquations::rl(double x) {

    PRINT_DEBUG_3("<AnisotropicFluidEquations::rl>");

	if (x==0) return 1;
	if (x>0) return (3*(-(sqrt(x)/(1 + x)) + atan(sqrt(x))))/(2.*pow(x,1.5));
	if (x<0) return 3/(2*fabs(x) - 2*pow(fabs(x),2)) - (3*atanh(sqrt(fabs(x))))/(2.*pow(fabs(x),1.5));
    return 1;
}

double AnisotropicFluidEquations::rp(double x) {

    PRINT_DEBUG_3("<AnisotropicFluidEquations::rp>");

	if (x==0) return -2./3.;
	if (x>0) return -(-1 + x)/(4.*x*pow(1 + x,2)) - atan(sqrt(x))/(4.*pow(x,1.5));
	if (x<0) return (-2/pow(-1 + fabs(x),2) + 1/((-1 + fabs(x))*fabs(x)) + atanh(sqrt(fabs(x)))/pow(fabs(x),1.5))/4.;
    return -2./3.;
}

double AnisotropicFluidEquations::rtp(double x) {

    PRINT_DEBUG_3("<AnisotropicFluidEquations::rtp>");

	if (x==0) return -2./5.;
	if (x>0) return (-3*(3 + x))/(8.*pow(x,2)*(1 + x)) - (3*(-3 + x)*atan(sqrt(x)))/(8.*pow(x,2.5));
	if (x<0) return (3*(3*sqrt(fabs(x)) - pow(fabs(x),1.5) - 3*atanh(sqrt(fabs(x))) + pow(x,2)*atanh(sqrt(fabs(x))) + 2*fabs(x)*atanh(sqrt(fabs(x)))))/(8.*(-1 + fabs(x))*pow(fabs(x),2.5));
    return -2./5.;
}

double AnisotropicFluidEquations::rlp(double x) {

    PRINT_DEBUG_3("<AnisotropicFluidEquations::rlp>");

	if (x==0) return -6./5.;
	if (x>0) return (3*(3 + 5*x))/(4.*pow(x,2)*pow(1 + x,2)) - (9*atan(sqrt(x)))/(4.*pow(x,2.5));
	if (x<0) return (-3*(-3*sqrt(fabs(x)) + 5*pow(fabs(x),1.5) + 3*atanh(sqrt(fabs(x))) + 3*pow(x,2)*atanh(sqrt(fabs(x))) - 6*fabs(x)*atanh(sqrt(fabs(x)))))/(4.*pow(-1 + fabs(x),2)*pow(fabs(x),2.5));
    return -6./5.;
}

inline void AnisotropicFluidEquations::PrintInfo(void){

    PRINT_DEBUG_3("<AnisotropicFluidEquations::PrintInfo>");

    PrintRawMessage("-> Anisotropic Fluid (LO)");
    PrintRawMessage("\tEta bar", d2s(eb));

}

/*!
*
*  \brief Formats output
*
*  \li aX = output value to be formated
*  \li aN = number of the equation whose output is being formated
*
*/

double AnisotropicFluidEquations::FormatOutput(double aX, int aN){

    PRINT_DEBUG_3("<AnisotropicFluidEquations::FormatOutput>");

    switch(aN){

        case 0: return UnitsFactor * aX;

        default: return aX;
    }
}

/*!
*
*  \brief Formats input
*
*  \li aX = output value to be formated
*  \li aN = number of the equation whose input is being formated
*
*/

double AnisotropicFluidEquations::FormatInput(double aX, int aN){

    PRINT_DEBUG_3("<AnisotropicFluidEquations::FormatOutput>");

    switch(aN){

        default: return aX;
    }
}
