/**
 *
 *   AnisotropicFluidEquationsNew.cxx
 *
 *   Copyright (C) 2017 Mubarak Alqahtani, Mohammad Nopoush, Radoslaw Ryblewski, and Michael Strickland
 *
 *   This file is part of aHydro3p1.
 *
 *   aHydro3p1 is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   aHydro3p1 is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with aHydro3p1.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "Global.h"
#include "RealisticEquationOfState.h"



/*!
 *
 *  \brief Default constructor
 *
 *  \li aConfigurator = pointer to configuration
 *  \li aGrid = pointer to grid
 *
 */

AnisotropicFluidEquationsNew::AnisotropicFluidEquationsNew(Configurator *aConfigurator, Grid *aGrid)
: mNames(0), mGrid(aGrid), mConfigurator(aConfigurator), mEquationOfState(0), eb(mConfigurator->GetDouble("etabar")),
mXP(aGrid->GetXP()), mYP(aGrid->GetYP()), mZP(aGrid->GetZP()),
mdX(aGrid->GetdX()), mdY(aGrid->GetdY()), mdZ(aGrid->GetdZ()),
m2dX(2 * mdX), m2dY(2 * mdY), m2dZ(2 * mdZ),
mdXi(1 / mdX), mdYi(1 / mdY), mdZi(1 / mdZ),
m2dXi(1 / m2dX), m2dYi(1 / m2dY), m2dZi(1 / m2dZ),
mXPm1(mXP - 1), mYPm1(mYP - 1), mZPm1(mZP - 1){
    
    PRINT_DEBUG_3("<AnisotropicFluidEquationsNew::AnisotropicFluidEquationsNew>");
    
    // initialization of names of equations
    
    mNames = new string [mNumberOfEquations];
    
    mNames[0] = "ax";
    mNames[1] = "ay";
    mNames[2] = "az";
    mNames[3] = "ux";
    mNames[4] = "uy";
    mNames[5] = "rap"; // vt
    mNames[6] = "lambda";
    mNames[7] = "temp";
    
    mEOSChoice = mConfigurator->GetInt("EquationOfState");
    
    // setup hfunc p and pm integral interpolations
    hfuncs.loadInterpolations();
    
    // load m(T) from file
    mass.loadInterpolations(mConfigurator->GetParameter("RealisticEquationOfStateDirectory"));
    
    // load etabar class based on parameterization type
    switch (mConfigurator->GetInt("Etabar_type")) {
        case 0:
            etabar = LinearTransportCoefficient(eb,mConfigurator->GetDouble("Slope1"),mConfigurator->GetDouble("Slope2"),mConfigurator->GetDouble("Tc"));
            break;
        default:
            PrintRawMessage("Unknown Etabar_type; using constant etabar instead");
            etabar = LinearTransportCoefficient(eb,0,0,0);
            break;
    }
    
 
}

/*!
 *
 *  \brief Default destructor
 *
 */

AnisotropicFluidEquationsNew::~AnisotropicFluidEquationsNew(){
    
    PRINT_DEBUG_3("<AnisotropicFluidEquationsNew::~AnisotropicFluidEquationsNew>");
    
    if(mNames){
        delete [] mNames;
    }
    mNames = 0;
    
    // free hfunc associated interpolations
    hfuncs.freeInterpolations();
    
    // free thermal mass interpolations
    mass.freeInterpolations();
    
    // free relaxation time interpolation
    tq.freeInterpolation();
    
}

/*!
 *
 *  \brief Sets equation of state
 *
 *  \li aEquationOfState = pointer to equation of state
 *
 */

void AnisotropicFluidEquationsNew::SetEquationOfState(EquationOfState *aEquationOfState){
    
    PRINT_DEBUG_3("<AnisotropicFluidEquationsNew::SetEquationOfState>");
    
    mEquationOfState = aEquationOfState;
    
    tq.setup(hfuncs,mass,etabar);
    
}

/*!
 *
 *  \brief Returns number of equations
 *
 */

usi AnisotropicFluidEquationsNew::GetNumberOfEquations(void){
    
    PRINT_DEBUG_3("<AnisotropicFluidEquationsNew::GetNumberOfEquations>");
    
    return mNumberOfEquations;
    
}

/*!
 *
 *  \brief Returns name of equation
 *
 *  \li aNumberOfEquation = number of equation
 *
 */

string AnisotropicFluidEquationsNew::GetName(usi aNumberOfEquation){
    
    PRINT_DEBUG_3("<AnisotropicFluidEquationsNew::GetName>");
    
    return mNames[aNumberOfEquation];
    
}


/*!
 *
 *  \brief Load square matrix A and column vector b for solution of A x = B
 *
 *  \li A = pointer to flattened double array of dim*dim matrix
 *  \li B = pointer to doubled array for dim column vector
 *
 */

void AnisotropicFluidEquationsNew::LoadAB(double * A, double * B) {
    
    PRINT_DEBUG_3("<AnisotropicFluidEquationsNew::LoadAB>");
    
    A[0] = Ch*d1H3*L4*u0; // 1,1
    A[1] = Ch*d2H3*L4*u0; // 1,2
    A[2] = Ch*d3H3*L4*u0; // 1,3
    A[3] = (Ch*(H3 + H3X)*L4*ux)/u0; // 1,4
    A[4] = (Ch*(H3 + H3X)*L4*uy)/u0; // 1,5
    A[5] = -((H3 + H3L)*L4*Sh*u0); // 1,6
    A[6] = Ch*L2*(4*H3*L - d4H3*m)*u0; // 1,7
    A[7] = Ch*(dTB + d4H3*dTm*L3)*u0; // 1,8
    A[8] = -(Ch*d1H3X*L4*up); // 2,1
    A[9] = -(Ch*d2H3X*L4*up); // 2,2
    A[10] = -(Ch*d3H3X*L4*up); // 2,3
    A[11] = -((Ch*(H3 + H3X)*L4*ux)/up); // 2,4
    A[12] = -((Ch*(H3 + H3X)*L4*uy)/up); // 2,5
    A[13] = -((H3L - H3X)*L4*Sh*up); // 2,6
    A[14] = Ch*L2*(-4*H3X*L + d4H3X*m)*up; // 2,7
    A[15] = Ch*(dTB - d4H3X*dTm*L3)*up; // 2,8
    A[16] = 0; // 3,1
    A[17] = 0; // 3,2
    A[18] = 0; // 3,3
    A[19] = (Ch*(H3 + H3X)*L4*u0*uy)/up; // 3,4
    A[20] = -((Ch*(H3 + H3X)*L4*u0*ux)/up); // 3,5
    A[21] = 0; // 3,6
    A[22] = 0; // 3,7
    A[23] = 0; // 3,8
    A[24] = d1H3L*L4*Sh; // 4,1
    A[25] = d2H3L*L4*Sh; // 4,2
    A[26] = d3H3L*L4*Sh; // 4,3
    A[27] = 0; // 4,4
    A[28] = 0; // 4,5
    A[29] = -(Ch*L4*(H3L + H3*u02 + H3X*up2)); // 4,6
    A[30] = L2*(4*H3L*L - d4H3L*m)*Sh; // 4,7
    A[31] = (-dTB + d4H3L*dTm*L3)*Sh; // 4,8
    A[32] = (3*ax2*ay*az*Ch*gFac*K3*L2*m3*u0)/8.; // 5,1
    A[33] = (ax3*az*Ch*gFac*K3*L2*m3*u0)/8.; // 5,2
    A[34] = (ax3*ay*Ch*gFac*K3*L2*m3*u0)/8.; // 5,3
    A[35] = (3*ax3*ay*az*Ch*gFac*K3*L2*m3*ux)/(8.*u0); // 5,4
    A[36] = (3*ax3*ay*az*Ch*gFac*K3*L2*m3*uy)/(8.*u0); // 5,5
    A[37] = -(ax3*ay*az*gFac*K3*L2*m3*Sh*u0)/8.; // 5,6
    A[38] = (ax3*ay*az*Ch*gFac*(4*K3*L + (K2 + K4)*m)*m3*u0)/16.; // 5,7
    A[39] = (ax3*ay*az*Ch*dTm*gFac*L*(6*K3*L - (K2 + K4)*m)*m2*u0)/16.; // 5,8
    A[40] = (ay3*az*Ch*gFac*K3*L2*m3*u0)/8.; // 6,1
    A[41] = (3*ax*ay2*az*Ch*gFac*K3*L2*m3*u0)/8.; // 6,2
    A[42] = (ax*ay3*Ch*gFac*K3*L2*m3*u0)/8.; // 6,3
    A[43] = (ax*ay3*az*Ch*gFac*K3*L2*m3*ux)/(8.*u0); // 6,4
    A[44] = (ax*ay3*az*Ch*gFac*K3*L2*m3*uy)/(8.*u0); // 6,5
    A[45] = -(ax*ay3*az*gFac*K3*L2*m3*Sh*u0)/8.; // 6,6
    A[46] = (ax*ay3*az*Ch*gFac*(4*K3*L + (K2 + K4)*m)*m3*u0)/16.; // 6,7
    A[47] = (ax*ay3*az*Ch*dTm*gFac*L*(6*K3*L - (K2 + K4)*m)*m2*u0)/16.; // 6,8
    A[48] = (ay*az3*Ch*gFac*K3*L2*m3*u0)/8.; // 7,1
    A[49] = (ax*az3*Ch*gFac*K3*L2*m3*u0)/8.; // 7,2
    A[50] = (3*ax*ay*az2*Ch*gFac*K3*L2*m3*u0)/8.; // 7,3
    A[51] = (ax*ay*az3*Ch*gFac*K3*L2*m3*ux)/(8.*u0); // 7,4
    A[52] = (ax*ay*az3*Ch*gFac*K3*L2*m3*uy)/(8.*u0); // 7,5
    A[53] = (-3*ax*ay*az3*gFac*K3*L2*m3*Sh*u0)/8.; // 7,6
    A[54] = (ax*ay*az3*Ch*gFac*(4*K3*L + (K2 + K4)*m)*m3*u0)/16.; // 7,7
    A[55] = (ax*ay*az3*Ch*dTm*gFac*L*(6*K3*L - (K2 + K4)*m)*m2*u0)/16.; // 7,8
    A[56] = d1H3*L4; // 8,1
    A[57] = d2H3*L4; // 8,2
    A[58] = d3H3*L4; // 8,3
    A[59] = 0; // 8,4
    A[60] = 0; // 8,5
    A[61] = 0; // 8,6
    A[62] = L2*(4*H3*L - d4H3*m); // 8,7
    A[63] = d4H3*dTm*L3 + (-4*H3eq*T + d1H3eq*(m - dTm*T))*T2; // 8,8
    
    B[0] = -((Ch*d4vt*(H3 + H3L)*L4*u02*up2 - Sh*up2*(d4T*(dTB + d4H3*dTm*L3)*u02 + L2*(d4L*(4*H3*L - d4H3*m)*u02 + L2*(d1H3*d4ax*u02 + d2H3*d4ay*u02 + d3H3*d4az*u02 + d4ux*H3*ux + d4ux*H3X*ux + d4uy*H3*uy + d4uy*H3X*uy))) + tau*u0*(d2T*dTB*ux3 + d2T*d4H3*dTm*L3*ux3 + 4*d2L*H3*L3*ux3 + d1H3*d2ax*L4*ux3 + d2ay*d2H3*L4*ux3 + d2az*d3H3*L4*ux3 - d2L*d4H3*L2*m*ux3 + d2uy*H3X*L4*ux*uy + d3ux*H3X*L4*ux*uy - d2uy*H3Y*L4*ux*uy - d3ux*H3Y*L4*ux*uy + d3T*dTB*ux2*uy + d3T*d4H3*dTm*L3*ux2*uy + 4*d3L*H3*L3*ux2*uy + d1H3*d3ax*L4*ux2*uy + d2H3*d3ay*L4*ux2*uy + d3az*d3H3*L4*ux2*uy - d3L*d4H3*L2*m*ux2*uy + d2T*dTB*ux*uy2 + d2T*d4H3*dTm*L3*ux*uy2 + 4*d2L*H3*L3*ux*uy2 + d1H3*d2ax*L4*ux*uy2 + d2ay*d2H3*L4*ux*uy2 + d2az*d3H3*L4*ux*uy2 - d2L*d4H3*L2*m*ux*uy2 + d3uy*L4*(H3*up2 + H3Y*ux2 + H3X*uy2) + d2ux*L4*(H3*up2 + H3X*ux2 + H3Y*uy2) + d3T*dTB*uy3 + d3T*d4H3*dTm*L3*uy3 + 4*d3L*H3*L3*uy3 + d1H3*d3ax*L4*uy3 + d2H3*d3ay*L4*uy3 + d3az*d3H3*L4*uy3 - d3L*d4H3*L2*m*uy3))/(tau*u0*up2)); // 1,9
    
    
    B[1] = -((Ch*d4vt*(H3L - H3X)*L4*up5 + Sh*up3*(d4T*(-dTB + d4H3X*dTm*L3)*up2 + L2*(d4L*(4*H3X*L - d4H3X*m)*up2 + L2*(d1H3X*d4ax*up2 + d2H3X*d4ay*up2 + d4ux*H3*ux + d4ux*H3X*ux + d3H3X*d4az*ux2 + d4uy*H3*uy + d4uy*H3X*uy + d3H3X*d4az*uy2))) - (tau*up*(-(d2T*dTB*u02*ux3) + d2T*d4H3X*dTm*L3*u02*ux3 + 4*d2L*H3X*L3*u02*ux3 + d1H3X*d2ax*L4*u02*ux3 + d2ay*d2H3X*L4*u02*ux3 + d2az*d3H3X*L4*u02*ux3 - d2L*d4H3X*L2*m*u02*ux3 - d2uy*H3X*L4*ux*uy - d3ux*H3X*L4*ux*uy + d2uy*H3Y*L4*ux*uy + d3ux*H3Y*L4*ux*uy - d3T*dTB*u02*ux2*uy + d3T*d4H3X*dTm*L3*u02*ux2*uy + 4*d3L*H3X*L3*u02*ux2*uy + d1H3X*d3ax*L4*u02*ux2*uy + d2H3X*d3ay*L4*u02*ux2*uy + d3az*d3H3X*L4*u02*ux2*uy - d3L*d4H3X*L2*m*u02*ux2*uy + d2uy*H3*L4*ux3*uy + d3ux*H3*L4*ux3*uy + d2uy*H3Y*L4*ux3*uy + d3ux*H3Y*L4*ux3*uy - d2T*dTB*u02*ux*uy2 + d2T*d4H3X*dTm*L3*u02*ux*uy2 + 4*d2L*H3X*L3*u02*ux*uy2 + d1H3X*d2ax*L4*u02*ux*uy2 + d2ay*d2H3X*L4*u02*ux*uy2 + d2az*d3H3X*L4*u02*ux*uy2 - d2L*d4H3X*L2*m*u02*ux*uy2 + d3uy*L4*(-(H3Y*u02*ux2) + H3X*(up4 + ux2) + H3*up2*uy2) + d2ux*L4*(H3*up2*ux2 - H3Y*u02*uy2 + H3X*(up4 + uy2)) - d3T*dTB*u02*uy3 + d3T*d4H3X*dTm*L3*u02*uy3 + 4*d3L*H3X*L3*u02*uy3 + d1H3X*d3ax*L4*u02*uy3 + d2H3X*d3ay*L4*u02*uy3 + d3az*d3H3X*L4*u02*uy3 - d3L*d4H3X*L2*m*u02*uy3 + d2uy*H3*L4*ux*uy3 + d3ux*H3*L4*ux*uy3 + d2uy*H3Y*L4*ux*uy3 + d3ux*H3Y*L4*ux*uy3))/u0)/(tau*up4)); // 2,9
    
    
    B[2] = -(((H3 + H3X)*L4*Sh*u0*up2*(d4uy*ux - d4ux*uy) + tau*(d3T*(dTB - d4H3Y*dTm*L3)*up2*ux + d2uy*L4*(H3Y - H3X*u02 - H3*up2)*ux2 - 4*d3L*H3Y*L3*ux3 - d1H3Y*d3ax*L4*ux3 - d2H3Y*d3ay*L4*ux3 - d3az*d3H3Y*L4*ux3 + d3L*d4H3Y*L2*m*ux3 + d2ux*H3X*L4*ux*uy - d3uy*H3X*L4*ux*uy - d2ux*H3Y*L4*ux*uy + d3uy*H3Y*L4*ux*uy - d2T*dTB*ux2*uy + d2T*d4H3Y*dTm*L3*ux2*uy + 4*d2L*H3Y*L3*ux2*uy + d1H3Y*d2ax*L4*ux2*uy + d2ay*d2H3Y*L4*ux2*uy + d2az*d3H3Y*L4*ux2*uy - d2L*d4H3Y*L2*m*ux2*uy + d2ux*H3*L4*ux3*uy - d3uy*H3*L4*ux3*uy + d2ux*H3X*L4*ux3*uy - d3uy*H3X*L4*ux3*uy + d3ux*H3X*L4*uy2 - d3ux*H3Y*L4*uy2 - 4*d3L*H3Y*L3*ux*uy2 - d1H3Y*d3ax*L4*ux*uy2 - d2H3Y*d3ay*L4*ux*uy2 - d3az*d3H3Y*L4*ux*uy2 + d3L*d4H3Y*L2*m*ux*uy2 + d3ux*H3*L4*ux2*uy2 + d3ux*H3X*L4*ux2*uy2 - d2T*dTB*uy3 + d2T*d4H3Y*dTm*L3*uy3 + 4*d2L*H3Y*L3*uy3 + d1H3Y*d2ax*L4*uy3 + d2ay*d2H3Y*L4*uy3 + d2az*d3H3Y*L4*uy3 - d2L*d4H3Y*L2*m*uy3 + d2ux*H3*L4*ux*uy3 - d3uy*H3*L4*ux*uy3 + d2ux*H3X*L4*ux*uy3 - d3uy*H3X*L4*ux*uy3 + d3ux*H3*L4*uy4 + d3ux*H3X*L4*uy4))/(tau*up3)); // 3,9
    
    
    B[3] = -((Ch*(d4T*(dTB - d4H3L*dTm*L3) - L2*(4*d4L*H3L*L + d1H3L*d4ax*L2 + d2H3L*d4ay*L2 + d3H3L*d4az*L2 - d4H3L*d4L*m)) + L4*(d4vt*Sh*(H3L + H3*u02 + H3X*up2) - (H3 + H3X)*tau*u0*(d2vt*ux + d3vt*uy)))/tau); // 4,9
    
    
    B[4] = -(gFac*m2*(2*ax3*ay*az*Ch*d4vt*K3*L2*m*teq*u02*up2 - ax2*Sh*teq*up2*(6*ay*az*d4ax*K3*L2*m*u02 + ax*(2*az*d4ay*K3*L2*m*u02 + ay*(2*d4az*K3*L2*m*u02 + az*(d4T*dTm*L*(6*K3*L - (K2 + K4)*m)*u02 + m*(d4L*(4*K3*L + (K2 + K4)*m)*u02 + 6*K3*L2*(d4ux*ux + d4uy*uy)))))) + tau*u0*(-2*K3eq*m*T2*up2 + 6*ax2*ay*az*K3*L2*m*teq*up2*(d2ax*ux + d3ax*uy) + ax3*(2*az*K3*L2*m*teq*up2*(d2ay*ux + d3ay*uy) + ay*(2*K3*L2*m*teq*up2*(d2az*ux + d3az*uy) + az*((K2 + K4)*m*teq*up2*(-(d2T*dTm*L*ux) + d2L*m*ux - d3T*dTm*L*uy + d3L*m*uy) + 2*K3*L*(2*m*teq*up2*(d2L*ux + d3L*uy) + L*(3*dTm*teq*up2*(d2T*ux + d3T*uy) + m*((1 + 3*d2ux*teq + d3uy*teq)*ux2 + 2*(d2uy + d3ux)*teq*ux*uy + (1 + d2ux*teq + 3*d3uy*teq)*uy2)))))))))/(16.*tau*teq*u0*up2); // 5,9
    
    
    B[5] = -(gFac*m2*(2*ax*ay3*az*Ch*d4vt*K3*L2*m*teq*u02*up2 - ay2*Sh*teq*up2*(2*ay*az*d4ax*K3*L2*m*u02 + ax*(6*az*d4ay*K3*L2*m*u02 + ay*(2*d4az*K3*L2*m*u02 + az*(d4T*dTm*L*(6*K3*L - (K2 + K4)*m)*u02 + m*(d4L*(4*K3*L + (K2 + K4)*m)*u02 + 2*K3*L2*(d4ux*ux + d4uy*uy)))))) + tau*u0*(-2*m*up2*(K3eq*T2 - ay3*az*K3*L2*teq*(d2ax*ux + d3ax*uy)) + ax*ay2*(6*az*K3*L2*m*teq*up2*(d2ay*ux + d3ay*uy) + ay*(2*K3*L2*m*teq*up2*(d2az*ux + d3az*uy) + az*((K2 + K4)*m*teq*up2*(-(d2T*dTm*L*ux) + d2L*m*ux - d3T*dTm*L*uy + d3L*m*uy) + 2*K3*L*(2*m*teq*up2*(d2L*ux + d3L*uy) + L*(3*dTm*teq*up2*(d2T*ux + d3T*uy) + m*((1 + d2ux*teq + 3*d3uy*teq)*ux2 - 2*(d2uy + d3ux)*teq*ux*uy + (1 + 3*d2ux*teq + d3uy*teq)*uy2)))))))))/(16.*tau*teq*u0*up2); // 6,9
    
    
    B[6] = -(gFac*m2*(2*m*(-(K3eq*T2*tau*u0) + ay*K3*L2*teq*(-(az*az2*d4ax*Sh*u02) + az3*tau*u0*(d2ax*ux + d3ax*uy))) + ax*(2*az*az2*K3*L2*m*teq*(-(d4ay*Sh*u02) + tau*u0*(d2ay*ux + d3ay*uy)) + ay*(6*K3*L2*m*teq*(az3*Ch*d4vt*u02 + az2*(-(d4az*Sh*u02) + tau*u0*(d2az*ux + d3az*uy))) + az*az2*((K2 + K4)*m*teq*(d4T*dTm*L*Sh*u02 - d4L*m*Sh*u02 + tau*u0*(-(d2T*dTm*L*ux) + d2L*m*ux - d3T*dTm*L*uy + d3L*m*uy)) + 2*K3*(-(L2*m*Sh*teq*(d4ux*ux + d4uy*uy)) + 2*L*m*teq*(-(d4L*Sh*u02) + tau*u0*(d2L*ux + d3L*uy)) + L2*(m*tau*(1 + d2ux*teq + d3uy*teq)*u0 + 3*dTm*teq*(-(d4T*Sh*u02) + d2T*tau*u0*ux + d3T*tau*u0*uy))))))))/(16.*tau*teq*u0); // 7,9
    
    
    B[7] = 0; // 8,9
    
    
}



/*!
 *
 *  \brief Loads RHS of the equation
 *
 *  \li aTau = time
 *  \li aVArr = arrays of 3D vectors
 *  \li aArg = number of array which contains values of the fields being arguments for RHS equation
 *  \li aTo = number of array where RHS equation will be loaded in
 *
 */

void AnisotropicFluidEquationsNew::LoadRHS(const double aTau, Vector3D *** const aVArr, const usi aArg, const usi aTo){
    
    PRINT_DEBUG_3("<AnisotropicFluidEquationsNew::LoadRHS>");
    
    mVArr = aVArr; // set mVArr to access it in GetTeff()
    
    register Vector3D * const tVec0A = aVArr[0][aArg];
    register Vector3D * const tVec1A = aVArr[1][aArg];
    register Vector3D * const tVec2A = aVArr[2][aArg];
    register Vector3D * const tVec3A = aVArr[3][aArg];
    register Vector3D * const tVec4A = aVArr[4][aArg];
    register Vector3D * const tVec5A = aVArr[5][aArg];
    register Vector3D * const tVec6A = aVArr[6][aArg];
    register Vector3D * const tVec7A = aVArr[7][aArg];
    
    register Vector3D * const tVec0 = aVArr[0][aTo];
    register Vector3D * const tVec1 = aVArr[1][aTo];
    register Vector3D * const tVec2 = aVArr[2][aTo];
    register Vector3D * const tVec3 = aVArr[3][aTo];
    register Vector3D * const tVec4 = aVArr[4][aTo];
    register Vector3D * const tVec5 = aVArr[5][aTo];
    register Vector3D * const tVec6 = aVArr[6][aTo];
    register Vector3D * const tVec7 = aVArr[7][aTo];
    
    for (usi i = 0; i < mXP; i++)
        for (usi j = 0; j < mYP; j++)
            for (usi k = 0; k < mZP; k++){
                
                SetVariables(aTau,i,j,k,tVec0A,tVec1A,tVec2A,tVec3A,tVec4A,tVec5A,tVec6A,tVec7A);
                
                double A[64],B[8];
                
                LoadAB(A,B);
                
                // gsl stuff goes here
                
                gsl_matrix_view m = gsl_matrix_view_array (A, 8, 8);
                
                gsl_vector_view b = gsl_vector_view_array (B, 8);
                
                gsl_vector *x = gsl_vector_alloc (8);
                
                int s;
                
                gsl_permutation * p = gsl_permutation_alloc (8);
                
                gsl_linalg_LU_decomp (&m.matrix, p, &s);
                
                gsl_linalg_LU_solve (&m.matrix, p, &b.vector, x);
                
                
                (*tVec0)(i,j,k) = gsl_vector_get (x, 0);
                (*tVec1)(i,j,k) = gsl_vector_get (x, 1);
                (*tVec2)(i,j,k) = gsl_vector_get (x, 2);
                (*tVec3)(i,j,k) = gsl_vector_get (x, 3);
                (*tVec4)(i,j,k) = gsl_vector_get (x, 4);
                (*tVec5)(i,j,k) = gsl_vector_get (x, 5);
                (*tVec6)(i,j,k) = gsl_vector_get (x, 6);
                (*tVec7)(i,j,k) = gsl_vector_get (x, 7);
                
                
                gsl_permutation_free (p);
                gsl_vector_free (x);
                
                CheckValues(tVec0,tVec1,tVec2,tVec3,tVec4,tVec5,tVec6,tVec7,i,j,k);
                
                /*
                 if (i==mXP/2 && j==mYP/2 && k==mZP/2) {
                 PrintRawMessage("T",(*tVec7A)(i,j,k)*hbarc);
                 PrintRawMessage("T_rf",GetTeffpt(i,j,k)*hbarc);
                 PrintRawMessage("T_rf2",temperature(Geteeff(i,j,k))*hbarc);
                 PrintRawMessage("E",energy(T));
                 PrintRawMessage("E_rf",Geteeff(i,j,k));
                 cout << endl << flush;
                 
                 cout << "-------------------------------------------" << endl;
                 for (int l=0;l<=10;l++) {
                 double T = 0.01 + 0.1*l;
                 cout << T << "\t";
                 cout << energy(T) << "\t";
                 cout << hfuncs.H3(1,1,1,mass.mass(T)/T)*T*T*T*T + mass.Beq(T);
                 cout << endl;
                 }
                 cout << "-------------------------------------------" << endl;=
                 }
                 */
                
            }
}


inline void AnisotropicFluidEquationsNew::CheckValues( Vector3D * const tVec0,
                                                      Vector3D * const tVec1,
                                                      Vector3D * const tVec2,
                                                      Vector3D * const tVec3,
                                                      Vector3D * const tVec4,
                                                      Vector3D * const tVec5,
                                                      Vector3D * const tVec6,
                                                      Vector3D * const tVec7, int i, int j, int k){
    
    PRINT_DEBUG_3("<AnisotropicFluidEquationsNew::CheckValues>");
}


/*  begin compiler-level switch for derivative order */

#if DERIVORDER==4

inline double AnisotropicFluidEquationsNew::Dx(Vector3D * const aVec, const usi i, const usi j, const usi k) {
    
    PRINT_DEBUG_3("<AnisotropicFluidEquationsNew::Dx>");
    
    if (i == 0) return ((*aVec)(1,j,k) - (*aVec)(0,j,k)) * mdXi;                             // right difference
    else if (i == mXPm1) return ((*aVec)(mXPm1,j,k) - (*aVec)(mXP - 2,j,k)) * mdXi;          // left difference
    else if (i == 1) return ((*aVec)(i + 1,j,k) - (*aVec)(i - 1,j,k)) * m2dXi;               // second order centered difference
    else if (i == mXPm1-1) return ((*aVec)(i + 1,j,k) - (*aVec)(i - 1,j,k)) * m2dXi;         // second order centered difference
    else return (8*((*aVec)(i + 1,j,k) - (*aVec)(i - 1,j,k)) - ((*aVec)(i + 2,j,k) - (*aVec)(i - 2,j,k)) ) * m2dXi / 6; // fourth order centered difference
}

inline double AnisotropicFluidEquationsNew::Dy(Vector3D * const aVec, const usi i, const usi j, const usi k) {
    
    PRINT_DEBUG_3("<AnisotropicFluidEquationsNew::Dy>");
    
    if (j == 0) return ((*aVec)(i,1,k) - (*aVec)(i,0,k)) * mdYi;                         // right difference
    else if (j == mYPm1) return ((*aVec)(i,mYPm1,k) - (*aVec)(i,mYP -2,k)) * mdYi;     // left difference
    else if (j == 1) return ((*aVec)(i,j+1,k) - (*aVec)(i,j-1,k)) * m2dYi;               // second order centered difference
    else if (j == mYPm1-1) return ((*aVec)(i,j+1,k) - (*aVec)(i,j-1,k)) * m2dYi;         // second order centered difference
    else return (8*((*aVec)(i,j+1,k) - (*aVec)(i,j-1,k)) - ((*aVec)(i,j+2,k) - (*aVec)(i,j-2,k)) ) * m2dYi / 6; // fourth order centered difference
}

inline double AnisotropicFluidEquationsNew::Dz(Vector3D * const aVec, const usi i, const usi j, const usi k) {
    
    PRINT_DEBUG_3("<AnisotropicFluidEquationsNew::Dz>");
    
    if (k == 0) return ((*aVec)(i,j,1) - (*aVec)(i,j,0)) * mdZi;                         // right difference
    else if (k == mZPm1) return ((*aVec)(i,j,mZPm1) - (*aVec)(i,j,mZP - 2)) * mdZi;    // left difference
    else if (k == 1) return ((*aVec)(i,j,k+1) - (*aVec)(i,j,k-1)) * m2dZi;               // second order centered difference
    else if (k == mZPm1-1) return ((*aVec)(i,j,k+1) - (*aVec)(i,j,k-1)) * m2dZi;         // second order centered difference
    else return (8*((*aVec)(i,j,k+1) - (*aVec)(i,j,k-1)) - ((*aVec)(i,j,k+2) - (*aVec)(i,j,k-2)) ) * m2dZi / 6; // fourth order centered difference
}

#else

inline double AnisotropicFluidEquationsNew::Dx(Vector3D * const aVec, const usi i, const usi j, const usi k) {
    
    PRINT_DEBUG_3("<AnisotropicFluidEquationsNew::Dx>");
    
    if (i == 0) return ((*aVec)(1,j,k) - (*aVec)(0,j,k)) * mdXi;                             // right difference
    else if (i == mXPm1) return ((*aVec)(mXPm1,j,k) - (*aVec)(mXP - 2,j,k)) * mdXi;          // left difference
    else return ((*aVec)(i + 1,j,k) - (*aVec)(i - 1,j,k)) * m2dXi;                           // centered difference
}

inline double AnisotropicFluidEquationsNew::Dy(Vector3D * const aVec, const usi i, const usi j, const usi k) {
    
    PRINT_DEBUG_3("<AnisotropicFluidEquationsNew::Dy>");
    
    if (j == 0) return ((*aVec)(i,1,k) - (*aVec)(i,0,k)) * mdYi;                         // right difference
    else if (j == mYPm1) return ((*aVec)(i,mYPm1,k) - (*aVec)(i,mYP -2,k)) * mdYi;     // left difference
    else return ((*aVec)(i,j + 1,k) - (*aVec)(i,j - 1,k)) * m2dYi;                       // centered difference
}

inline double AnisotropicFluidEquationsNew::Dz(Vector3D * const aVec, const usi i, const usi j, const usi k) {
    
    PRINT_DEBUG_3("<AnisotropicFluidEquationsNew::Dz>");
    
    if (k == 0) return ((*aVec)(i,j,1) - (*aVec)(i,j,0)) * mdZi;                         // right difference
    else if (k == mZPm1) return ((*aVec)(i,j,mZPm1) - (*aVec)(i,j,mZP - 2)) * mdZi;    // left difference
    else return ((*aVec)(i,j,k + 1) - (*aVec)(i,j,k - 1)) * m2dZi;                       // centered difference
}

#endif

/*  end compiler-level switch for derivative order */


void AnisotropicFluidEquationsNew::SetVariables(const double aTau, const usi i, const usi j, const usi k, Vector3D * const aVec0, Vector3D * const aVec1,Vector3D * const aVec2,Vector3D * const aVec3,Vector3D * const aVec4,Vector3D * const aVec5,Vector3D * const aVec6,Vector3D * const aVec7){
    
    PRINT_DEBUG_3("<AnisotropicFluidEquationsNew::SetVariables>");
    
    // set functions values
    
    ax   = (*aVec0)(i,j,k);  // alpha_x
    ay   = (*aVec1)(i,j,k);  // alpha_y
    az   = (*aVec2)(i,j,k);  // alpha_z
    ux   = (*aVec3)(i,j,k);  // ux
    uy   = (*aVec4)(i,j,k);  // uy
    vt   = (*aVec5)(i,j,k);  // vartheta
    L    = (*aVec6)(i,j,k);  // lambda
    T    = (*aVec7)(i,j,k);  // temperature
    
    //*************************
    m = mass.mass(T);
    mh = m/L;
    mheq = m/T;
    dTm = mass.dTmass(T);
    //*************************
    
    
    //****************/
    
    //K1=K(1,mh);
    K2=K(2,mh);
    K3=K(3,mh);
    K4=K(4,mh);
    //K1eq=K(1,mheq);
    //K2eq=K(2,mheq);
    K3eq=K(3,mheq);
    //K4eq=K(4,mheq);
    
    // set eta, tau and Y
    
    tau = aTau;
    vs = mGrid->GetZ(k); // load varsigma
    
    Ch = cosh(vs-vt);
    Sh = sinh(vs-vt);
    
    // set derivatives
    
    d2ax   =  Dx(aVec0,i,j,k);
    d3ax   =  Dy(aVec0,i,j,k);
    d4ax   =  Dz(aVec0,i,j,k);
    
    d2ay   =  Dx(aVec1,i,j,k);
    d3ay   =  Dy(aVec1,i,j,k);
    d4ay   =  Dz(aVec1,i,j,k);
    
    d2az   =  Dx(aVec2,i,j,k);
    d3az   =  Dy(aVec2,i,j,k);
    d4az   =  Dz(aVec2,i,j,k);
    
    d2ux   =  Dx(aVec3,i,j,k);
    d3ux   =  Dy(aVec3,i,j,k);
    d4ux   =  Dz(aVec3,i,j,k);
    
    d2uy   =  Dx(aVec4,i,j,k);
    d3uy   =  Dy(aVec4,i,j,k);
    d4uy   =  Dz(aVec4,i,j,k);
    
    d2vt   =  Dx(aVec5,i,j,k);
    d3vt   =  Dy(aVec5,i,j,k);
    d4vt   =  Dz(aVec5,i,j,k);
    
    d2L   =  Dx(aVec6,i,j,k);
    d3L   =  Dy(aVec6,i,j,k);
    d4L   =  Dz(aVec6,i,j,k);
    
    d2T   =  Dx(aVec7,i,j,k);
    d3T   =  Dy(aVec7,i,j,k);
    d4T   =  Dz(aVec7,i,j,k);
    
    
    // set EOS
    
    H3     =   hfuncs.H3(ax,ay,az,mh);
    d1H3   =   hfuncs.d1H3(ax,ay,az,mh);
    d2H3   =   hfuncs.d2H3(ax,ay,az,mh);
    d3H3   =   hfuncs.d3H3(ax,ay,az,mh);
    d4H3   =   hfuncs.d4H3(ax,ay,az,mh);
    
    H3X     =   hfuncs.H3X(ax,ay,az,mh);
    d1H3X   =   hfuncs.d1H3X(ax,ay,az,mh);
    d2H3X   =   hfuncs.d2H3X(ax,ay,az,mh);
    d3H3X   =   hfuncs.d3H3X(ax,ay,az,mh);
    d4H3X   =   hfuncs.d4H3X(ax,ay,az,mh);
    
    H3Y     =   hfuncs.H3Y(ax,ay,az,mh);
    d1H3Y   =   hfuncs.d1H3Y(ax,ay,az,mh);
    d2H3Y   =   hfuncs.d2H3Y(ax,ay,az,mh);
    d3H3Y   =   hfuncs.d3H3Y(ax,ay,az,mh);
    d4H3Y   =   hfuncs.d4H3Y(ax,ay,az,mh);
    
    H3L     =   hfuncs.H3L(ax,ay,az,mh);
    d1H3L   =   hfuncs.d1H3L(ax,ay,az,mh);
    d2H3L   =   hfuncs.d2H3L(ax,ay,az,mh);
    d3H3L   =   hfuncs.d3H3L(ax,ay,az,mh);
    d4H3L   =   hfuncs.d4H3L(ax,ay,az,mh);
    
    H3eq    =   hfuncs.H3(1.,1.,1.,mheq);
    d1H3eq  =   hfuncs.d4H3(1.,1.,1.,mheq);
    
    H3B     =   hfuncs.H3B(ax,ay,az,mh);
    dTB     =   -m*dTm*L*L*H3B;
    
    gFac = 32*M_PI*Nt;
    
    // now the boring part comes
    
    ax2 = ax*ax;
    ax3 = ax2*ax;
    ax4 = ax3*ax;
    
    ay2 = ay*ay;
    ay3 = ay2*ay;
    ay4 = ay3*ay;
    
    az2 = az*az;
    az3 = az2*az;
    az4 = az3*az;
    
    ux2 = ux*ux;
    ux3 = ux2*ux;
    ux4 = ux3*ux;
    
    uy2 = uy*uy;
    uy3 = uy2*uy;
    uy4 = uy3*uy;
    
    L2 = L*L;
    L3 = L2*L;
    L4 = L3*L;
    L5 = L4*L;
    
    T2 = T*T;
    T3 = T2*T;
    T4 = T3*T;
    T5 = T4*T;
    
    m2 = m*m;
    m3 = m2*m;
    m4 = m3*m;
    
    up = sqrt(ux2+uy2);
    up2 = ux2+uy2;
    up3 = up2*up;
    up4 = up3*up;
    up5 = up4*up;
    
    u0 = sqrt(1+up2);
    u02 = 1 + up2;
    u03 = u02*u0;
    u04 = u03*u0;
    
    // tau_eq
    
    Eeq = energy(T);
    Peq = pressure(T);
    teq=tq.teq(T);
    
}

inline double AnisotropicFluidEquationsNew::energy(const double aT){
    
    PRINT_DEBUG_3("<AnisotropicFluidEquationsNew::energy>");
    
    return hfuncs.H3(1,1,1,mass.mass(aT)/aT)*aT*aT*aT*aT + mass.Beq(aT);
    
}

inline double AnisotropicFluidEquationsNew::pressure(const double aT){
    
    PRINT_DEBUG_3("<AnisotropicFluidEquationsNew::pressure>");
    
    return hfuncs.H3X(1,1,1,mass.mass(aT)/aT)*aT*aT*aT*aT - mass.Beq(aT);
    
}


inline void AnisotropicFluidEquationsNew::PrintInfo(void){
    
    PRINT_DEBUG_3("<AnisotropicFluidEquationsNew::PrintInfo>");
    
    PrintRawMessage("-> Anisotropic Fluid (LO)");
    PrintRawMessage("\tEta bar", d2s(eb));
    
}

/*!
 *
 *  \brief Formats output
 *
 *  \li aX = output value to be formated
 *  \li aN = number of the equation whose output is being formated
 *
 */

double AnisotropicFluidEquationsNew::FormatOutput(double aX, int aN){
    
    PRINT_DEBUG_3("<AnisotropicFluidEquationsNew::FormatOutput>");
    
    switch(aN){
            
        case 0: return UnitsFactor * aX;
            
        default: return aX;
    }
}

/*!
 *
 *  \brief Formats input
 *
 *  \li aX = output value to be formated
 *  \li aN = number of the equation whose input is being formated
 *
 */

double AnisotropicFluidEquationsNew::FormatInput(double aX, int aN){
    
    PRINT_DEBUG_3("<AnisotropicFluidEquationsNew::FormatOutput>");
    
    switch(aN){
            
        default: return aX;
    }
}

/*!
 *
 *  \brief Enforces Landau matching by setting the temperature
 *
 *  \li aVArr = arrays associated with the grid
 *  \li mXP, mYP, mZP = grid dimensions
 *
 */
void AnisotropicFluidEquationsNew::EnforceLandauMatching(Vector3D *** const aVArr) {
    Vector3D* tVec = aVArr[7][0]; // pointer to temperature
    usi mXP = mGrid->GetXP();
    usi mYP = mGrid->GetYP();
    usi mZP = mGrid->GetZP();
    for (usi i = 0; i < mXP; i++)
        for (usi j = 0; j < mYP; j++)
            for (usi k = 0; k < mZP; k++)
                (*tVec)(i,j,k)  = GetTeffpt(i,j,k);
}

/*!
 *
 *  \brief Returns energy density
 *
 *  \li i,j,k = grid point indexes
 *
 */
double AnisotropicFluidEquationsNew::Geteeff(int i, int j, int k){
    
    PRINT_DEBUG_3("<AnisotropicFluidEquationsNew::GetTeff>");
    
    return   hfuncs.H3(
                       (*mVArr[0][0])(i,j,k),
                       (*mVArr[1][0])(i,j,k),
                       (*mVArr[2][0])(i,j,k),
                       mass.mass((*mVArr[7][0])(i,j,k))/(*mVArr[6][0])(i,j,k)
                       ) * (*mVArr[6][0])(i,j,k)*(*mVArr[6][0])(i,j,k)*(*mVArr[6][0])(i,j,k)*(*mVArr[6][0])(i,j,k) ;
    
}

/*!
 *
 *  \brief Returns effective temperature
 *
 *  \li i,j,k = grid point indexes
 *
 */
double AnisotropicFluidEquationsNew::GetTeffpt(int i, int j, int k){
    
    PRINT_DEBUG_3("<AnisotropicFluidEquationsNew::GetTeff>");
    
    return   findT((*mVArr[6][0])(i,j,k), // lambda
                   (*mVArr[0][0])(i,j,k), // ax
                   (*mVArr[1][0])(i,j,k), // ay
                   (*mVArr[2][0])(i,j,k), // az
                   (*mVArr[7][0])(i,j,k), // T - used as guess for root finder
                   1.1); // multiplier used for solution bracketing > 1
}

/*!
 *
 *  \brief Returns effective temperature
 *
 *  \li x,y,e = position in simulation volume
 *
 */
double AnisotropicFluidEquationsNew::GetTeff(double x, double y, double e){
    
    PRINT_DEBUG_3("<AnisotropicFluidEquationsNew::GetTeff>");
    
    return   findT(mVArr[6][0]->Interpolate(x,y,e),
                   mVArr[0][0]->Interpolate(x,y,e),
                   mVArr[1][0]->Interpolate(x,y,e),
                   mVArr[2][0]->Interpolate(x,y,e),
                   mVArr[7][0]->Interpolate(x,y,e),
                   1.1);
}

double AnisotropicFluidEquationsNew::lambdaFunc(double T, void *params)
{
    struct lambdaFunc_params *p = (struct lambdaFunc_params *) params;
    
    double lambda = p->lambda;
    double ax = p->ax;
    double ay = p->ay;
    double az = p->az;
    
    Hfuncs *hfuncs = p->hfuncs;
    ThermalMass *mass = p->mass;
    
    double m = mass->mass(T);
    
    return hfuncs->H3(1,1,1,m/T)*T*T*T*T - hfuncs->H3(ax,ay,az,m/lambda)*lambda*lambda*lambda*lambda;
}



double AnisotropicFluidEquationsNew::findT(double lambda, double ax, double ay, double az, double tguess, double delta) {
    
    /* cout << "findT: ";
     cout << setw(10) << T;
     cout << setw(10) << ax;
     cout << setw(10) << ay;
     cout << setw(10) << az;
     cout << setw(10) << m << endl; */
    
    int status;
    int iter = 0, max_iter = 100;
    
    const gsl_root_fsolver_type *lT;
    gsl_root_fsolver *s;
    
    double r = fabs(tguess);
    double x_lo = tguess/delta, x_hi = delta*tguess;
    
    gsl_function F;
    struct lambdaFunc_params params = {lambda, ax, ay, az, &hfuncs, &mass};
    
    F.function = &AnisotropicFluidEquationsNew::lambdaFunc;
    F.params = &params;
    
    lT = gsl_root_fsolver_brent;
    s = gsl_root_fsolver_alloc(lT);
    gsl_root_fsolver_set(s, &F, x_lo, x_hi);
   
    gsl_set_error_handler_off();
 
    do
    {
        iter++;
        status = gsl_root_fsolver_iterate(s);
        r = gsl_root_fsolver_root(s);
        x_lo = gsl_root_fsolver_x_lower(s);
        x_hi = gsl_root_fsolver_x_upper(s);
        status = gsl_root_test_interval(x_lo, x_hi,0, 0.001);
        if (status == GSL_SUCCESS) break;
    }
    while (status == GSL_CONTINUE && iter < max_iter);
    
    gsl_root_fsolver_free(s);
    
    if (status==GSL_SUCCESS) return r;
    else {
        
        PrintRawMessage("Warning: Landau Matching failed for some reason");
        PrintRawMessage("T_guess",tguess);
        return fabs(tguess);
    }
}

double AnisotropicFluidEquationsNew::findT(double lambda, double ax, double ay, double az) {
    return findT(lambda,ax,ay,az,lambda,2.);
}



