/** 
*
*   EulerStepper.cxx
*	
*   Copyright (C) 2017 Radoslaw Ryblewski and Michael Strickland
*
*   This file is part of aHydro3p1.
*
*   aHydro3p1 is free software: you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation, either version 3 of the License, or
*   (at your option) any later version.
*
*   aHydro3p1 is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with aHydro3p1.  If not, see <http://www.gnu.org/licenses/>.
*    
*/

#include "Global.h"

/*!
 *
 *  \brief Default constructor
 *
 *  \li *aConfigurator = pointer to the configuration
 *  \li *aGrid = pointer to the grid
 *
 */

EulerStepper::EulerStepper(Configurator *aConfigurator, Grid *aGrid)
: mVArr(0), mEquations(0), mGrid(aGrid),
mXP(aGrid->GetXP()), mYP(aGrid->GetYP()), mZP(aGrid->GetZP()), mdT(aGrid->GetdT()){

    PRINT_DEBUG_3("EulerStepper::EulerStepper");

}

/*!
 *
 *  \brief Destructor
 *
 */

EulerStepper::~EulerStepper(){

    PRINT_DEBUG_3("EulerStepper::~EulerStepper");

}

/*!
 *
 *  \brief Allocates memory
 *
 */

void    EulerStepper::AllocateMemory(Equations *aEquations){

    PRINT_DEBUG_3("EulerStepper::AllocateMemory");

    // initilize with zeros

    try{
    mEquations = aEquations;
    mVArr = new Vector3D** [mEquations->GetNumberOfEquations()];
        for (int i = 0; i < aEquations->GetNumberOfEquations(); i++){
            mVArr[i] = new Vector3D*[mNumberOfVectors];
            for (int j = 0; j < mNumberOfVectors; j++){
            mVArr[i][j] = new Vector3D(
                mEquations->GetName(i),
                mGrid->GetXMin(),mGrid->GetXMax(),mGrid->GetXP(),
                mGrid->GetYMin(),mGrid->GetYMax(),mGrid->GetYP(),
                mGrid->GetZMin(),mGrid->GetZMax(),mGrid->GetZP()
                );
            }
        }
    } catch (std::bad_alloc) {

        exit(ProgramStatus(ERROR_MEMORY_BAD_ALLOCATION, "EulerStepper::AllocateMemory"));

    }
}

/*!
 *
 *  \brief Initializes vectors
 *
 */

void    EulerStepper::Initialize(Initializer **aIArr){

    PRINT_DEBUG_3("EulerStepper::Initialize");

    // initilize with Initializers

        for (int n = 0; n < mEquations->GetNumberOfEquations(); n++){
                PrintRawMessage("Initializing", mEquations->GetName(n));
                Vector3D* tVec = mVArr[n][0];
                for (usi i = 0; i < mXP; i++)
                    for (usi j = 0; j < mYP; j++)
                        for (usi k = 0; k < mZP; k++)
                            (*tVec)(i,j,k) += mEquations->FormatInput(aIArr[n]->GetValue(i,j,k), n);
        }
}

/*!
 *
 *  \brief Frees memory that is used by the class
 *
 */

void EulerStepper::FreeMemory(){

    PRINT_DEBUG_3("EulerStepper::FreeMemory");

    if(mVArr) {
        for (int i = 0; i < mEquations->GetNumberOfEquations(); i++) {
            for (int j = 0; j < mNumberOfVectors; j++){
                delete mVArr[i][j];
            }
            delete[] mVArr[i];
        }
        delete[] mVArr;

    } else{
    PrintMessage("Nothing to Do");
    }
    mVArr = 0;

}

/*!
 *
 *  \brief Return arrays used by the stepper
 *
 */

inline Vector3D *** EulerStepper::GetArrays(void){

    PRINT_DEBUG_3("EulerStepper::GetArrays");

    return mVArr;

}

/*!
 *
 *  \brief Make an Euler step
 *
 *  \li aTau = proper time at which to evaluate the RHS of the evolution equations (positive definite real number)
 *
 */

void    EulerStepper::MakeTimeStep(const double aTau){

    PRINT_DEBUG_3("EulerStepper::MakeTimeStep");

    // copy data from arrays No.0 to arrays No.1

    CopyArrays(0,1);

    // make one time step dT by doing 1 step of Euler algorithm

    MakeStep(aTau,0);

    // swap pointers of arrays No.0 with arrays No.1

    SwapArrays(0,1);

}

/*!
 *
 *  \brief Copy arrays
 *
 *  \li aFrom = index of array to copy from
 *  \li aTo = index of array to copy to
 *
 */

inline void EulerStepper::CopyArrays(const usi aFrom, const usi aTo){

    PRINT_DEBUG_3("EulerStepper::CopyArrays");

    const usi mNEq = mEquations->GetNumberOfEquations();
    Vector3D *tVecT(0), *tVecF(0);

    for(usi n = 0; n < mNEq; n++){
    tVecF = mVArr[n][aFrom];
    tVecT = mVArr[n][aTo];
    for (usi i = 0; i < mXP; i++)
        for (usi j = 0; j < mYP; j++)
            for (usi k = 0; k < mZP; k++){

                (*tVecT)(i,j,k) = (*tVecF)(i,j,k);

            }
    }
}

/*!
 *
 *  \brief Swap arrays using a pointer swap
 *
 *  \li aArrA = index of array A
 *  \li aArrB = index of array B
 *
 */

inline void EulerStepper::SwapArrays(const usi aArrA, const usi aArrB){

    PRINT_DEBUG_3("EulerStepper::SwapArrays");

    Vector3D *tVecTemp = 0;
    usi n;
    const usi mNEq = mEquations->GetNumberOfEquations();

    for(n = 0; n < mNEq; n++){

        tVecTemp = mVArr[n][aArrA];
        mVArr[n][aArrA] = mVArr[n][aArrB];
        mVArr[n][aArrB] = tVecTemp;

    }
}

/*!
 *
 *  \brief Make an intermediate Euler step
 *
 *  \li aTau = proper time at which to evaluate the RHS of the evolution equations
 *  \li aArg = index of the mVArr to use to load the RHS of the evolution equations
 *
*/

void EulerStepper::MakeStep(const double aTau, const usi aArg) {

    PRINT_DEBUG_3("<EulerStepper::MakeStep>");

	mEquations->LoadRHS(aTau, mVArr, aArg);

    const usi mNEq = mEquations->GetNumberOfEquations();
    Vector3D *tVec0(0), *tVec1(0), *tVec2(0), *tVec3(0);

    for(usi n = 0; n < mNEq; n++){

        tVec1 = mVArr[n][1];
        tVec2 = mVArr[n][2];

        for (usi i = 0; i < mXP; i++)
            for (usi j = 0; j < mYP; j++)
                for (usi k = 0; k < mZP; k++)
                    (*tVec1)(i,j,k) += mdT * (*tVec2)(i,j,k);
    }
}

/*!
 *
 *  \brief Returns the number of vector arrays that will be used for the evolution
 *
 */

usi EulerStepper::GetNumberOfVectors(void){

    PRINT_DEBUG_3("<EulerStepper::GetNumberOfVectors>");

    return mNumberOfVectors;

}

/*!
 *
 *  \brief Prints Info
 *
 */

void EulerStepper::PrintInfo(void){

    PRINT_DEBUG_3("<EulerStepper::PrintInfo>");

    PrintRawMessage("-> 1st order Euler");

}
