/** 
*
*   GaussianBlobsInitialSource.cxx
*	
*   Copyright (C) 2017 Radoslaw Ryblewski and Michael Strickland
*
*   This file is part of aHydro3p1.
*
*   aHydro3p1 is free software: you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation, either version 3 of the License, or
*   (at your option) any later version.
*
*   aHydro3p1 is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with aHydro3p1.  If not, see <http://www.gnu.org/licenses/>.
*    
*/

#include "Global.h"

GaussianBlobsInitialSource::GaussianBlobsInitialSource(Configurator *aConfigurator,  Grid *aGrid)
: mConfigurator(aConfigurator), mGrid(aGrid){

    PRINT_DEBUG_3("<GaussianBlobsInitialSource::GaussianBlobsInitialSource>");

    // InitialTemperatureGauss in GeV
    // all variables are transformed to 1/fm

    mNormalizationFactor = mConfigurator->GetDouble("InitialTemperature") / hbarc;

    mSigmaX = mConfigurator->GetDouble("SigmaX");
    mSigmaY = mConfigurator->GetDouble("SigmaY");
    mSigmaZ = mConfigurator->GetDouble("SigmaZ");

    // choose random points distributed isotropically in a fixed radius sphere

    mnbumps = 100;
    mrmax = 6.;

    try{

        mxp = new double[mnbumps];
        myp = new double[mnbumps];
        mzp = new double[mnbumps];

    } catch (std::bad_alloc) {

        exit(ProgramStatus(ERROR_MEMORY_BAD_ALLOCATION, "<GaussianBlobsInitialSource::GaussianBlobsInitialSource>"));

    }

    double r,theta,phi;

    // seed the random number generator with the current time
    srand((unsigned int)time(0));

    for (usi i=0; i<mnbumps; i++) {
        r = mrmax*pow(((double) rand())/((double)RAND_MAX),1./3.);
        theta = asin(2*((double) rand())/((double)RAND_MAX)-1)+0.5*M_PI;
        phi = 2*M_PI*((double) rand())/((double)RAND_MAX);
        mxp[i] = r*sin(theta)*cos(phi);
        myp[i] = r*sin(theta)*sin(phi);
        mzp[i] = r*cos(theta);
    }
//
//    // equations
//
//    mEquationsChoice    = aConfigurator->GetInt("Equations");
//
//    switch(mEquationsChoice){
//
//        case 1:
//            // for PerfectFluidEquations
//            aFunction = &GaussianBlobsInitialSource::GetLogTemperature;
//            break;
//
//        case 2:
//            // for AnisotropicFluidEquations
//            aFunction = &GaussianBlobsInitialSource::GetTemperature;
//            break;
//
//        default:
//
//            exit(ProgramStatus(ERROR_OPTION_MISSING,
//                               "FlatInitialSource::GaussianInitialSource: Cannot find Equations for " + d2s(mEquationsChoice)));
//    }

}

GaussianBlobsInitialSource::~GaussianBlobsInitialSource(){

    PRINT_DEBUG_3("<GaussianBlobsInitialSource::~GaussianBlobsInitialSource>");

    delete[] mxp;
    delete[] myp;
    delete[] mzp;

}

double GaussianBlobsInitialSource::GetValue(int i, int j, int k){

    PRINT_DEBUG_3("<GaussianBlobsInitialSource::GetValue>");

    return mNormalizationFactor * GaussBlobs3D(mGrid->GetX(i),mGrid->GetY(j),mGrid->GetZ(k));
//
//    return (this->*aFunction)(mGrid->GetX(i),mGrid->GetY(j),mGrid->GetZ(k));

}
//
//double GaussianBlobsInitialSource::GetTemperature(double x, double y, double z){
//
//    PRINT_DEBUG_3("<GaussianBlobsInitialSource::GetTemperature>");
//
//    return mNormalizationFactor * GaussBlobs3D(x, y, z);
//
//}
//
//double GaussianBlobsInitialSource::GetLogTemperature(double x, double y, double z){
//
//    PRINT_DEBUG_3("<GaussianBlobsInitialSource::GetLogTemperature>");
//
//    return log(mNormalizationFactor * GaussBlobs3D(x, y, z));
//
//}

double GaussianBlobsInitialSource::GaussBlobs3D(double x, double y, double z){

    PRINT_DEBUG_3("<GaussianBlobsInitialSource::Gauss3D>");

    double lambda = verysmall;

    for (usi i=0; i<mnbumps; i++) {
        lambda += (1 - verysmall) * exp(-0.5 * (pow((x-mxp[i]) / mSigmaX,2) + pow((y-myp[i]) / mSigmaY,2) + pow((z-mzp[i]) / mSigmaZ,2))) + verysmall;
    }

    lambda /= mnbumps;
    lambda *= pow(mrmax,2);

    return lambda;
}

void GaussianBlobsInitialSource::PrintInfo(void){

    PRINT_DEBUG_3("<GaussianBlobsInitialSource::PrintInfo>");

    PrintRawMessage("-> Random 3D Gaussian Blobs");
    PrintRawMessage("\tInitial central Lambda", d2s(mNormalizationFactor) + " [1/fm] (" + d2s(mNormalizationFactor * hbarc) + " [GeV])");
    PrintRawMessage("\tsigmaX", d2s(mSigmaX));
    PrintRawMessage("\tsigmaY", d2s(mSigmaY));
    PrintRawMessage("\tsigmaE", d2s(mSigmaZ));

}
