/** 
*
*   GlauberInspiredInitialSource.cxx
*	
*   Copyright (C) 2017 Radoslaw Ryblewski and Michael Strickland
*
*   This file is part of aHydro3p1.
*
*   aHydro3p1 is free software: you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation, either version 3 of the License, or
*   (at your option) any later version.
*
*   aHydro3p1 is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with aHydro3p1.  If not, see <http://www.gnu.org/licenses/>.
*    
*/

#include "Global.h"

/*!
*
*  \brief Default constructor
*
*  \li aConfigurator = pointer to configuration
*  \li aEquationOfState = pointer to the equation of state
*  \li aGrid = pointer to grid
*
*/

GlauberInspiredInitialSource::GlauberInspiredInitialSource(Configurator *aConfigurator, EquationOfState *aEquationOfState,  Grid *aGrid)
: mConfigurator(aConfigurator), mEOS(aEquationOfState), mGrid(aGrid){

    PRINT_DEBUG_3("<GlauberInspiredInitialSource::GlauberInspiredInitialSource>");

    // create optical glauber for impact parameter set by user

    mGlau    = new OpticalGlauber(mConfigurator);

    // this glauber will be used for calculating Scale

    mGlauC   = 0;

    // initial temperature in GeV transformed to fm

    mNormalizationFactor            = mConfigurator->GetDouble("InitialTemperature") / hbarc;
    mMix                            = mConfigurator->GetDouble("MixingFactor");
    mCollEn                         = mConfigurator->GetDouble("CollisionEnergy");
    mGaussianWidth                  = mConfigurator->GetDouble("GaussianHalfWidth");
    mPlateauWidth                   = mConfigurator->GetDouble("PlateauHalfWidth");

    mInitialSourceChoice            = mConfigurator->GetInt("InitialSource");
    mSourcesNormalizationChoice     = mConfigurator->GetInt("DensityOfSourcesNormalization");
    mSourcesScalingChoice           = mConfigurator->GetInt("DensityOfSourcesScaling");
//    mEquationsChoice                = mConfigurator->GetInt("Equations");

    switch(mInitialSourceChoice){

        case 2: // shifted

            SetParticipantVelocity(mCollEn);
            aSourceFunction = &GlauberInspiredInitialSource::ShiftedSource;
            aSourceFunction2 = &GlauberInspiredInitialSource::ShiftedSource2;
            mInitialSourceDesc = "Shifted Source";
            break;

        case 3: // tilted

            mShift = mConfigurator->GetDouble("ShiftInRapidity");
            SetParticipantRapidity(mCollEn);
            mRapCorr = mParticipantR - mShift;
            aSourceFunction = &GlauberInspiredInitialSource::TiltedSource;
            aSourceFunction2 = &GlauberInspiredInitialSource::TiltedSource2;
            mInitialSourceDesc = "Tilted Source";
            break;

        default:

        exit(ProgramStatus(ERROR_OPTION_MISSING,
        "GlauberInspiredInitialSource::GlauberInspiredInitialSource: Cannot find Initial source for " + d2s(mInitialSourceChoice)));

    }

    // create arrays for wounded nucleons of nuclei A & B, binary collisions and rapidity shift

    try {

        WNAArr  = 0;
        WNBArr  = 0;
        BCArr   = 0;
        RSArr   = 0;

        WNAArr  = AllocateArray2D(1);
        WNBArr  = AllocateArray2D(2);
        BCArr   = AllocateArray2D(3);
        if(mInitialSourceChoice == 2) RSArr   = AllocateArray2D(4);

    } catch(std::bad_alloc){

        exit(ProgramStatus(ERROR_MEMORY_BAD_ALLOCATION,
        "Cannot create allocate memory for arrays for GlauberInspiredInitialSource."));

    }

    // choose normalization

    switch(mSourcesNormalizationChoice){

     case 1:

        // normalize to the density in center in central collisions (x=0,y=0,e=0,b=0),
        mSourcesNormalizationDesc = "b = 0";
        mGlauC = new OpticalGlauber(mConfigurator);
        mGlauC->SetImpactParameter(0.0);
        mScale = 1 / GetSource2(mGlauC, 0.0, 0.0, 0.0);
        break;

     case 2:

        // normalize to the density in center in the collision with certain b (x=0,y=0,e=0,b=ImpactParameter)
        mSourcesNormalizationDesc = "b = " + d2s(mGlau->GetImpactParameter());
        mScale = 1 / GetSource2(mGlau, 0.0, 0.0, 0.0);
        break;

     default:

        exit(ProgramStatus(ERROR_OPTION_MISSING,
        "GlauberInspiredInitialSource::GlauberInspiredInitialSource: Cannot find Normalization Of Sources for " + d2s(mSourcesNormalizationChoice)));
    }

    switch(mSourcesScalingChoice){

            case 1: mSourcesScalingDesc = "entropy";  break;
            case 2: mSourcesScalingDesc = "energy";  break;
            default:
                exit(ProgramStatus(ERROR_OPTION_MISSING,
                "GlauberInspiredInitialSource::GetValue: Cannot find Initial Density Scaling for " + d2s(mSourcesScalingChoice)));
    }
}

/*!
 *
 *  \brief Destructor
 *
 */

GlauberInspiredInitialSource::~GlauberInspiredInitialSource(){

    PRINT_DEBUG_3("<GlauberInspiredInitialSource::~GlauberInspiredInitialSource>");

    delete mGlau;
    if(mGlauC) delete mGlauC;

    FreeArray2D(WNAArr);
    FreeArray2D(WNBArr);
    FreeArray2D(BCArr);
    if(RSArr) FreeArray2D(RSArr);
}

/*!
 *
 *  \brief Calculates the initial condition for PDE
 *
 *  \li mGlau - pointer to Optical Glauber
 *  \li x - x coordinate of the point in space
 *  \li y - y coordinate of the point in space
 *  \li x - x coordinate of the point in space
 *
 */

double GlauberInspiredInitialSource::GetValue(int i, int j, int k){

    PRINT_DEBUG_3("<GlauberInspiredInitialSource::GetValue>");

        switch(mSourcesScalingChoice){

            case 1: return mEOS->TS(mEOS->ST(mNormalizationFactor) * mScale * GetSource(i,j,k));  break;

            case 2: return mEOS->TE(mEOS->ET(mNormalizationFactor) * mScale * GetSource(i,j,k));  break;

            default: exit(ProgramStatus(ERROR_OPTION_MISSING, "GlauberInspiredInitialSource::GetValue switch reached default"));
        }
}

/*!
 *
 *  \brief Calculates the source profile
 *
 *  \li mGlau - pointer to Optical Glauber
 *  \li x - x coordinate of the point in space
 *  \li y - y coordinate of the point in space
 *  \li x - x coordinate of the point in space
 *
 */

double GlauberInspiredInitialSource::GetSource2(OpticalGlauber* mGlau, double x, double y, double z){

    PRINT_DEBUG_3("<GlauberInspiredInitialSource::GetSource>");

    return (this->*aSourceFunction2)(mGlau,x,y,z);
}

/*!
 *
 *  \brief Calculates the source profile
 *
 *  \li i - i-th point in x direction on grid
 *  \li j - j-th point in y direction on grid
 *  \li k - k-th point in z direction on grid
 *
 */

double GlauberInspiredInitialSource::GetSource(int i, int j, int k){

    PRINT_DEBUG_3("<GlauberInspiredInitialSource::GetSource>");

    return (this->*aSourceFunction)(i,j,k);

}

/*!
 *
 *  \brief Calculates shifted source
 *
 *  \li i - i-th point in x direction on grid
 *  \li j - j-th point in y direction on grid
 *  \li k - k-th point in z direction on grid
 *
 */

double GlauberInspiredInitialSource::ShiftedSource(int i, int j, int k){

    PRINT_DEBUG_3("<GlauberInspiredInitialSource::ShiftedSource>");

    return
            (
            (1 -  mMix) * (WNAArr[i][j] + WNBArr[i][j])  +
              2 * mMix  *  BCArr[i][j]
            ) * PseudorapidtyProfile(mGrid->GetZ(k) - RSArr[i][j]);

}

/*!
 *
 *  \brief Calculates shifted source
 *
 *  \li mGlau - pointer to Optical Glauber
 *  \li x - x coordinate of the point in space
 *  \li y - y coordinate of the point in space
 *  \li x - x coordinate of the point in space
 *
 */

double GlauberInspiredInitialSource::ShiftedSource2(OpticalGlauber* mGlau, double x, double y, double z){

    PRINT_DEBUG_3("<GlauberInspiredInitialSource::GetSource>");

    return (
            (1 -  mMix) * (mGlau->GetWNA(x,y) + mGlau->GetWNB(x,y))  +
              2 * mMix  *  mGlau->GetBC(x,y)
            ) * PseudorapidtyProfile(z - GetRapidityShift(mGlau,x,y));
}

/*!
 *
 *  \brief Calculates tilted source
 *
 *  \li i - i-th point in x direction on grid
 *  \li j - j-th point in y direction on grid
 *  \li k - k-th point in z direction on grid
 *
 */

double GlauberInspiredInitialSource::TiltedSource(int i, int j, int k){

    PRINT_DEBUG_3("<GlauberInspiredInitialSource::TiltedSource>");

    return
            (1 -  mMix) * (WNAArr[i][j] * PseudorapidtyProfile(mGrid->GetZ(k)) * FunctionF( mGrid->GetZ(k))  +
                           WNBArr[i][j] * PseudorapidtyProfile(mGrid->GetZ(k)) * FunctionF(-mGrid->GetZ(k))) +
                  mMix  *   BCArr[i][j] * PseudorapidtyProfile(mGrid->GetZ(k));

}

/*!
 *
 *  \brief Calculates tilted source
 *
 *  \li mGlau - pointer to Optical Glauber
 *  \li x - x coordinate of the point in space
 *  \li y - y coordinate of the point in space
 *  \li x - x coordinate of the point in space
 *
 */

double GlauberInspiredInitialSource::TiltedSource2(OpticalGlauber* mGlau, double x, double y, double z){

    PRINT_DEBUG_3("<GlauberInspiredInitialSource::GetSource>");

    return
            (1 -  mMix) * (mGlau->GetWNA(x,y) * PseudorapidtyProfile(z) * FunctionF( z) +
                           mGlau->GetWNB(x,y) * PseudorapidtyProfile(z) * FunctionF(-z)) +
                  mMix  *  mGlau->GetBC(x,y)  * PseudorapidtyProfile(z);

}

/*!
 *
 *  \brief Calculates pseudorapidity profile
 *
 *  \li z - pseudorapidity
 *
 */

inline double GlauberInspiredInitialSource::PseudorapidtyProfile(double z){

    PRINT_DEBUG_3("<GlauberInspiredInitialSource::PseudorapidtyProfile>");
    double verysmall = (double)1e-10;
    return (1 - verysmall) * exp(-pow((fabs(z) - mPlateauWidth),2.0) / (2.0 * pow(mGaussianWidth,2.0)) * unitstep(fabs(z) - mPlateauWidth)) + verysmall;

}

/*!
 *
 *  \brief Calculates function F for tilted source
 *
 *  \li z - pseudorapidity
 *
 */

inline double GlauberInspiredInitialSource::FunctionF(double z){

 if(z < - mRapCorr)   return 0;
 else if((- mRapCorr <= z) && (z <= mRapCorr)) return (z + mRapCorr) / (2 * mRapCorr);
 else return 1;

}

/*!
 *
 *  \brief Allocates 2dim array and initializes it
 *
 *  \li aInitializer - choice for array initializer
 *
 */

double**   GlauberInspiredInitialSource::AllocateArray2D(int aInitializer){

    PRINT_DEBUG_3("<GlauberInspiredInitialSource::AllocateArray2D>");

    double **tArray2D = 0;

    tArray2D = new double* [mGrid->GetXP()];
    for (int i = 0; i < mGrid->GetXP(); i++) {
        tArray2D[i] = new double [mGrid->GetYP()];
        for (int j = 0; j < mGrid->GetYP(); j++) {
            switch(aInitializer){
                case 1:
                tArray2D[i][j] = mGlau->GetWNA(mGrid->GetX(i),mGrid->GetY(j));
                break;

                case 2:
                tArray2D[i][j] = mGlau->GetWNB(mGrid->GetX(i),mGrid->GetY(j));
                break;

                case 3:
                tArray2D[i][j] = mGlau->GetBC(mGrid->GetX(i),mGrid->GetY(j));
                break;

                case 4:
                tArray2D[i][j] = GetRapidityShift(mGlau, mGrid->GetX(i),mGrid->GetY(j));
                break;

                default:
                tArray2D[i][j] = 0.0;
            }
        }
    }
    return tArray2D;
}

/*!
 *
 *  \brief Free memory used by 2dim array
 *
 *  \li aArray2D - pointer to the 2dim array
 *
 */

void GlauberInspiredInitialSource::FreeArray2D(double **aArray2D){

    PRINT_DEBUG_3("<GlauberInspiredInitialSource::FreeArray2D>");

      if(aArray2D) {
        for (int i = 0; i < mGrid->GetXP(); i++) {
          delete[] aArray2D[i];
        }
        delete[] aArray2D;
      }
}

/*!
 *
 *  \brief Calculates rapidity shift for shifted initial source
 *
 *  \li z - pseudorapidity
 *
 */

inline double GlauberInspiredInitialSource::GetRapidityShift(OpticalGlauber *aGlau, double x, double y){

    PRINT_DEBUG_3("<GlauberInspiredInitialSource::GetRapidityShift>");

    return 0.5 * log(
            (
                aGlau->GetWNA(x,y) + aGlau->GetWNB(x,y) +
                mParticipantV * (aGlau->GetWNA(x,y) - aGlau->GetWNB(x,y))
            ) / (
                aGlau->GetWNA(x,y) + aGlau->GetWNB(x,y) -
                mParticipantV * (aGlau->GetWNA(x,y) - aGlau->GetWNB(x,y))
            )
        );
}

/*!
 *
 *  \brief Calculates and sets participant velocity for shifted source
 *
 *  \li aCollEn - collision energy
 *
 */

inline double GlauberInspiredInitialSource::SetParticipantVelocity(double aCollEn){

    PRINT_DEBUG_3("<GlauberInspiredInitialSource::SetParticipantVelocity>");

    mParticipantV = sqrt(pow(aCollEn / 2.0,2) - pow((ProtonMass + NeutronMass) / 2.0, 2)) / (aCollEn / 2.0);
    return mParticipantV;
}

/*!
 *
 *  \brief Calculates and sets participant rapidity for tilted source
 *
 *  \li aCollEn - collision energy
 *
 */

inline double GlauberInspiredInitialSource::SetParticipantRapidity(double aCollEn){

    PRINT_DEBUG_3("<GlauberInspiredInitialSource::GetParticipantRapidity>");

    mParticipantR = log(aCollEn / ((ProtonMass + NeutronMass) / 2.0));
    return mParticipantR;
}

/*!
 *
 *  \brief Print information
 *
 */

void GlauberInspiredInitialSource::PrintInfo(void){

    PRINT_DEBUG_3("<GlauberInspiredInitialSource::PrintInfo>");

    PrintRawMessage("-> " + mInitialSourceDesc);
    PrintRawMessage("\tNormalize to " + mSourcesNormalizationDesc);
    PrintRawMessage("\tDensity(x,y,z) ~ " + mSourcesScalingDesc);
    PrintRawMessage("\tInitial central Lambda", d2s(mNormalizationFactor) + " [1/fm] (" + d2s(mNormalizationFactor * hbarc) + " [GeV])");
    PrintRawMessage("\tMixing Factor", mMix);
    PrintRawMessage("\tCollision Energy [GeV]", mCollEn);
    PrintRawMessage("\tGaussian Half Width", mGaussianWidth);
    PrintRawMessage("\tPlateau Half Width", mPlateauWidth);
    mGlau->PrintInfo();
}
