/** 
*
*   IdealEquationOfState.cxx
*	
*   Copyright (C) 2017 Radoslaw Ryblewski and Michael Strickland
*
*   This file is part of aHydro3p1.
*
*   aHydro3p1 is free software: you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation, either version 3 of the License, or
*   (at your option) any later version.
*
*   aHydro3p1 is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with aHydro3p1.  If not, see <http://www.gnu.org/licenses/>.
*    
*/

#include "Global.h"

/*!
*
*  \brief Default constructor
*
*  \li aStatistics = statistics, choices: 1 - Boltzmann (default), 2 - Bose-Einstein, 3- Fermi-Dirac, default:1
*  \li aDegeneracy = degeneracy factor, default:0 (if default then calculate degeneracy with built-in values)
*  \li aQuarkFlavours = quark flavours, default:2
*
*/

IdealEquationOfState::IdealEquationOfState(int aStatistics, double aDegeneracy, int aQuarkFlavours)
:mStatistics(aStatistics), mQuarkFlavours(aQuarkFlavours), mGluonColors(8), mColors(3), mSpin(2), mAntiParticle_Particle(2), BuiltIn(true){

    PRINT_DEBUG_3("<IdealEquationOfState::IdealEquationOfState>");

    switch(mStatistics){

        case 1:

            //mDegeneracy = mGluonColors * mSpin; // Boltzmann
	    mDegeneracy = 51.41035360127906;
            mMultiplierB= 1.000;
            mStatisticsDesc = "Boltzmann";
            break;

        case 2:

            mDegeneracy = mGluonColors * mSpin; // Bose-Einstein
            mMultiplierBE= Pi4 / 90;
            mStatisticsDesc = "Bose-Einstein";
            break;

        case 3:

            mDegeneracy = mColors * mSpin * mQuarkFlavours * mAntiParticle_Particle; // Fermi-Dirac
            mMultiplierFD = (7 * Pi4) / 720;
            mStatisticsDesc = "Fermi-Dirac";
            break;

        default:

            exit(ProgramStatus(ERROR_OPTION_MISSING,
                "IdealEquationOfState::IdealEquationOfState: Cannot find particle statistics for " + d2s(mStatistics)));
     }

    mM = *(GetMultiplier());

    if(aDegeneracy) {
        mDegeneracy = aDegeneracy;
        BuiltIn = false;
    }
}

/*!
*
*  \brief Default destructor
*
*/

IdealEquationOfState::~IdealEquationOfState(){

    PRINT_DEBUG_3("<IdealEquationOfState::~IdealEquationOfState>");

}

/*!
*
*  \brief Calculates speed of sound for given temperature
*
*  \li T - temperature
*
*/

double IdealEquationOfState::CsT(double T){

    PRINT_DEBUG_3("<IdealEquationOfState::CsT>");

    return 1 / sqrt(3.0); // the same for all statistics

}

/*!
*
*  \brief Calculates entropy density for given temperature
*
*  \li T - temperature
*
*/

double IdealEquationOfState::ST(double T){

    PRINT_DEBUG_3("<IdealEquationOfState::ST>");

    return (4 * mDegeneracy) / Pi2 * pow(T,3) * mM;

}

/*!
*
*  \brief Calculates energy density for given temperature
*
*  \li T - temperature
*
*/

double IdealEquationOfState::ET(double T){

    PRINT_DEBUG_3("<IdealEquationOfState::ET>");

    return (3. * mDegeneracy) / Pi2 * pow(T,4) * mM;
//return pow(T,4);

}

/*!
*
*  \brief Calculates pressure for given temperature
*
*  \li T - temperature
*
*/

double IdealEquationOfState::PT(double T){

    PRINT_DEBUG_3("<IdealEquationOfState::PT>");

    return mDegeneracy / Pi2 * pow(T,4) * mM;

}

/*!
*
*  \brief Calculates temperature for given entropy density
*
*  \li S - entropy density
*
*/

double IdealEquationOfState::TS(double S){

    PRINT_DEBUG_3("<IdealEquationOfState::TS>");

    return pow((Pi2 * S) / (4 * mDegeneracy * mM), 1 / 3.0);

}

/*!
*
*  \brief Calculates temperature for given energy density
*
*  \li E - energy density
*
*/

double IdealEquationOfState::TE(double E){

    PRINT_DEBUG_3("<IdealEquationOfState::TE>");

    return pow((Pi2 * E) / (3 * mDegeneracy * mM), 1 / 4.0);

}

/*!
*
*  \brief Calculates temperature for given pressure
*
*  \li P - pressure
*
*/

double IdealEquationOfState::TP(double P){

    PRINT_DEBUG_3("<IdealEquationOfState::TP>");

    return pow((Pi2 * P) / (mDegeneracy * mM), 1 / 4.0);

}

/*!
*
*  \brief Calculates derivative of energy density with respect to temperature for given temperature
*
*  \li T - etemperature
*
*/

double IdealEquationOfState::EpT(double T){

    PRINT_DEBUG_3("<IdealEquationOfState::EpT>");

    return 4 * ET(T) / T;

}

/*!
*
*  \brief Calculates derivative of entropy density with respect to temperature for given temperature
*
*  \li T - etemperature
*
*/

double IdealEquationOfState::SpT(double T){

    PRINT_DEBUG_3("<IdealEquationOfState::SpT>");

    return 3 * ST(T) / T;

}

/*!
*
*  \brief Returns the multiplier for given statistics
*
*/

inline double* IdealEquationOfState::GetMultiplier(void){

    PRINT_DEBUG_3("<IdealEquationOfState::GetMultiplier>");

    if(mStatistics == 1) return (&mMultiplierB);
    else if(mStatistics == 2) return (&mMultiplierBE);
    else return (&mMultiplierFD);

}

/*!
*
*  \brief Prints Info
*
*/

inline void IdealEquationOfState::PrintInfo(void){

    PrintRawMessage("-> Ideal (e = 3P)");

    PrintRawMessage("Particle statistics", mStatisticsDesc);

    if(BuiltIn) {
        switch(mStatistics){

            case 1:
            case 2:

                PrintRawMessage("Gluon Colors", mGluonColors);
                PrintRawMessage("Spin", mSpin);
                break;

            case 3:

                PrintRawMessage("Colors", mGluonColors);
                PrintRawMessage("Spin", mSpin);
                PrintRawMessage("Quark Flavours", mQuarkFlavours);
                PrintRawMessage("P-A", mAntiParticle_Particle);
                break;
        }
    }else{
        PrintRawMessage("User Defined Degeneracy Factor");
    }
    PrintRawMessage("Degeneracy Factor", mDegeneracy);
}
