/**
*
*   MCGlauber.cxx
*	
*   Copyright (C) 2017 Radoslaw Ryblewski and Michael Strickland
*
*   This file is part of aHydro3p1.
*
*   aHydro3p1 is free software: you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation, either version 3 of the License, or
*   (at your option) any later version.
*
*   aHydro3p1 is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with aHydro3p1.  If not, see <http://www.gnu.org/licenses/>.
*    
*/

#include "Global.h"

/*!
*
*  \brief Default constructor
*
*  \li aConfigurator = pointer to configuration
*
*/

MCGlauber::MCGlauber(Configurator *aConfigurator, usi aSeed) {

    PRINT_DEBUG_3("<MCGlauber::MCGlauber>");

    // define nulei A and B

    mNucleusA = new Nucleus(
                            aConfigurator->GetInt("NumberOfNucleonsA"),
                            aConfigurator->GetDouble("SurfaceDiffusenessA"),
                            aConfigurator->GetDouble("NuclearSaturationDensityA"),
                            aConfigurator->GetDouble("RadiusA"));
    
    mNucleusB = new Nucleus(
                            aConfigurator->GetInt("NumberOfNucleonsB"),
                            aConfigurator->GetDouble("SurfaceDiffusenessB"),
                            aConfigurator->GetDouble("NuclearSaturationDensityB"),
                            aConfigurator->GetDouble("RadiusB"));
    
    mb      = aConfigurator->GetDouble("ImpactParameter");
    mcross  = aConfigurator->GetDouble("InelasticCrossSection");

    // seed the random number generator with the current time
    srand(aSeed);
    
    // allocate memory for the collision points array
    usi ApB = mNucleusA->GetNumberOfNucleons() + mNucleusB->GetNumberOfNucleons();
    mxp = new double[ApB];
    myp = new double[ApB];
    
    // reset
    mNpart=0;
    
    // load the collision point arrays
    while (mNpart==0) mNpart = MCGlauber::LoadCollisionPointsArrays();    
}


/*!
 *
 *  \brief Destructor
 *
 */

MCGlauber::~MCGlauber(){

    PRINT_DEBUG_3("<MCGlauber::~MCGlauber>");

    // delete nuclei
    delete mNucleusA;
    delete mNucleusB;
    
    // delete collision point arrays
    delete[] mxp;
    delete[] myp;

}

/*!
 *
 * \brief Generates A sampled nucleon positions from the nuclear density profile
 *
 * Points are loaded into x and y arrays (z is not used in current context)
 * Note: This routine assumes spherical nuclei.  It also also assumes that 
 * the random number generator has already been seeded
 *
 * \li nucleus = Nucleus object
 * \li *x = pointer to an array of doubles of size A
 * \li *y = pointer to an array of doubles of size A
 *
 */

void MCGlauber::SampleNuclearDensity(Nucleus *nucleus, double *x, double *y) {
	
    const double M = 5; // number should be > max of the function; used for accept/reject test
	const double rmax = 20; // maximum possible radius to even consider as relevant
	
    usi A = nucleus->GetNumberOfNucleons();
    
    double r,v,f,u;
    usi m = 0;
    
	while (m<A) {
        r = rmax*((double) rand())/((double)RAND_MAX);
        u = ((double) rand())/((double)RAND_MAX);
        r /= sqrt(3.); // adjust since routine below needs separate x, y, and z components
        if (u < r*r*(nucleus->GetNuclearDensity(r,r,r))/M) {
            v = 2*(((double) rand())/((double)RAND_MAX)-0.5);
            f = 2*M_PI*((double) rand())/((double)RAND_MAX);
            x[m] = r*sqrt(1-v*v)*cos(f);
            y[m] = r*sqrt(1-v*v)*sin(f);
            m++;
        }
	}
    
}

/*!
 *
 * \brief Loads the collision point arrays with the collision points and returns the number of collisions found
 *
 * Uses hard-sphere model for determining if collision occured or not
 *
 */

usi MCGlauber::LoadCollisionPointsArrays() {
    
	double *x1,*y1,*x2,*y2,dn,dist;
	int *l1,*l2;
    
    // allocate memory and sample Nucleus A
    usi A = mNucleusA->GetNumberOfNucleons();
	x1 = new double[A];
	y1 = new double[A];
	SampleNuclearDensity(mNucleusA,x1,y1);
    
    // allocate participants list for Nucleus A
    l1 = new int[A];

    // allocate memory and sample Nucleus B
    usi B = mNucleusB->GetNumberOfNucleons();
	x2 = new double[B];
	y2 = new double[B];
    SampleNuclearDensity(mNucleusB,x2,y2);
    
    // allocate participants list Nucleus B
	l2 = new int[B];
    
	for (usi i=0; i<A; i++) {
        x1[i] -= mb/2;
        l1[i] = 0;
    }
    for (usi i=0; i<B; i++) {
		x2[i] += mb/2;
        l2[i] = 0;
	}
    
	//dn = sqrt(0.1*mcross/M_PI); // seems to be a disagreement about units with Radek here...
    dn = sqrt(mcross/M_PI); // hack around it for now
    
	for (usi i=0; i<A; i++)
        for (usi j=0; j<B; j++) {
            dist = pow(x1[i]-x2[j],2);
            dist += pow(y1[i]-y2[j],2);
            dist = sqrt(dist);
            if (dist<dn) {
                l1[i] += 1;
                l2[j] += 1;
            }
        }
    
	usi n = 0;
    
	for (usi i=0; i<A; i++) {
        if (l1[i]>0) {
            mxp[n] = x1[i];
            myp[n] = y1[i];
            n++;
        }
    }
    for (usi i=0; i<B; i++) {
        if (l2[i]>0) {
            mxp[n] = x2[i];
            myp[n] = y2[i];
            n++;
        }
	}
    
	delete[] x1;
	delete[] y1;
	delete[] l1;
	delete[] x2;
	delete[] y2;
	delete[] l2;
    
	return n;
    
}

/*!
 *
 *  \brief Returns the x points array
 *
 */

double * MCGlauber::GetXPointsArray(void){
    return mxp;
}

/*!
 *
 *  \brief Returns the y points array
 *
 */

double * MCGlauber::GetYPointsArray(void){
    return myp;
}

/*!
 *
 *  \brief Returns the number of participants in the event
 *
 */

usi MCGlauber::GetNumberOfParticipants(void){
    return mNpart;
}

/*!
 *
 *  \brief Prints info about the class
 *
 */

void MCGlauber::PrintInfo(void){

    PRINT_DEBUG_3("<MCGlauber::PrintInfo>");

    PrintRawMessage("");
    PrintRawMessage("\t-> Monte Carlo Glauber");
    PrintRawMessage("\tImpact Parameter [fm]",mb);
    PrintRawMessage("\tInelastic Cross Section [mb]",mcross);
    PrintRawMessage("");
    PrintRawMessage("\tNucleus A");
    if (mNucleusA==NULL) PrintRawMessage("\tTHE MYSTERY BUG STRIKES AGAIN");
    else mNucleusA->PrintInfo();
    PrintRawMessage("");
    PrintRawMessage("\tNucleus B");
    mNucleusB->PrintInfo();
    PrintRawMessage("");
    PrintRawMessage("\tNumber of Participants Found",mNpart);

}
