/**
*
*   MCGlauberInitialSource.cxx
*	
*   Copyright (C) 2017 Radoslaw Ryblewski and Michael Strickland
*
*   This file is part of aHydro3p1.
*
*   aHydro3p1 is free software: you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation, either version 3 of the License, or
*   (at your option) any later version.
*
*   aHydro3p1 is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with aHydro3p1.  If not, see <http://www.gnu.org/licenses/>.
*    
*/

#include "Global.h"

/*!
 *
 *  \brief Default constructor
 *
 *  \li *aConfigurator = pointer to the configuration
 *  \li *aGrid = pointer to the grid
 *
 */

MCGlauberInitialSource::MCGlauberInitialSource(Configurator *aConfigurator,  Grid *aGrid, usi aSeed)
: mConfigurator(aConfigurator), mGrid(aGrid) {

    PRINT_DEBUG_3("<MCGlauberInitialSource::MCGlauberInitialSource>");

    // InitialTemperatureGauss in GeV
    // all variables are transformed to 1/fm

    mNormalizationFactor = mConfigurator->GetDouble("InitialTemperature") / hbarc;

    // set std dev for gaussians

    mMCSig0 = mConfigurator->GetDouble("MCSig0");

    // parameters that enter the rapidity dependence of the profile

    mGaussianWidth                  = mConfigurator->GetDouble("GaussianHalfWidth");
    mPlateauWidth                   = mConfigurator->GetDouble("PlateauHalfWidth");

    // load a MCGlauber object

    mMCGlauber = new MCGlauber(mConfigurator,aSeed);

}

/*!
 *
 *  \brief Destructor
 *
 */

MCGlauberInitialSource::~MCGlauberInitialSource(){

    PRINT_DEBUG_3("<MCGlauberInitialSource::~MCGlauberInitialSource>");

    delete mMCGlauber;

}

/*!
 *
 *  \brief Returns the source profile
 *
 *  \li i = i index of the point in space
 *  \li j = j index of the point in space
 *  \li k = k index of the point in space
 *
 */

double MCGlauberInitialSource::GetValue(int i, int j, int k){

    PRINT_DEBUG_3("<MCGlauberInitialSource::GetValue>");

    //PrintMessage("i",i);
    //PrintMessage("j",j);
    //PrintMessage("k",k);
    //if (i==32 && j==32 && k==32) PrintRawMessage("Boo",mNormalizationFactor * GetProfile(mGrid->GetX(i),mGrid->GetY(j),mGrid->GetZ(k)));
    
    return mNormalizationFactor * GetProfile(mGrid->GetX(i),mGrid->GetY(j),mGrid->GetZ(k));
}
//
///*!
// *
// *  \brief Returns the temperature
// *
// *  \li x = x coordinate of the point in space
// *  \li y = y coordinate of the point in space
// *  \li x = x coordinate of the point in space
// *
// */
//
//double MCGlauberInitialSource::GetTemperature(double x, double y, double z){
//
//    PRINT_DEBUG_3("<MCGlauberInitialSource::GetTemperature>");
//
//    return mNormalizationFactor * GetProfile(x, y, z);
//
//}
//
///*!
// *
// *  \brief Returns the natural logarithm of the temperature
// *
// *  \li x = x coordinate of the point in space
// *  \li y = y coordinate of the point in space
// *  \li x = x coordinate of the point in space
// *
// */
//
//double MCGlauberInitialSource::GetLogTemperature(double x, double y, double z){
//
//    PRINT_DEBUG_3("<MCGlauberInitialSource::GetLogTemperature>");
//
//    return log(mNormalizationFactor * GetProfile(x, y, z));
//
//}

/*!
 *
 *  \brief Calculates the spatial rapidity profile
 *
 *  \li z = spatial rapidity
 *
 */

inline double MCGlauberInitialSource::SpatialRapidityProfile(double z){

    PRINT_DEBUG_3("<MCGlauberInitialSource::SpatialRapidityProfile>");

    return exp(-pow((fabs(z) - mPlateauWidth),2.0) / (2.0 * pow(mGaussianWidth,2.0)) * unitstep(fabs(z) - mPlateauWidth));

}

/*!
 *
 *  \brief Calculates the MCGlauber source proofile
 *
 *  \li x = x coordinate of the point in space
 *  \li y = y coordinate of the point in space
 *  \li x = x coordinate of the point in space
 *
 *  As implemented below, we put a gaussian at the position of every nucleon that underwent a collision
 *
 */

double MCGlauberInitialSource::GetProfile(double x, double y, double z){

    PRINT_DEBUG_3("<MCGlauberInitialSource::Gauss3D>");

    double *xp = mMCGlauber->GetXPointsArray();
    double *yp = mMCGlauber->GetYPointsArray();
    usi nPart = mMCGlauber->GetNumberOfParticipants();
    
    //PrintRawMessage("found nPart",nPart);

    double temp = 0;

    for (usi i=0; i<nPart; i++) {
        //PrintRawMessage("xp",xp[i]);
        //PrintRawMessage("yp",yp[i]);
        temp += exp(-0.5 * (pow((x-xp[i]) / mMCSig0,2) + pow((y-yp[i]) / mMCSig0,2)));
    }

    // set by ideal number density for now with rather arbitary normalization
    // also multiply by spatial rapidity profile
    // TODO:  Sync this with realistic EOS + options + more thought about the normalization
    //        + better/more clever regulation at very large r
    temp = pow(temp*SpatialRapidityProfile(z)/nPart,1./3.) + 1.0e-3;

    return temp;

}

/*!
 *
 *  \brief Prints info about the class
 *
*/

void MCGlauberInitialSource::PrintInfo(void){

    PRINT_DEBUG_3("<MCGlauberInitialSource::PrintInfo>");

    mMCGlauber->PrintInfo();
    PrintRawMessage("");
    PrintRawMessage("\tNucleonic 2D Gaussian Std Dev",mMCSig0);
    PrintRawMessage("\tRapidity Plateau Half Width",mPlateauWidth);
    PrintRawMessage("\tRapidity Tails Std Dev",mGaussianWidth);
    PrintRawMessage("");

}
