/** 
*
*   Nucleus.cxx
*	
*   Copyright (C) 2017 Radoslaw Ryblewski and Michael Strickland
*
*   This file is part of aHydro3p1.
*
*   aHydro3p1 is free software: you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation, either version 3 of the License, or
*   (at your option) any later version.
*
*   aHydro3p1 is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with aHydro3p1.  If not, see <http://www.gnu.org/licenses/>.
*    
*/

#include "Global.h"

/*!
*
*  \brief Default constructor
*
*  \li aNumberOfNucleons = number of nucleons
*  \li aSurfaceDiffuseness = surface diffuseness
*  \li aNuclearSaturationDensity = nuclear saturation density
*  \li aRadius = radius of the nucleus (optional)
*
*/

Nucleus::Nucleus(unsigned short int aNumberOfNucleons, double aSurfaceDiffuseness, double aNuclearSaturationDensity, double aRadius)
: mA(aNumberOfNucleons), ma(aSurfaceDiffuseness), mrho0(aNuclearSaturationDensity), mR(aRadius) {

    PRINT_DEBUG_3("<Nucleus::Nucleus>");

    if(!mR) mR = SetRadius(mA);

}

/*!
 *
 *  \brief Destructor
 *
 */

Nucleus::~Nucleus(){

    PRINT_DEBUG_3("<Nucleus::~Nucleus>");

}

/*!
 *
 *  \brief Calculates and sets the radius of the nucleus
 *
 *  \li aA = number of nucleons
 *
 */

double Nucleus::SetRadius(int aA){

    PRINT_DEBUG_3("<Nucleus::Nucleus>");

    mR = 1.12 * pow(aA, 1.0 / 3.0) - 0.86 * pow(aA, - 1.0 / 3.0);
    return mR;

}

/*!
 *
 *  \brief Returns the number of nucleons contained in the nucleus
 *
 */
usi Nucleus::GetNumberOfNucleons() {
    return mA;
}

/*!
 *
 *  \brief Calculates nuclear thickness of the nucleus
 *
 *  \li x - x coordinate of the point in transverse plane
 *  \li y - y coordinate of the point in transverse plane
 *
 */

double Nucleus::GetNuclearThickness(double x, double y){

    PRINT_DEBUG_3("<Nucleus::GetNuclearThickness>");

    return (
        Integrate(&Nucleus::GetNuclearDensity,x,y,-50.0,+50.0,300) // do not use &() http://www.elektroda.pl/rtvforum/topic539315.html   http://stackoverflow.com/questions/7134197/error-with-address-of-parenthesized-member-function
    );
}

/*!
 *
 *  \brief Calculates nuclear density of the nucleus (Woods-Saxon Parameterization)
 *
 *  \li x - x coordinate of the point in space
 *  \li y - y coordinate of the point in space
 *  \li z - z coordinate of the point in space
 *
 */

double Nucleus::GetNuclearDensity(double x, double y, double z){

    PRINT_DEBUG_3("<Nucleus::GetNuclearDensity>");

    return mrho0 / (1 + exp((sqrt(pow(x,2) + pow(y,2) + pow(z,2)) - mR) / ma));

}

/*!
 *
 *  \brief Perform 1dim integral in the z direction
 *
 *  \li double (Nucleus::*aFunction)(double, double, double) - function of three variables
 *  \li double x - x coordinate of the point in space
 *  \li double y - y coordinate of the point in space
 *  \li double zMin - minimal value of the integration range
 *  \li double zMax - maximal value of the integration range
 *  \li double zPts - number of integration points
 *
 */

double Nucleus::Integrate(double (Nucleus::*aFunction)(double, double, double), double x, double y, double zMin, double zMax, double zPoints){ // argument is a pointer to the integrated function

    PRINT_DEBUG_3("<Nucleus::Integrate>");

    double Sum = 0;
    double dz = fabs((zMax - zMin)) / zPoints;
    for (int i = 0; i < zPoints; ++i) Sum += ((this->*aFunction)(x,y,zMin + dz * i) + (this->*aFunction)(x,y,zMin + dz * (i + 1))) / 2.0;

    return dz * Sum;
}

/*!
 *
 *  \brief Print information
 *
 */

void Nucleus::PrintInfo(void){

    PRINT_DEBUG_3("<Nucleus::PrintInfo>");

    PrintRawMessage("\tNucleons",mA);
    PrintRawMessage("\tSurface Diffuseness",ma);
    PrintRawMessage("\tNuclear Saturation Density",mrho0);
    PrintRawMessage("\tRadius [fm]",mR);
}
