/** 
*
*   OpticalGlauber.cxx
*	
*   Copyright (C) 2017 Radoslaw Ryblewski and Michael Strickland
*
*   This file is part of aHydro3p1.
*
*   aHydro3p1 is free software: you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation, either version 3 of the License, or
*   (at your option) any later version.
*
*   aHydro3p1 is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with aHydro3p1.  If not, see <http://www.gnu.org/licenses/>.
*    
*/

#include "Global.h"

/*!
*
*  \brief Default constructor
*
*  \li aConfigurator = pointer to configuration
*
*/

OpticalGlauber::OpticalGlauber(Configurator *aConfigurator){

    PRINT_DEBUG_3("<OpticalGlauber::OpticalGlauber>");

    // define nulei A and B

    mNucleusA = new Nucleus(
        aConfigurator->GetInt("NumberOfNucleonsA"),
        aConfigurator->GetDouble("SurfaceDiffusenessA"),
        aConfigurator->GetDouble("NuclearSaturationDensityA"),
        aConfigurator->GetDouble("RadiusA"));

    mNucleusB = new Nucleus(
        aConfigurator->GetInt("NumberOfNucleonsB"),
        aConfigurator->GetDouble("SurfaceDiffusenessB"),
        aConfigurator->GetDouble("NuclearSaturationDensityB"),
        aConfigurator->GetDouble("RadiusB"));

    mb      = aConfigurator->GetDouble("ImpactParameter");
    mcross  = aConfigurator->GetDouble("InelasticCrossSection");

}

/*!
 *
 *  \brief Destructor
 *
 */

OpticalGlauber::~OpticalGlauber(){

    PRINT_DEBUG_3("<OpticalGlauber::~OpticalGlauber>");

    delete mNucleusA;
    delete mNucleusB;

}

/*!
 *
 *  \brief Calculates density of binary collisions
 *
 *  \li x - x coordinate of the point in transverse plane
 *  \li y - y coordinate of the point in transverse plane
 *
 */

double OpticalGlauber::GetBC(double x, double y){

    PRINT_DEBUG_3("<OpticalGlauber::GetBC>");

    return mcross * mNucleusA->GetNuclearThickness(x - mb / 2.0,y) * mNucleusB->GetNuclearThickness(x + mb / 2.0,y);

}

/*!
 *
 *  \brief Calculates density of wounded nucleons in nucleus A
 *
 *  \li x - x coordinate of the point in transverse plane
 *  \li y - y coordinate of the point in transverse plane
 *
 */

double OpticalGlauber::GetWNA(double x, double y){

    PRINT_DEBUG_3("<OpticalGlauber::GetWNA>");

    return mNucleusA->GetNuclearThickness(x - mb / 2.0,y) * (1 - exp(- mcross * mNucleusB->GetNuclearThickness(x + mb / 2.0,y)));

}

/*!
 *
 *  \brief Calculates density of wounded nucleons in nucleus B
 *
 *  \li x - x coordinate of the point in transverse plane
 *  \li y - y coordinate of the point in transverse plane
 *
 */

double OpticalGlauber::GetWNB(double x, double y){

    PRINT_DEBUG_3("<OpticalGlauber::GetWNB>");

    return mNucleusB->GetNuclearThickness(x + mb / 2.0,y) * (1 - exp(- mcross * mNucleusA->GetNuclearThickness(x - mb / 2.0,y)));

}

/*!
 *
 *  \brief Sets the impact parameter of the nucleus
 *
 *  \li ab = impact parameter of the nucleus
 *
 */

double OpticalGlauber::SetImpactParameter(double ab){

    PRINT_DEBUG_3("<OpticalGlauber::SetImpactParameter>");

    mb = ab;
    return mb;

}

/*!
 *
 *  \brief Returns the impact parameter of the nucleus
 *
 *  \li ab = impact parameter of the nucleus
 *
 */

double OpticalGlauber::GetImpactParameter(void){

    PRINT_DEBUG_3("<OpticalGlauber::GetImpactParameter>");

    return mb;

}

/*!
 *
 *  \brief Print information
 *
 */

void OpticalGlauber::PrintInfo(void){

    PRINT_DEBUG_3("<OpticalGlauber::PrintInfo>");

    PrintRawMessage("");
    PrintRawMessage("\t- > Optical Glauber");
    PrintRawMessage("\tImpact Parameter [fm]",mb);
    PrintRawMessage("\tInelastic Cross Section [mb]",mcross * 10);
    PrintRawMessage("");
    PrintRawMessage("\tNucleus A");
    mNucleusA->PrintInfo();
    PrintRawMessage("");
    PrintRawMessage("\tNucleus B");
    mNucleusB->PrintInfo();

}
