/**
*
*   ParametrizationFreezer.cxx
*
*   Copyright (C) 2017 Radoslaw Ryblewski and Michael Strickland
*
*   This file is part of aHydro3p1.
*
*   aHydro3p1 is free software: you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation, either version 3 of the License, or
*   (at your option) any later version.
*
*   aHydro3p1 is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with aHydro3p1.  If not, see <http://www.gnu.org/licenses/>.
*
*/

#include "Global.h"

/*!
*
*  \brief Default constructor
*
*  \li aConfigurator = pointer to configuration
*  \li aBINDirectory = directory containing the binary file to perform freeze-out extraction on
*
*/

ParametrizationFreezer::ParametrizationFreezer(Configurator *aConfigurator, string aBINDir)
: mConfigurator(aConfigurator), mBINDir(aBINDir), mGrid(0), mField(0), mVArr(0), mNumberOfEquations(0){

    PRINT_DEBUG_3("<ParametrizationFreezer::ParametrizationFreezer>");

    mTSVDir  = mConfigurator->GetParameter("TSVDir");
    mXMLDir  = mConfigurator->GetParameter("XMLDir");
    mHydroID = GetFilename(GetDirectory(aBINDir));


    // error for freeze-out extraction by findroot()

    epsilon = (double)1e-10;

    //options for freeze-out extraction

    TFO = mConfigurator->GetDouble("TFO") / hbarc; // GeV -> fm^-1
    NZ  = mConfigurator->GetInt("NZ");
    NP  = mConfigurator->GetInt("NP");
    NT  = mConfigurator->GetInt("NT");
    lambda  = mConfigurator->GetDouble("Lambda");


    // ranges for angles phi, zeta, theta

    zetamin    = Pi * mConfigurator->GetDouble("Zmin");
    zetamax    = Pi * mConfigurator->GetDouble("Zmax");
    phimin     = Pi * mConfigurator->GetDouble("Pmin");
    phimax     = Pi * mConfigurator->GetDouble("Pmax");
    thetamin   = Pi * mConfigurator->GetDouble("Tmin");
    thetamax   = Pi * mConfigurator->GetDouble("Tmax");

    // spacing in each direction

    dzeta  = (zetamax - zetamin) / (NZ - 1);
    dphi   = (phimax  - phimin ) / (NP - 1);
    dtheta = (thetamax - thetamin) / (NT - 1);

    // switch between different hydrodynamic equations types

    mEquationsChoice  = mConfigurator->GetInt("Equations");

    switch(mEquationsChoice){

        case 1: // perfect fluid

            mNumberOfEquations = 4;
            mNames   = new string [mNumberOfEquations];
            mNamesFO = new string [mNumberOfEquations];
            mUnitsFO = new string [mNumberOfEquations];

            mNames[0] = "lnt";
            mNames[1] = "rap";
            mNames[2] = "ux";
            mNames[3] = "uy";

            mNamesFO[0] = "Distance";
            mNamesFO[1] = "FluidRap";
            mNamesFO[2] = "FluidUx";
            mNamesFO[3] = "FluidUy";

            mUnitsFO[0] = "GeV^-1";
            mUnitsFO[1] = "1";
            mUnitsFO[2] = "1";
            mUnitsFO[3] = "1";

            break;

        case 2: // anisotropic fluid

            mNumberOfEquations = 5;
            mNames   = new string [mNumberOfEquations];
            mNamesFO = new string [mNumberOfEquations];
            mUnitsFO = new string [mNumberOfEquations];

            mNames[0] = "lambda";
            mNames[1] = "rap";
            mNames[2] = "ux";
            mNames[3] = "uy";
            mNames[4] = "xi";

            mNamesFO[0] = "Distance";
            mNamesFO[1] = "FluidRap";
            mNamesFO[2] = "FluidUx";
            mNamesFO[3] = "FluidUy";
            mNamesFO[4] = "FluidXi";

            mUnitsFO[0] = "GeV^-1";
            mUnitsFO[1] = "1";
            mUnitsFO[2] = "1";
            mUnitsFO[3] = "1";
            mUnitsFO[4] = "1";

            break;

        case 3: // anisotropic fluid - new

            mNumberOfEquations = 9; // 8 + additional lambda table for THERMINATOR
            mNames   = new string [mNumberOfEquations];
            mNamesFO = new string [mNumberOfEquations];
            mUnitsFO = new string [mNumberOfEquations];

            mNames[0] = "temp";
            mNames[1] = "ax";
            mNames[2] = "ay";
            mNames[3] = "az";
            mNames[4] = "ux";
            mNames[5] = "uy";
            mNames[6] = "rap";
            mNames[7] = "lambda";
            mNames[8] = "temp";

            mNamesFO[0] = "Distance"; //?
            mNamesFO[1] = "FluidAx";
            mNamesFO[2] = "FluidAy";
            mNamesFO[3] = "FluidAz";
            mNamesFO[4] = "FluidUx";
            mNamesFO[5] = "FluidUy";
            mNamesFO[6] = "FluidRap";
            mNamesFO[7] = "FluidLambda";
            mNamesFO[8] = "FluidTeff";
            
            mUnitsFO[0] = "GeV^-1";
            mUnitsFO[1] = "1";
            mUnitsFO[2] = "1";
            mUnitsFO[3] = "1";
            mUnitsFO[4] = "1";
            mUnitsFO[5] = "1";
            mUnitsFO[6] = "1";
            mUnitsFO[7] = "GeV";
            mUnitsFO[8] = "GeV";

            break;

            default:

                exit(ProgramStatus(ERROR_OPTION_MISSING,
                "ParametrizationFreezer::ParametrizationFreezer: Cannot find Equations for " + d2s(mEquationsChoice)));
    }

    // read .bin file containing "temperature"-like quantity into mField, be carefull about units

    Parser *tParser = 0;
    tParser = new Parser(mBINDir + mNames[0] + ".bin");
    mField = tParser->ReadBIN(mConfigurator);
    delete tParser;

    // ranges of hydro grid

    tmin = mField->GetTMin();
    tmax = mField->GetTMax();
    xmax = mField->GetXMax();
    ymax = mField->GetYMax();
    emax = mField->GetZMax();

    deltat = (mField->GetTMax() - mField->GetTMin()) / (mField->GetTPts() - 1);
    deltax = 2 * mField->GetXMax() / (mField->GetXPts() - 1);
    deltay = 2 * mField->GetYMax() / (mField->GetYPts() - 1);
    deltae = 2 * mField->GetZMax() / (mField->GetZPts() - 1);


    PrintMessage("EQUATION OF STATE");
    mEOS = GetEquationOfState();
    meFO = mEOS->ET(TFO);

    PrintInfo();

    PrintRawMessage("");
    PrintRawMessage("Allocate memory for output arrays.");
    PrintRawMessage("Memory needed [MB]", mNumberOfEquations * NZ * NP * NT * sizeof(double) / (1024 * 1024));

    // allocate arrays for extracted values

    mVArr = new Vector3D* [mNumberOfEquations];
    for(int n = 0; n < mNumberOfEquations; ++n){
        mVArr[n] = new Vector3D(mNamesFO[n],
                        phimin,   phimax,   NP,
                        zetamin,  zetamax,  NZ,
                        thetamin, thetamax, NT);
    }

}

/*!
*
*  \brief Default destructor
*
*/

ParametrizationFreezer::~ParametrizationFreezer(){

    PRINT_DEBUG_3("<ParametrizationFreezer::~ParametrizationFreezer>");

    // free memory

    if(mField){
        delete mField;
        mField = 0;
    }
    
    if(mVArr){
        delete [] mVArr;
        mVArr = 0;
    }

    if(mNames){
        delete [] mNames;
        mNames = 0;
    }

    if(mNamesFO){
        delete [] mNamesFO;
        mNamesFO = 0;
    }

    if(mUnitsFO){
        delete [] mUnitsFO;
        mUnitsFO = 0;
    }

}

void ParametrizationFreezer::ExtractFOHS(void){

    switch(mEquationsChoice){

        case 1: // perfect fluid
            ExtractFOHSPerfectFluid();
            break;
        case 2: // anisotropic fluid
            ExtractFOHSAnisotropicFluid();
            break;
        case 3: // anisotropic fluid new
            ExtractFOHSAnisotropicFluidNew();
            break;

        default:

            exit(ProgramStatus(ERROR_OPTION_MISSING,
            "ParametrizationFreezer::ExtractFOHS: Cannot find Equations for " + d2s(mEquationsChoice)));
    }
}

void ParametrizationFreezer::ExtractFOHSPerfectFluid(void){

    // main loop over angles which parametrize entire space-time

    // for fancy bar
    char limit = '|';
    int full = 100;
    int prog = 0;

    // for simplicity

    Vector3D *tVec0 = mVArr[0];

    PrintMessage("EXTRACTION");

    double zeta, phi, theta;
    double dmin, dmid, dmax;
    int MaxI = 100;

    cout << "phi" << /* "\tzeta" << "\ttheta" << */endl;
    phi = phimin;
    for(int i = 0; i < NP; ++i){
    	cout << "\r" << phi * rad2deg << flush;
        // fancy bar begin
        if(prog != (int)(100 * phi / phimax)) {
            prog = (int)(100 * phi / phimax);
            cout << "\r\t" << limit;
            for(int m = 0; m < prog; m++) cout << '*';
            for(int m = prog; m < full; m++) cout << ' ';
            cout << limit << prog << "%" << flush;
        }
        // fancy bar end

        zeta = zetamin;
        for(int j = 0; j < NZ; ++j){
            //cout << "\r\t" << zeta * rad2deg << flush;
            theta = thetamin;
            for(int k = 0; k < NT; ++k){
                //cout << "\r\t\t" << theta * rad2deg << flush;

                // set the range for search of d: <dmin,dmax>

                dmin = 0.0;
                dmax = dmaxt(phi, zeta, theta) - epsilon; // minus small number to be sure dmax is in the field domain


                // extract value of d at which T = T_F0

                dmid = FindRoot(&ParametrizationFreezer::f1, phi, zeta, theta, MaxI, dmin, dmax, epsilon);

                // put the extracted value into the table

                (*tVec0)(i,j,k) = dmid / hbarc;
                theta += dtheta;
            }
            zeta += dzeta;
        }
        phi += dphi;
    }
    cout << endl << endl;

    PrintMessage("Done");

    // doing loops for n = 0 and 1 <= n < mNumberOfEquations> independently saves the RAM memory during freeze-out extraction

    // loop over all remaining functions

    Vector3D *tVec = 0;

    for(int n = 1; n < mNumberOfEquations; ++n){

        tVec = mVArr[n];

        // load data into tField

        Parser *tParser = 0;
        tParser = new Parser(mBINDir + mNames[n] + ".bin");
        mField = tParser->ReadBIN(mConfigurator, 0);
        delete tParser;

        PrintRawMessage("Extracting freeze-out data from " + mNames[n] + " table.");
        double temp = 0;
        phi = phimin;
        for(int i = 0; i < NP; ++i){
            zeta = zetamin;
            for(int j = 0; j < NZ; ++j){
                theta = thetamin;
                for(int k = 0; k < NT; ++k){
                    dmid = (*tVec0)(i,j,k) * hbarc;
                    temp = static_cast<double>(mField->Interpolate(
                    tau(dmid, phi, zeta, theta),
                    x(  dmid, phi, zeta, theta),
                    y(  dmid, phi, zeta, theta),
                    eta(dmid, phi, zeta, theta)
                    ));
                    (*tVec)(i,j,k) = temp; //? temp : 1e-10;
                    theta += dtheta;
                }
                zeta += dzeta;
            }
            phi += dphi;
        }
    }
    PrintMessage("Done");

    // export arrays to .xml file

    MakeDirectory(mXMLDir);
    SetDescriptions();
    ExportFOHSXML(mXMLDir + mHydroID + "_FOHS.xml");

        // output test data
    ExportVector3D("d", tVec0, 1);
//    ExportVector3D("r", mVArr[1], 1.);
//    ExportVector3D("ux", mVArr[2], 1.);
//    ExportVector3D("uy", mVArr[3], 1.);
//    ExportAllVectors3D(); // this function may be used to export raw fohs data

    PrintMessage("Done");

}

void ParametrizationFreezer::ExtractFOHSAnisotropicFluid(void){

    // read .bin file containing "xi" field into mField, be carefull about units

    Parser *tParser = 0;
    tParser = new Parser(mBINDir + mNames[4] + ".bin");
    mFieldX = tParser->ReadBIN(mConfigurator);
    delete tParser;

    // main loop over angles which parametrize entire space-time

    // for fancy bar
    char limit = '|';
    int full = 100;
    int prog = 0;

    // for simplicity

    Vector3D *tVec0 = mVArr[0];

    PrintMessage("EXTRACTION");

    double zeta, phi, theta;
    double dmin, dmid, dmax;
    int MaxI = 100;

    cout << "phi" << /* "\tzeta" << "\ttheta" << */endl;
    phi = phimin;
    for(int i = 0; i < NP; ++i){
    	cout << "\r" << phi * rad2deg << flush;
        // fancy bar begin
        if(prog != (int)(100 * phi / phimax)) {
            prog = (int)(100 * phi / phimax);
            cout << "\r\t" << limit;
            for(int m = 0; m < prog; m++) cout << '*';
            for(int m = prog; m < full; m++) cout << ' ';
            cout << limit << prog << "%" << flush;
        }
        // fancy bar end

        zeta = zetamin;
        for(int j = 0; j < NZ; ++j){
            //cout << "\r\t" << zeta * rad2deg << flush;
            theta = thetamin;
            for(int k = 0; k < NT; ++k){
                //cout << "\r\t\t" << theta * rad2deg << flush;

                // set the range for search of d: <dmin,dmax>

                dmin = 0.0;
                dmax = dmaxt(phi, zeta, theta) - epsilon; // minus small number to be sure dmax is in the field domain

                // extract value of d at which T = T_F0


                dmid = FindRoot(&ParametrizationFreezer::f2, phi, zeta, theta, MaxI, dmin, dmax, epsilon);

                // put the extracted value into the table

                (*tVec0)(i,j,k) = dmid / hbarc;
                theta += dtheta;
            }
            zeta += dzeta;
        }
        phi += dphi;
    }
    cout << endl << endl;

    PrintMessage("Done");

    // doing loops for n = 0 and n=<1,mNumberOfEquations> independently saves the RAM memory during freeze-out extraction

    // loop over all remaining functions

    Vector3D *tVec = 0;

    for(int n = 1; n < mNumberOfEquations; ++n){

        tVec = mVArr[n];

        // load data into tField

        Parser *tParser = 0;
        tParser = new Parser(mBINDir + mNames[n] + ".bin");
        mField = tParser->ReadBIN(mConfigurator, 0);
        delete tParser;

        PrintRawMessage("Extracting freeze-out data from " + mNames[n] + " table.");
        double temp = 0;
        phi = phimin;
        for(int i = 0; i < NP; ++i){
            zeta = zetamin;
            for(int j = 0; j < NZ; ++j){
                theta = thetamin;
                for(int k = 0; k < NT; ++k){
                    dmid = (*tVec0)(i,j,k) * hbarc;
                    temp = static_cast<double>(mField->Interpolate(
                    tau(dmid, phi, zeta, theta),
                    x(  dmid, phi, zeta, theta),
                    y(  dmid, phi, zeta, theta),
                    eta(dmid, phi, zeta, theta)
                    ));
                    (*tVec)(i,j,k) = temp;// ? temp : 1e-10;
                    theta += dtheta;
                }
                zeta += dzeta;
            }
            phi += dphi;
        }
    }
    PrintMessage("Done");

    // export arrays to .xml file

    MakeDirectory(mXMLDir);
    SetDescriptions();
    ExportFOHSXML(mXMLDir + mHydroID + "_FOHS.xml");

    // output test data
//    ExportVector3D("d", tVec0, 1);
//    ExportVector3D("r", mVArr[1], 1.);
//    ExportVector3D("ux", mVArr[2], 1.);
//    ExportVector3D("uy", mVArr[3], 1.);
//    ExportAllVectors3D();

    PrintMessage("Done");
}


void ParametrizationFreezer::ExtractFOHSAnisotropicFluidNew(void){

    // read .bin file containing the effective temperature be careful about units

    Parser *tParser = 0;
    tParser = new Parser(mBINDir + mNames[0] + ".bin");
    mField = tParser->ReadBIN(mConfigurator);
    delete tParser;

    // main loop over angles which parametrize entire space-time

    // for fancy bar
    char limit = '|';
    int full = 100;
    int prog = 0;

    // for simplicity

    Vector3D *tVec0 = mVArr[0];

    PrintMessage("EXTRACTION");

    double zeta, phi, theta;
    double dmin, dmid, dmax;
    int MaxI = 100;

    cout << "phi" << /* "\tzeta" << "\ttheta" << */endl;
    phi = phimin;
    for(int i = 0; i < NP; ++i){
    	cout << "\r" << phi * rad2deg << flush;
        // fancy bar begin
        if(prog != (int)(100 * phi / phimax)) {
            prog = (int)(100 * phi / phimax);
            cout << "\r\t" << limit;
            for(int m = 0; m < prog; m++) cout << '*';
            for(int m = prog; m < full; m++) cout << ' ';
            cout << limit << prog << "%" << flush;
        }
        // fancy bar end

        zeta = zetamin;
        for(int j = 0; j < NZ; ++j){
            //cout << "\r\t" << zeta * rad2deg << flush;
            theta = thetamin;
            for(int k = 0; k < NT; ++k){
                //cout << "\r\t\t" << theta * rad2deg << flush;

                // set the range for search of d: <dmin,dmax>

                dmin = 0.0 + epsilon;
                dmax = dmaxt(phi, zeta, theta) - epsilon; // minus small number to be sure dmax is in the field domain

                // check if the system droped down below T_F0

                if(f3(dmax, phi, zeta, theta) > 0 )    {
                    PrintRawMessage(
                    " Fireball temperature too HIGH to freeze out at :\t tau = "
                    + d2s(tau(dmid, phi, zeta, theta)) + "\t x = "
                    + d2s(x(dmid, phi, zeta, theta)) + "\t y = "
                    + d2s(y(dmid, phi, zeta, theta))+ "\t eta = "
                    + d2s(eta(dmid, phi, zeta, theta))
                    );
                    exit(ProgramStatus(ERROR_VALUE_NOT_FOUND,"Exit."));
                }
                if(f3(dmin, phi, zeta, theta) < 0)    {
                    PrintRawMessage(
                    " Fireball temperature too LOW to freeze out at :\t tau = "
                    + d2s(tau(dmid, phi, zeta, theta)) + "\t x = "
                    + d2s(x(dmid, phi, zeta, theta)) + "\t y = "
                    + d2s(y(dmid, phi, zeta, theta))+ "\t eta = "
                    + d2s(eta(dmid, phi, zeta, theta))
                    );
                    exit(ProgramStatus(ERROR_VALUE_NOT_FOUND,"Exit."));

                }


                // extract value of d at which T = T_F0

                dmid = FindRoot(&ParametrizationFreezer::f3, phi, zeta, theta, MaxI, dmin, dmax, epsilon);

                // put the extracted value into the table

                (*tVec0)(i,j,k) = dmid / hbarc;
                theta += dtheta;
            }
            zeta += dzeta;
        }
        phi += dphi;
    }
    cout << endl << endl;

    PrintMessage("Done");

    // doing loops for n = 0 and n=<1,mNumberOfEquations> independently saves the RAM memory during freeze-out extraction

    // loop over all remaining functions  - > UP TO n=7 SINCE WE NEED lambda IN THERMINATOR TOO

    Vector3D *tVec = 0;

    for(int n = 1; n < mNumberOfEquations; ++n){

        double tmult = 1;
        if(n > mNumberOfEquations -3) tmult*=hbarc; // for setting GeV units for lambda

        tVec = mVArr[n];

        // load data into mField (to save memory)

        Parser *tParser = 0;
        tParser = new Parser(mBINDir + mNames[n] + ".bin");
        mField = tParser->ReadBIN(mConfigurator, 0);
        delete tParser;

        PrintRawMessage("Extracting freeze-out data from " + mNames[n] + " table.");
        double temp = 0;
        phi = phimin;
        for(int i = 0; i < NP; ++i){
            zeta = zetamin;
            for(int j = 0; j < NZ; ++j){
                theta = thetamin;
                for(int k = 0; k < NT; ++k){
                    dmid = (*tVec0)(i,j,k) * hbarc;
                    temp = static_cast<double>(mField->Interpolate(
                    tau(dmid, phi, zeta, theta),
                    x(  dmid, phi, zeta, theta),
                    y(  dmid, phi, zeta, theta),
                    eta(dmid, phi, zeta, theta)
                    ));
                    (*tVec)(i,j,k) = temp * tmult;// ? temp : 1e-10;
                    theta += dtheta;
                }
                zeta += dzeta;
            }
            phi += dphi;
        }
    }
    PrintMessage("Done");

    // export arrays to .xml file

    MakeDirectory(mXMLDir);
    SetDescriptions();
    ExportFOHSXML(mXMLDir + mHydroID + "_FOHS.xml");

    // output test data
    ExportVector3D("d", tVec0, 1);
//    ExportVector3D("r", mVArr[1], 1.);
//    ExportVector3D("ux", mVArr[2], 1.);
//    ExportVector3D("uy", mVArr[3], 1.);
//    ExportAllVectors3D();

    PrintMessage("Done");
}

double ParametrizationFreezer::f1(double d, double p, double z, double t){

    return mField->Interpolate(tau(d, p, z, t), x(d, p, z, t), y(d, p, z, t), eta(d, p, z, t)) - TFO; // note the field should be in fm here

}

double ParametrizationFreezer::f2(double d, double p, double z, double t){

    return mField->Interpolate(tau(d, p, z, t), x(d, p, z, t), y(d, p, z, t), eta(d, p, z, t))
            * pow(R(mFieldX->Interpolate(tau(d, p, z, t), x(d, p, z, t), y(d, p, z, t), eta(d, p, z, t))),0.25) - TFO; // note the field should be in fm here

}

double ParametrizationFreezer::f3(double d, double p, double z, double t){

    return mField->Interpolate(tau(d, p, z, t), x(d, p, z, t), y(d, p, z, t), eta(d, p, z, t)) - TFO;
}

EquationOfState* ParametrizationFreezer::GetEquationOfState(){

    EquationOfState *tEquationOfState = 0;

    switch(mConfigurator->GetInt("EquationOfState")){

        case 1:

            tEquationOfState = new IdealEquationOfState(
                                                        mConfigurator->GetInt("IdealEquationOfStateStatistics"),
                                                        mConfigurator->GetDouble("IdealEquationOfStateDegeneracy"),
                                                        mConfigurator->GetInt("IdealEquationOfStateQuarkFlavors"));

            break;

        case 2:

            tEquationOfState = new RealisticEquationOfState(
                                                            mConfigurator->GetParameter("RealisticEquationOfStateDirectory"));

            break;

        default:

            exit(ProgramStatus(ERROR_OPTION_MISSING,
                               "ParametrizationFreezer::GetEquationOfState: Cannot find equation of state for " + mConfigurator->GetParameter("EquationOfState")));
    }

    tEquationOfState->PrintInfo();

    return tEquationOfState;
}

double ParametrizationFreezer::R(double x) {

    PRINT_DEBUG_3("<ParametrizationFreezer::R>");

	if (x==0) return 1;
	if (x>0) return (1/(1 + x) + atan(sqrt(x))/sqrt(x))/2.;
	if (x<0) return (1/(1 - fabs(x)) + atanh(sqrt(fabs(x)))/sqrt(fabs(x)))/2.;
    return 1;
}

inline double ParametrizationFreezer::phi0(void){

    return atan(ymax / xmax);
}

inline double ParametrizationFreezer::zeta0(double phi){

    return atan((tmax - tmin) / dmaxp(phi));
}

inline double ParametrizationFreezer::theta0(double phi, double zeta){

    return atan(dmaxz(phi, zeta) / (lambda * emax));
}

inline double ParametrizationFreezer::dmaxp(double phi){

    if(phi < phi0() || 2 * Pi - phi0() < phi || (Pi - phi0() < phi && phi < Pi + phi0())) return fabs(xmax / cos(phi));
    else return fabs(ymax / sin(phi));
}

inline double ParametrizationFreezer::dmaxz(double phi, double zeta){

    if(zeta < zeta0(phi)) return fabs(dmaxp(phi) / cos(zeta));
    else return fabs((tmax -tmin) / sin(zeta));
}

inline double ParametrizationFreezer::dmaxt(double phi, double zeta, double theta){

    if(theta < theta0(phi, zeta) || Pi - theta0(phi, zeta) < theta) return fabs((lambda * emax) / cos(theta));
    else return fabs(dmaxz(phi, zeta) / sin(theta));
}

inline double ParametrizationFreezer::tau(double d, double phi, double zeta, double theta){

    return tmin + d * sin(theta) * sin(zeta) ;
}

inline double ParametrizationFreezer::x(double d, double phi, double zeta, double theta){

    return d * sin(theta) * cos(zeta) * cos(phi);
}

inline double ParametrizationFreezer::y(double d, double phi, double zeta, double theta){

    return d * sin(theta) * cos(zeta) * sin(phi);
}

inline double ParametrizationFreezer::eta(double d, double phi, double zeta, double theta){

    return d * cos(theta) / lambda;
}

// find root with bisection method
inline double ParametrizationFreezer::FindRoot(double (ParametrizationFreezer::*aFunction)(double, double, double, double), double p, double z, double t, int MaxIterations, double xmin, double xmax, double eps){

    double xmid;

    while (fabs(xmax - xmin) > eps){
        --MaxIterations;
        xmid = 0.5 * (xmin + xmax);
        if((this->*aFunction)(xmid, p, z, t) * (this->*aFunction)(xmin, p, z, t) < 0) xmax = xmid;
        else xmin = xmid;
    }
    if(MaxIterations < 0){
        PrintRawMessage("FindRoot: Max_iterations i = " + d2s(MaxIterations) + " exceeded.\n");
        PrintRawMessage("FindRoot: Can't find root at the given level of accuracy in this range. \n");
    }
    return xmid;
}

/*!
*
*  \brief Prints Info
*
*/

void ParametrizationFreezer::PrintInfo(void){

    PrintMessage(" FREEZE-OUT SETTINGS ");
    PrintRawMessage("Freeze-out method", "(d,zeta,phi,theta) parametrization");
    PrintRawMessage("Freeze-out temperature [GeV]", TFO * hbarc);
    PrintRawMessage("Freeze-out energy density [GeV/fm^3]", meFO * hbarc);
    PrintRawMessage("");
    PrintRawMessage("Field domain:");
    PrintRawMessage("tau", "(" + d2s(tmin ) + "," + d2s(tmax) + ")");
    PrintRawMessage("x",   "(" + d2s(-xmax) + "," + d2s(xmax) + ")");
    PrintRawMessage("y",   "(" + d2s(-ymax) + "," + d2s(ymax) + ")");
    PrintRawMessage("eta", "(" + d2s(-emax) + "," + d2s(emax) + ")");
    PrintRawMessage("Field grid:");
    PrintRawMessage("tau", deltat);
    PrintRawMessage("x",   deltax);
    PrintRawMessage("y",   deltay);
    PrintRawMessage("eta", deltae);
    PrintRawMessage("");
    PrintRawMessage("Parametrization ranges [deg]:");
    PrintRawMessage("zeta", "(" + d2s(zetamin * rad2deg) + ", " + d2s(zetamax * rad2deg) + ")");
    PrintRawMessage("phi", "(" + d2s(phimin * rad2deg) + ", " + d2s(phimax * rad2deg) + ")");
    PrintRawMessage("theta", "(" + d2s(thetamin * rad2deg) + ", " + d2s(thetamax * rad2deg) + ")");
    PrintRawMessage("");
    PrintRawMessage("Parametrization spacing [deg]:");
    PrintRawMessage("zeta", dzeta * rad2deg);
    PrintRawMessage("phi", dphi * rad2deg);
    PrintRawMessage("theta", dtheta * rad2deg);
    PrintRawMessage("");
    PrintRawMessage("Number of points:");
    PrintRawMessage("zeta",  NZ);
    PrintRawMessage("phi",   NP);
    PrintRawMessage("theta", NT);
    PrintRawMessage("Total", NZ * NP * NT );
    PrintRawMessage("");
    PrintRawMessage("Lambda", lambda);
    PrintRawMessage("Accuracy", epsilon);
    PrintRawMessage("");

}

void ParametrizationFreezer::SetDescriptions(void){

    // set descriptions of parameters

    mEquations = "UNKNOWN";
    switch(mConfigurator->GetInt("Equations")){

        case 1:
            mEquations = "3+1D Perfect Fluid";
            break;

        case 2:
            mEquations = "3+1D Anisotropic Fluid (LO)";
            break;

        case 3:
            mEquations = "3+1D Anisotropic Fluid (LO) New";
            break;

        default:
            PrintRawMessage("Warning: Unknown option for", "<Equations>.");
    }

    mEquationOfState = "UNKNOWN";
    switch(mConfigurator->GetInt("EquationOfState")){

        case 1:
            mEquationOfState = "Ideal";
            break;

        case 2:
            mEquationOfState = "Realistic";
            break;

        default:
            PrintRawMessage("Warning: Unknown option for", "<EquationOfState>.");
    }

    mDeviceType = "UNKNOWN";
//    switch(mConfigurator->GetInt("Device")){

//        case 1:
//            mDeviceType = "RHIC";
//            break;
//
//        case 2:
//            mDeviceType = "LHC";
//            break;
//
//        default:
//            PrintRawMessage("Warning: Unknown option for", "<Device>.");
//    }

    mCollisionalSystem = "UNKNOWN";
//    switch(mConfigurator->GetInt("CollisionalSystem")){
//
//        case 1:
//            mCollisionalSystem = "AuAu";
//            break;
//
//        case 2:
//            mCollisionalSystem = "PbPb";
//            break;
//
//        default:
//            PrintRawMessage("Warning: Unknown option for", "<CollisionalSystem>.");
//    }

    mCMin = "0";
    mCMax = "0";

    mInitialSource = "UNKNOWN";
    switch(mConfigurator->GetInt("InitialSource")){

        case 0:
            mInitialSource = "Flat";
            break;

        case 1:
            mInitialSource = "3D Gauss";
            break;

        case 2:
            mInitialSource = "Glauber Inspired - Shifted source";
            break;

        case 3:
            mInitialSource = "Glauber Inspired - Tilted source";
            break;

        case 4:
            mInitialSource = "Sum of Random 3D Gaussian Bumps";
            break;

        case 5:
            mInitialSource = "Monte Carlo Glauber";
            break;

        default:
            PrintRawMessage("Warning: Unknown option for", "<InitialSource>.");
    }

    mDensityScaling = "UNKNOWN";
    switch(mConfigurator->GetInt("DensityOfSourcesScaling")){

        case 1:
            mDensityScaling = "Entropy density";
            break;

        case 2:
            mDensityScaling = "Energy density";
            break;

        default:
            PrintRawMessage("Warning: Unknown option for", "<DensityOfSourcesScaling>.");
    }

    mDensityNormalization = "UNKNOWN";
    switch(mConfigurator->GetInt("DensityOfSourcesNormalization")){

        case 1:
            mDensityNormalization = "Normalize to central";
            break;

        case 2:
            mDensityNormalization = "Normalize to non-central";
            break;

        default:
            PrintRawMessage("Warning: Unknown option for", "<DensityOfSourcesNormalization>.");
    }

    mUxInitialFlow = "UNKNOWN";
    switch(mConfigurator->GetInt("UxInitialFlow")){

        case 1:
            mUxInitialFlow= "No flow";
            break;

        default:
            PrintRawMessage("Warning: Unknown option for", "<UxInitialFlow>.");
    }

    mUyInitialFlow = "UNKNOWN";
    switch(mConfigurator->GetInt("UyInitialFlow")){

        case 1:
            mUyInitialFlow = "No flow";
            break;

        default:
            PrintRawMessage("Warning: Unknown option for", "<UyInitialFlow>.");
    }

    mRInitialFlow = "UNKNOWN";
    switch(mConfigurator->GetInt("RInitialFlow")){

        case 1:
            mRInitialFlow = "Bjorken";
            break;

        default:
            PrintRawMessage("Warning: Unknown option for", "<RInitialFlow>.");
    }

    mIdealStatistics = "UNKNOWN";
    switch(mConfigurator->GetInt("IdealEquationOfStateStatistics")){

        case 1:
            mIdealStatistics = "Boltzmann";
            break;

        case 2:
            mIdealStatistics = "BE";
            break;

        case 3:
            mIdealStatistics = "FD";
            break;

        default:
            PrintRawMessage("Warning: Unknown option for", "<IdealEquationOfStateStatistics>.");
    }
}

void ParametrizationFreezer::ExportFOHSXML(string aFileName){

    ofstream tFile;
    ostringstream oss;

    PrintRawMessage("Exporting freeze-out hypersurface to " + aFileName + " file.");

	tFile.open(aFileName.c_str(), ios::trunc | ios::out);

    //export the XML tags to the .xml file

    oss << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" << endl;
    oss << "<HYPERSURFACE filename=\"" << mHydroID << "\" version=\"1.0\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xsi:noNamespaceSchemaLocation=\"http://therminator2.ifj.edu.pl/hypersurface.xsd\">" << endl;
    oss << GetPARAMETERS("\t") << endl;
    oss << GetDESCRIPTION("\t") << endl;
    oss << "<!-- Included VECTOR3D names : ";
    for(int n = 0; n < mNumberOfEquations;++n) oss << mNamesFO[n] << " ";
    oss << " -->" << endl;
    tFile << oss.str(); oss.str("");
    for(int n = 0; n < mNumberOfEquations;++n) ExportVector("\t",tFile,n);
    oss << "</HYPERSURFACE>" << endl;
    tFile << oss.str(); oss.str("");
	tFile.close();

}

void ParametrizationFreezer::ExportVector(string Tab, ofstream &aFile, int n){

    Vector3D* const tVec = mVArr[n];
    ostringstream oss;
    oss << "<VECTOR3D name=\"" + mNamesFO[n] + "\" unit=\"" + mUnitsFO[n] + "\">" << endl;
    oss << GetAXIS("\t","Zeta", tVec->GetYMin(),tVec->GetYMax(),tVec->GetYPts(),"rad") << endl;
    oss << GetAXIS("\t","Phi",  tVec->GetXMin(),tVec->GetXMax(),tVec->GetXPts(),"rad") << endl;
    oss << GetAXIS("\t","Theta",tVec->GetZMin(),tVec->GetZMax(),tVec->GetZPts(),"rad") << endl;
    aFile << oss.str(); oss.str("");
    GetDATA("\t",aFile, mVArr[n], tVec->GetXPts() * tVec->GetYPts() * tVec->GetZPts());
    oss << "</VECTOR3D>";
    aFile << oss.str(); oss.str("");
}

void ParametrizationFreezer::GetDATA(string Tab,ofstream &aFile, Vector3D* const aVec, long long int aPts){

    ostringstream oss;
    oss << Tab << "<DATA points=\"" << aPts << "\">" << endl;
    aFile << oss.str();oss.str("");
    ExportVector3DXml(aFile, aVec);
    oss << Tab << "</DATA>" << endl;
    aFile << oss.str();oss.str("");

}

void ParametrizationFreezer::ExportVector3DXml(ofstream &aFile, Vector3D* const aVec){
//
//    aFile.precision(6);
    stringstream tBuffer;

    for (int i = 0; i < NZ; ++i){
        for (int j = 0; j < NP; ++j)
            for (int k = 0; k < NT; ++k){
			tBuffer << (*aVec)(j,i,k) << "\n"; // carefull, different order of indices required by THERMINATOR
            }
    }
    aFile << tBuffer.rdbuf();
    tBuffer.str("");

}

string ParametrizationFreezer::GetAXIS(string Tab, string aAxisName, double aMin, double aMax, int aPts, string aUnit){

    ostringstream oss;
    oss << Tab <<"<AXIS name=\"" << aAxisName << "\">" << endl;
    oss << Tab << Tab << GetTAG("DETAIL","min", aUnit, aMin) << endl;
    oss << Tab << Tab << GetTAG("DETAIL","max", aUnit, aMax) << endl;
    oss << Tab << Tab << GetTAG("DETAIL","pts", "1", static_cast<double>(aPts)) << endl;
    oss << Tab <<"</AXIS>";

    return oss.str();
}

string ParametrizationFreezer::GetPARAMETERS(string Tab){

    ostringstream oss;

    // export tags

    oss << "<PARAMETERS>" << endl;
    oss << Tab << GetTAG("PARAMETER","Tau_i","fm", mConfigurator->GetDouble("TauMin")) << endl;
    oss << Tab << GetTAG("PARAMETER","Lambda", "fm", mConfigurator->GetDouble("Lambda")) << endl;
    oss << Tab << GetTAG("PARAMETER","Temperature", "MeV", mConfigurator->GetDouble("TFO") * 1000) << endl;
    oss << Tab << GetTAG("PARAMETER","Mu_B", "MeV", mConfigurator->GetDouble("Mu_B")) << endl;
    oss << Tab << GetTAG("PARAMETER","Mu_I", "MeV", mConfigurator->GetDouble("Mu_I")) << endl;
    oss << Tab << GetTAG("PARAMETER","Mu_S", "MeV", mConfigurator->GetDouble("Mu_S")) << endl;
    oss << Tab << GetTAG("PARAMETER","Mu_C", "MeV", mConfigurator->GetDouble("Mu_C")) << endl;
    oss << Tab << GetTAG("PARAMETER","device", "1", mDeviceType) << endl;
    oss << Tab << GetTAG("PARAMETER","colliding_system", "1", mCollisionalSystem) << endl;
    oss << Tab << GetTAG("PARAMETER","colliding_energy", "GeV", mConfigurator->GetDouble("CollisionEnergy")) << endl;
    oss << Tab << GetTAG("PARAMETER","centrality_min", "%", mCMin) << endl;
    oss << Tab << GetTAG("PARAMETER","centrality_max", "%", mCMax) << endl;
    oss << Tab << GetTAG("PARAMETER","impact_parameter", "fm", mConfigurator->GetDouble("ImpactParameter")) << endl;
    oss << Tab << GetTAG("PARAMETER","temperature_at_center", "MeV", mConfigurator->GetDouble("InitialTemperature") * 1000) << endl;
    oss << "</PARAMETERS>";

    return oss.str();
}

string ParametrizationFreezer::GetDESCRIPTION(string Tab){

    ostringstream oss;

    oss << "<DESCRIPTION>" << endl;
        oss << Tab << "<ITEM name=\"" << "initial_nuclear_profile" << "\">" << endl;
        oss << Tab << Tab << GetTAG("DETAIL","type","1", mInitialSource) << endl;

    switch(mConfigurator->GetInt("InitialSource")){

        case 0:

            //oss << Tab << Tab << GetTAG("DETAIL","T0","MeV", mConfigurator->GetDouble("InitialTemperature") * 1000) << endl;
            oss << Tab << "</ITEM>" << endl;
            break;

        case 1:

            oss << Tab << Tab << GetTAG("DETAIL","sigma_x","fm", mConfigurator->GetDouble("SigmaX")) << endl;
            oss << Tab << Tab << GetTAG("DETAIL","sigma_y","fm", mConfigurator->GetDouble("SigmaY")) << endl;
            oss << Tab << Tab << GetTAG("DETAIL","sigma_eta","1", mConfigurator->GetDouble("SigmaZ")) << endl;
            //oss << Tab << Tab << GetTAG("DETAIL","T0","MeV", mConfigurator->GetDouble("InitialTemperature") * 1000) << endl;
            oss << Tab << "</ITEM>" << endl;
            break;

        case 2: case 3:

            oss << Tab << Tab << GetTAG("DETAIL","impact_parameter", "fm", mConfigurator->GetDouble("ImpactParameter")) << endl;
            oss << Tab << Tab << GetTAG("DETAIL","shift_in_rapidity", "1", mConfigurator->GetDouble("ShiftInRapidity")) << endl;
            oss << Tab << Tab << GetTAG("DETAIL","plateau_half_width", "1", mConfigurator->GetDouble("PlateauHalfWidth")) << endl;
            oss << Tab << Tab << GetTAG("DETAIL","gauss_half_width", "1", mConfigurator->GetDouble("GaussianHalfWidth")) << endl;
            //oss << Tab << Tab << GetTAG("DETAIL","T0", "MeV", mConfigurator->GetDouble("InitialTemperature") * 1000) << endl;
            oss << Tab << "</ITEM>" << endl;

            oss << Tab << "<ITEM name=\"" << "eos_to_initial_condition" << "\">" << endl;
            oss << Tab << mDensityScaling << endl;
            oss << Tab << "</ITEM>" << endl;

            oss << Tab << "<ITEM name=\"" << "normalization" << "\">" << endl;
            oss << Tab << mDensityNormalization << endl;
            oss << Tab << "</ITEM>" << endl;

            oss << Tab << "<ITEM name=\"" << "glauber_optical" << "\">" << endl;
            oss << Tab << Tab << GetTAG("DETAIL","InelasticCrossSection","mb", mConfigurator->GetDouble("InelasticCrossSection")) << endl;
            oss << Tab << Tab << GetTAG("DETAIL","MixingFactor", "1", mConfigurator->GetDouble("MixingFactor")) << endl;
            oss << Tab << Tab << GetTAG("DETAIL","NumberOfNucleonsA", "1", mConfigurator->GetDouble("NumberOfNucleonsA")) << endl;
            oss << Tab << Tab << GetTAG("DETAIL","NumberOfNucleonsB", "1", mConfigurator->GetDouble("NumberOfNucleonsB")) << endl;
            oss << Tab << Tab << GetTAG("DETAIL","SurfaceDiffusenessA", "fm", mConfigurator->GetDouble("SurfaceDiffusenessA")) << endl;
            oss << Tab << Tab << GetTAG("DETAIL","SurfaceDiffusenessB", "fm", mConfigurator->GetDouble("SurfaceDiffusenessB")) << endl;
            oss << Tab << Tab << GetTAG("DETAIL","NuclearSaturationDensityA", "1", mConfigurator->GetDouble("NuclearSaturationDensityA")) << endl;
            oss << Tab << Tab << GetTAG("DETAIL","NuclearSaturationDensityB", "1", mConfigurator->GetDouble("NuclearSaturationDensityB")) << endl;
            oss << Tab << "</ITEM>" << endl;

            break;
            
        case 5:
            
            oss << Tab << Tab << GetTAG("DETAIL","impact_parameter", "fm", mConfigurator->GetDouble("ImpactParameter")) << endl;
            //oss << Tab << Tab << GetTAG("DETAIL","shift_in_rapidity", "1", mConfigurator->GetDouble("ShiftInRapidity")) << endl;
            //oss << Tab << Tab << GetTAG("DETAIL","plateau_half_width", "1", mConfigurator->GetDouble("PlateauHalfWidth")) << endl;
            //oss << Tab << Tab << GetTAG("DETAIL","gauss_half_width", "1", mConfigurator->GetDouble("GaussianHalfWidth")) << endl;
            //oss << Tab << Tab << GetTAG("DETAIL","T0", "MeV", mConfigurator->GetDouble("InitialTemperature") * 1000) << endl;
            oss << Tab << "</ITEM>" << endl;
            
            break;

    }

    oss << Tab << "<ITEM name=\"" << "initial_flow" << "\">" << endl;
    oss << Tab << Tab << GetTAG("DETAIL","x-direction","1", mUxInitialFlow) << endl;
    oss << Tab << Tab << GetTAG("DETAIL","y-direction","1", mUyInitialFlow) << endl;
    oss << Tab << Tab << GetTAG("DETAIL","eta-direction","1", mRInitialFlow) << endl;
    oss << Tab << "</ITEM>" << endl;

    oss << Tab << "<ITEM name=\"" << "equation_of_state" << "\">" << endl;
    oss << Tab << Tab << GetTAG("DETAIL","type","1", mEquationOfState) << endl;
    if(mConfigurator->GetInt("EquationOfState") == 1){
    oss << Tab << Tab << GetTAG("DETAIL","statistics","1", mIdealStatistics) << endl;
    oss << Tab << Tab << GetTAG("DETAIL","degeneracy_factor","1", mConfigurator->GetDouble("IdealEquationOfStateDegeneracy")) << endl;
    oss << Tab << Tab << GetTAG("DETAIL","quark_flavors","1", mConfigurator->GetInt("IdealEquationOfStateQuarkFlavors")) << endl;
    }
    if(mConfigurator->GetInt("EquationOfState") == 2){
    oss << Tab << Tab << GetTAG("DETAIL","T ht TC","1", "QGP_from_lattice_QCD") << endl;
    oss << Tab << Tab << GetTAG("DETAIL","T eq TC","1", "crossover") << endl;
    oss << Tab << Tab << GetTAG("DETAIL","T lt TC","1", "Hadron_gas") << endl;
    }
    oss << Tab << "</ITEM>" << endl;

    oss << Tab << "<ITEM name=\"" << "hydrodynamic_equations" << "\">" << endl;
    oss << Tab << Tab << GetTAG("DETAIL","type","1",mEquations) << endl;
    oss << Tab << Tab << GetTAG("DETAIL","tau_min","fm", tmin) << endl;
    oss << Tab << Tab << GetTAG("DETAIL","tau_max","fm", tmax) << endl;
    oss << Tab << Tab << GetTAG("DETAIL","tau_step_size","1", deltat) << endl;
    oss << Tab << Tab << GetTAG("DETAIL","x_min","fm", -xmax) << endl;
    oss << Tab << Tab << GetTAG("DETAIL","x_max","fm",  xmax) << endl;
    oss << Tab << Tab << GetTAG("DETAIL","x_spacing","1", deltax) << endl;
    oss << Tab << Tab << GetTAG("DETAIL","y_min","fm", -ymax) << endl;
    oss << Tab << Tab << GetTAG("DETAIL","y_max","fm",  ymax) << endl;
    oss << Tab << Tab << GetTAG("DETAIL","y_spacing","1", deltay) << endl;
    oss << Tab << Tab << GetTAG("DETAIL","eta_min","fm", -emax) << endl;
    oss << Tab << Tab << GetTAG("DETAIL","eta_max","fm",  emax) << endl;
    oss << Tab << Tab << GetTAG("DETAIL","eta_spacing","1", deltae) << endl;
    oss << Tab << "</ITEM>" << endl;

    oss << "</DESCRIPTION>";

    return oss.str();
}

string ParametrizationFreezer::GetTAG(string aTagName, string aDetailName, string aUnit, double aValue){

    ostringstream oss;
    oss << "<" << aTagName << " name=\"" << aDetailName << "\" unit=\"" << aUnit << "\">" << aValue << "</" << aTagName << ">";
    return oss.str();

}

string ParametrizationFreezer::GetTAG(string aTagName, string aDetailName, string aUnit, int aValue){

    ostringstream oss;
    oss << "<" << aTagName << " name=\"" << aDetailName << "\" unit=\"" << aUnit << "\">" << aValue << "</" << aTagName << ">";
    return oss.str();

}

string ParametrizationFreezer::GetTAG(string aTagName, string aDetailName, string aUnit, string aValue){

    ostringstream oss;
    oss << "<" << aTagName << " name=\"" << aDetailName << "\" unit=\"" << aUnit << "\">" << aValue << "</" << aTagName << ">";
    return oss.str();

}

void ParametrizationFreezer::ExportVector3D(string aVectorName, Vector3D *aVec, double aMultip){

    fstream tSnapshotFile;
    string tSnapshotFileName = mTSVDir + mHydroID + "/" + aVectorName + "FO.tsv";
    PrintMessage("File", tSnapshotFileName + " was created.");
	tSnapshotFile.open(tSnapshotFileName.c_str(), ios::trunc | ios::out);
    stringstream tBuffer;
    double zeta, phi, theta;
    phi = phimin;
    for(int i = 0; i < NP; ++i){
        zeta = zetamin;
        for(int j = 0; j < NZ; ++j){
            theta = thetamin;
            for(int k = 0; k < NT; ++k){
                tBuffer << phi << "\t" << zeta << "\t" << theta << "\t" << (*aVec)(i,j,k) * aMultip << "\n";
                theta += dtheta;
            }
            zeta += dzeta;
        }
        phi += dphi;
    }
    tSnapshotFile << tBuffer.rdbuf(); // faster than tBuffer.str();
    tBuffer.str("");
	tSnapshotFile.close();
}

void ParametrizationFreezer::ExportAllVectors3D(void){

    double aMultip = 1;
    for(int n = 0; n < mNumberOfEquations;++n){
        if(n == 0) aMultip = hbarc;
        else aMultip = 1;
        ofstream tFile;
        string aFileName = mXMLDir + mNamesFO[n] + "_" + mHydroID + "_FOHS.dat";
        tFile.open(aFileName.c_str(), ios::trunc | ios::out);

        PrintRawMessage("Exporting freeze-out hypersurface to " + aFileName + " file.");

        stringstream tBuffer;

        tBuffer << "phi" << "\t"<< phimin << "\t" << phimax << "\t"  << NP << "\n";
        tBuffer << "zeta" << "\t"<< zetamin << "\t" << zetamax << "\t"  << NZ << "\n";
        tBuffer << "theta" << "\t"<< thetamin << "\t" << thetamax << "\t"  << NT << "\n";

        for (int i = 0; i < NP; ++i)
            for (int j = 0; j < NZ; ++j){
                for (int k = 0; k < NT; ++k){
                tBuffer << (*mVArr[n])(i,j,k) * aMultip << "\n"; // be carefull, different order of indices required by THERMINATOR
                }
        }
        tFile << tBuffer.rdbuf();
        tBuffer.str("");
        tFile.close();
    }

}
