/**
 *
 *   Parser.cxx
 *
 *   Copyright (C) 2017 Radoslaw Ryblewski and Michael Strickland
 *
 *   This file is part of aHydro3p1.
 *
 *   aHydro3p1 is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   aHydro3p1 is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with aHydro3p1.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "Global.h"

/*!
 *
 *  \brief Default constructor
 *
 */

Parser::Parser()
: mFileName(""), mFile(0){
    
    PRINT_DEBUG_3("<Parser::Parser()>");
    
}

/*!
 *
 *  \brief Default constructor
 *
 *  \li aFileName = file name to be parsed
 *
 */

Parser::Parser(string aFileName)
: mFileName(aFileName){
    
    PRINT_DEBUG_3("<Parser::Parser(const char*)>");
    
    mFile.open(mFileName.c_str());
    if((!mFile) && (!mFile.is_open())) exit(ProgramStatus(ERROR_FILE_NOT_FOUND, mFileName.c_str()));
    PrintRawMessage("Reading from file", mFileName);
}

/*!
 *
 *  \brief Default destructor
 *
 */

Parser::~Parser(){
    
    PRINT_DEBUG_3("<Parser::~Parser>");
    
    mFile.close();
}

/*!
 *
 *  \brief Reads the .ini file and creates configuration based on it
 *
 */

Configurator* Parser::ReadINI(void){
    
    PRINT_DEBUG_3("<Parser::ReadINI>");
    
    Configurator *tConfigurator = new Configurator;
    
    Parameter  	tParameter;
    string	    tString;
    int		    tLineNumber;
    
    mFile.seekg(0, ios::beg);
    
    tLineNumber = 0;
    while ((!mFile.eof())) {
        tLineNumber++;
        getline(mFile,tString);                        // istream & std::getline(istream, string, char)  // not a member of sring
        
        // remove all white signs
        int tStringLength = tString.length();
        for(int i = 0; i < tStringLength + 1; i++){
            if((tString[i] == '\t') || (tString[i] == ' ') ) {
                tString.erase(i,1);
                tStringLength--;
                i--;
            }
        }
        
        // simple analysis of structure of the imported string
        
        if((tString.empty()) || (tString[0] == '#')) {
            PRINT_DEBUG_2("\tLine "<<tLineNumber<<" Ignoring  : "<<tString);
            continue;
        } else if(string::npos != tString.find('[')) {
            PRINT_DEBUG_2("\tLine "<<tLineNumber<<" Section   : "<<tString);
            continue;
        } else if(!(string::npos != tString.find('='))) {
            PRINT_DEBUG_2("\tLine "<<tLineNumber<<" WARNING   : "<<tString);
            PRINT_DEBUG_2("\tWrong format. Treating as commentary.")
            continue;
        }
        
        // clear the Parameter structure
        
        tParameter.keyword = "";
        tParameter.value = "";
        
        // read the keyword and its value
        
        int i;
        for(i = 0; tString[i] != '='; i++)
            tParameter.keyword += tString[i];
        for(i++; i < tString.length(); i++)
            tParameter.value += tString[i];
        
        // add the parameter to the configurator
        
        tConfigurator->AddParameter(&tParameter);
        
        PRINT_DEBUG_3("\tLine "<<tLineNumber<<" Parameter : "<<tParameter.keyword<<" = "<<tParameter.value);
    }
    return tConfigurator;
}

/*!
 *
 *  \brief Reads eos file and creates tables with thermodynamic functions
 *
 *  \li aVecs = table of vectors contraining thermodynamic functions
 *  \li aN = number of vectors to be read
 *  \li aNames = names of the vectors to be read
 *
 */

Vector1D ** Parser::ReadEOS(const int &aN, const string* aNames){
    
    PRINT_DEBUG_3("<Parser::ReadEOS>");
    
    //    {"#", "eos.tsv - File contains tables of termodynamic functions for realistic equation of state (interpolation between HG and LQGP)."},
    //    {"#", "File structure (T (temperature), E(energy density), P(pressure), S(entropy density) are in fm):"},
    //    {"#", "TPts", "ln(T[0])", "ln(T[TPts-1])"},
    //    {"#", "EPts", "ln(E[0])", "ln(E[EPts-1])"},
    //    {"#", "PPts", "ln(P[0])", "ln(P[PPts-1])"},
    //    {"#", "SPts", "ln(S[0])", "ln(S[SPts-1])"},
    //    {"#", "ln(T)", "ln(E)", "ln(P)", "ln(S)", "ln(Ep)", "ln(Sp)", "   == > TPts records, ln(T) forms a structured grid"},
    //    {"#", "ln(E)", "ln(T)", "ln(P)", "ln(S)", "    == > EPts records, ln(E) forms a structured grid"},
    //    {"#", "ln(P)", "ln(E)", "ln(T)", "ln(S)", "    == > PPts records, ln(P) forms a structured grid"},
    //    {"#", "ln(S)", "ln(E)", "ln(P)", "ln(T)", "    == > SPts records, ln(S) forms a structured grid"}
    
    // ignore description
    
    string line;
    while(1) {
        int pointer = mFile.tellg();
        getline (mFile,line);
        if(delwhite(line)[0] != '#') {
            mFile.seekg(pointer);
            break;
        }
        PRINT_DEBUG_2("Ignoring : " << line);
    }
    
    // read parameters and create vectors
    
    long int XPts; double XMin, XMax;
    
    Vector1D **Vecs = new Vector1D* [aN];
    
    mFile >> XPts >> XMin >> XMax;
    Vecs[0] = new Vector1D(aNames[0], XMin, XMax, XPts);
    Vecs[1] = new Vector1D(aNames[1], XMin, XMax, XPts);
    Vecs[2] = new Vector1D(aNames[2], XMin, XMax, XPts);
    Vecs[12] = new Vector1D(aNames[12], XMin, XMax, XPts);
    Vecs[13] = new Vector1D(aNames[13], XMin, XMax, XPts);
    
    mFile >> XPts >> XMin >> XMax;
    Vecs[3] = new Vector1D(aNames[3], XMin, XMax, XPts);
    Vecs[4] = new Vector1D(aNames[4], XMin, XMax, XPts);
    Vecs[5] = new Vector1D(aNames[5], XMin, XMax, XPts);
    
    mFile >> XPts >> XMin >> XMax;
    Vecs[6] = new Vector1D(aNames[6], XMin, XMax, XPts);
    Vecs[7] = new Vector1D(aNames[7], XMin, XMax, XPts);
    Vecs[8] = new Vector1D(aNames[8], XMin, XMax, XPts);
    
    mFile >> XPts >> XMin >> XMax;
    Vecs[9] = new Vector1D(aNames[9], XMin, XMax, XPts);
    Vecs[10] = new Vector1D(aNames[10], XMin, XMax, XPts);
    Vecs[11] = new Vector1D(aNames[11], XMin, XMax, XPts);
    
    // first column can be ignored since it lies on the structured grid
    
    double tfirstcolumn;
    for(int i = 0; i < Vecs[0]->GetXPts(); i++) mFile >> tfirstcolumn >> (*Vecs[0])(i) >> (*Vecs[1])(i) >> (*Vecs[2])(i) >> (*Vecs[12])(i) >> (*Vecs[13])(i);
    for(int i = 0; i < Vecs[3]->GetXPts(); i++) mFile >> tfirstcolumn >> (*Vecs[3])(i) >> (*Vecs[4])(i) >> (*Vecs[5])(i);
    for(int i = 0; i < Vecs[6]->GetXPts(); i++) mFile >> tfirstcolumn >> (*Vecs[6])(i) >> (*Vecs[7])(i) >> (*Vecs[8])(i);
    for(int i = 0; i < Vecs[11]->GetXPts(); i++) mFile >> tfirstcolumn >> (*Vecs[9])(i) >> (*Vecs[10])(i) >> (*Vecs[11])(i);
    
    return Vecs;
}


/*!
 *
 *  \brief Reads .bin file and returns Vector4D pointer
 *
 */

Vector4D* Parser::ReadBIN(Configurator *aConfigurator, bool PrintInfoQ){
    
    PRINT_DEBUG_3("<Parser::ReadBIN>");
    
    Grid *tGrid   = new Grid(aConfigurator);
    
    const int tXP = tGrid->GetXP(); // x points
    const int tYP = tGrid->GetYP(); // y points
    const int tZP = tGrid->GetZP(); // eta points
    
    struct stat tFileStatistics;
    const int tSizeOfType = sizeof(BINnumber); // c++ number type of data in bin file, has to be the same as hydro output, see global.h
    
    // find out what the number of tau points contains the .bin file
    // number of tau points exported to .bin file may be less than the number of steps to be taken by the solver
    // algorithm set in ini file
    
    stat(mFileName.c_str(), &tFileStatistics);
    
    // total grid points
    // <size of the bin file in Bajts> / <size of BINnumber in Bajts> = <total number of points>
    
    if(PrintInfoQ) PrintRawMessage("File size [MB]", tFileStatistics.st_size / (1024 * 1024));
    double tP = tFileStatistics.st_size / tSizeOfType;
    
    // tP has to be integer number, if not smth wroj
    
    if(tP -  static_cast<int>(tP)){
        PrintRawMessage("File: " + mFileName + "has wrong format");
        PrintRawMessage("      and/or incorrect .bin file numbers type specified.");
        PrintRawMessage("      Its size is not multiple of a size of fundamental type.");
        exit(ProgramStatus(ERROR_FILE_CORRUPTED, "Parser::ReadBIN"));
    }
    
    // calculate number of tau points
    // number of tau points in file can be evaluated in such a way
    
    double tTPts = tP / (tXP * tYP * tZP);
    if(tTPts -  static_cast<int>(tTPts)){
        PrintRawMessage("File: " + mFileName + "has wrong format");
        PrintRawMessage("      and/or incorrect .bin file numbers type specified.");
        PrintRawMessage("      Its size is not multiple of a size of fundamental type.");
        exit(ProgramStatus(ERROR_FILE_CORRUPTED, "Parser::ReadBIN"));
    }else{
        if(PrintInfoQ){
            PrintRawMessage("File content:");
            PrintRawMessage("Tau points", tTPts);
            PrintRawMessage("X points", tXP);
            PrintRawMessage("Y points", tYP);
            PrintRawMessage("Eta points", tZP);
        }
    }
    
    const int tTP =  static_cast<int>(tTPts);
    
    // find tau max and tau min
    // tGrid->GetT(tGrid->GetTP()) can give different value from the last eksported tau point in .bin file
    
    unsigned int tRecordingFrequency = fabs(aConfigurator->GetInt("RecordingFrequency"));
    const double ttmin = tGrid->GetT(0);
    const double ttmax = ttmin + tRecordingFrequency * (tTP - 1) * tGrid->GetdT();
    
    Vector4D* tField = new Vector4D(GetFilename(mFileName),
                                    ttmin, ttmax, tTP,
                                    tGrid->GetX(0), tGrid->GetX(tXP-1), tXP,
                                    tGrid->GetY(0), tGrid->GetY(tYP-1), tYP,
                                    tGrid->GetZ(0), tGrid->GetZ(tZP-1), tZP);
    
    
    // read .bin file into tField
    
    for (int n = 0; n < tTP; ++n)
        for (int i = 0; i < tXP; ++i)
            for (int j = 0; j < tYP; ++j)
                for (int k = 0; k < tZP; ++k){
                    mFile.read(reinterpret_cast<char*>(&((*tField)(n,i,j,k))), tSizeOfType);
                }
    
    return tField;
    
}

void Parser::IgnoreFileHeader(void){
    
    PRINT_DEBUG_3("<Parser::IgnoreFileHeader>");
    
    // ignore description
    
    
    string line;
    while(1) {
        int pointer = mFile.tellg();
        getline (mFile,line);
        if(delwhite(line)[0] != '#') {
            mFile.seekg(pointer);
            break;
        }
        PRINT_DEBUG_2("Ignoring : " << line);
    }
    
}

