/** 
*
*   PerfectFluidEquations.cxx
*	
*   Copyright (C) 2017 Radoslaw Ryblewski and Michael Strickland
*
*   This file is part of aHydro3p1.
*
*   aHydro3p1 is free software: you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation, either version 3 of the License, or
*   (at your option) any later version.
*
*   aHydro3p1 is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with aHydro3p1.  If not, see <http://www.gnu.org/licenses/>.
*    
*/

#include "Global.h"

PerfectFluidEquations::PerfectFluidEquations(Configurator *aConfigurator, Grid *aGrid)
: mNames(0), mGrid(aGrid), mEquationOfState(0),
mXP(aGrid->GetXP()), mYP(aGrid->GetYP()), mZP(aGrid->GetZP()),
mdX(aGrid->GetdX()), mdY(aGrid->GetdY()), mdZ(aGrid->GetdZ()),
m2dX(2 * mdX), m2dY(2 * mdY), m2dZ(2 * mdZ),
mdXi(1 / mdX), mdYi(1 / mdY), mdZi(1 / mdZ),
m2dXi(1 / m2dX), m2dYi(1 / m2dY), m2dZi(1 / m2dZ),
mXPm1(mXP - 1), mYPm1(mYP - 1), mZPm1(mZP - 1)
 {

    PRINT_DEBUG_3("<PerfectFluidEquations::PerfectFluidEquations>");

    mNames = new string [mNumberOfEquations];

    // initialization of names of equations

    mNames[0] = "lnt";
    mNames[1] = "rap";
    mNames[2] = "ux";
    mNames[3] = "uy";

    // zeroing the derivatives

    double DFDz(0),  DFDx(0),  DFDy(0);
    double DRDz(0),  DRDx(0),  DRDy(0);
    double DUxDz(0), DUxDx(0), DUxDy(0);
    double DUyDz(0), DUyDx(0), DUyDy(0);

    double Eta(0);
    double Tau(0);

}

PerfectFluidEquations::~PerfectFluidEquations(){

    PRINT_DEBUG_3("<PerfectFluidEquations::PerfectFluidEquations>");

    if(mNames){
    delete [] mNames;
    }

    mNames = 0;

}

void PerfectFluidEquations::SetEquationOfState(EquationOfState *aEquationOfState){

    PRINT_DEBUG_3("<PerfectFluidEquations::SetEquationOfState>");

    mEquationOfState = aEquationOfState;

}

usi PerfectFluidEquations::GetNumberOfEquations(void){

    PRINT_DEBUG_3("<PerfectFluidEquations::GetNumberOfEquations>");

    return mNumberOfEquations;

}

string PerfectFluidEquations::GetName(usi aNumberOfEquation){

    PRINT_DEBUG_3("<PerfectFluidEquations::GetName>");

    return mNames[aNumberOfEquation];

}

void PerfectFluidEquations::LoadRHS(const double aTau, Vector3D *** const aVArr, const usi aArg, const usi aTo){

    PRINT_DEBUG_3("<PerfectFluidEquations::LoadRHS>");

    register Vector3D * const tVec0A = aVArr[0][aArg];
    register Vector3D * const tVec1A = aVArr[1][aArg];
    register Vector3D * const tVec2A = aVArr[2][aArg];
    register Vector3D * const tVec3A = aVArr[3][aArg];

    register Vector3D * const tVec0 = aVArr[0][aTo];
    register Vector3D * const tVec1 = aVArr[1][aTo];
    register Vector3D * const tVec2 = aVArr[2][aTo];
    register Vector3D * const tVec3 = aVArr[3][aTo];

    for (usi i = 0; i < mXP; i++)
        for (usi j = 0; j < mYP; j++)
            for (usi k = 0; k < mZP; k++){

            SetVariables(aTau,i,j,k,tVec0A,tVec1A,tVec2A,tVec3A);

            (*tVec0)(i,j,k) = GetRHSValue(0);
            (*tVec1)(i,j,k) = GetRHSValue(1);
            (*tVec2)(i,j,k) = GetRHSValue(2);
            (*tVec3)(i,j,k) = GetRHSValue(3);

            CheckValues(tVec0,tVec1,tVec2,tVec3,i,j,k);
            }

}

inline void PerfectFluidEquations::CheckValues( Vector3D * const tVec0,
                                                Vector3D * const tVec1,
                                                Vector3D * const tVec2,
                                                Vector3D * const tVec3, int i, int j, int k){

    PRINT_DEBUG_3("<PerfectFluidEquations::CheckValues>");

        int flag = 0;
		if (isnan((*tVec0)(i,j,k))) {
			PrintRawMessage("Found nan in ", mNames[0]);
			flag=1;
		}
		if (exp((*tVec0)(i,j,k)) <= 0.0) {
			PrintRawMessage("Found negative in ", mNames[0]);
			flag=1;
		}
		if (isnan((*tVec1)(i,j,k))) {
			PrintRawMessage("Found nan in ", mNames[1]);
			flag=1;
		}
		if (isnan((*tVec2)(i,j,k))) {
			PrintRawMessage("Found nan in ", mNames[2]);
			flag=1;
		}
		if (isnan((*tVec3)(i,j,k))) {
			PrintRawMessage("Found nan in ", mNames[3]);
			flag=1;
		}
		if (flag) {
			PrintRawMessage("i = " , i);
			PrintRawMessage("j = " , j);
			PrintRawMessage("k = " , k);
			PrintRawMessage("x = " , mGrid->GetZ(i));
			PrintRawMessage("y = " , mGrid->GetZ(j));
			PrintRawMessage("e = " , mGrid->GetZ(k));
			exit(0);
		}
}

inline double PerfectFluidEquations::Dx(Vector3D * const aVec, const usi i, const usi j, const usi k) {

    PRINT_DEBUG_3("<PerfectFluidEquations::Dx>");

	if (i == 0) return ((*aVec)(1,j,k) - (*aVec)(0,j,k)) * mdXi;                         // right difference
	else if (i == mXPm1) return ((*aVec)(mXPm1,j,k) - (*aVec)(mXP - 2,j,k)) * mdXi;    // left difference
	else return ((*aVec)(i + 1,j,k) - (*aVec)(i - 1,j,k)) * m2dXi;                       // centered difference
}

inline double PerfectFluidEquations::Dy(Vector3D * const aVec, const usi i, const usi j, const usi k) {

    PRINT_DEBUG_3("<PerfectFluidEquations::Dy>");

	if (j == 0) return ((*aVec)(i,1,k) - (*aVec)(i,0,k)) * mdYi;                         // right difference
	else if (j == mYPm1) return ((*aVec)(i,mYPm1,k) - (*aVec)(i,mYP -2,k)) * mdYi;     // left difference
	else return ((*aVec)(i,j + 1,k) - (*aVec)(i,j - 1,k)) * m2dYi;                       // centered difference
}

inline double PerfectFluidEquations::Dz(Vector3D * const aVec, const usi i, const usi j, const usi k) {

    PRINT_DEBUG_3("<PerfectFluidEquations::Dz>");

	if (k == 0) return ((*aVec)(i,j,1) - (*aVec)(i,j,0)) * mdZi;                         // right difference
	else if (k == mZPm1) return ((*aVec)(i,j,mZPm1) - (*aVec)(i,j,mZP - 2)) * mdZi;    // left difference
	else return ((*aVec)(i,j,k + 1) - (*aVec)(i,j,k - 1)) * m2dZi;                       // centered difference
}

void PerfectFluidEquations::SetVariables(const double aTau, const usi i, const usi j, const usi k, Vector3D * const aVec0, Vector3D * const aVec1,Vector3D * const aVec2,Vector3D * const aVec3){

    PRINT_DEBUG_3("<PerfectFluidEquations::SetVariables>");

    Tau = aTau;
    Eta = mGrid->GetZ(k);

    // set derivatives

    DFDx  = Dx(aVec0,i,j,k);
    DFDy  = Dy(aVec0,i,j,k);
    DFDz  = Dz(aVec0,i,j,k);

    DRDx  = Dx(aVec1,i,j,k);
    DRDy  = Dy(aVec1,i,j,k);
    DRDz  = Dz(aVec1,i,j,k);

    DUxDx   = Dx(aVec2,i,j,k);
    DUxDy   = Dy(aVec2,i,j,k);
    DUxDz   = Dz(aVec2,i,j,k);

    DUyDx   = Dx(aVec3,i,j,k);
    DUyDy   = Dy(aVec3,i,j,k);
    DUyDz   = Dz(aVec3,i,j,k);

    // set functions values

    F = (*aVec0)(i,j,k);
    R = (*aVec1)(i,j,k);
    Ux = (*aVec2)(i,j,k);
    Uy = (*aVec3)(i,j,k);

    Y = Eta - R;

    // set other variables

    Ux2 = Ux * Ux;
    Uy2 = Uy * Uy;

    UP = sqrt(Ux2 + Uy2);
    U0 = sqrt(1 + Ux2 + Uy2);

    Ux3 = Ux2 * Ux;
    Ux4 = Ux3 * Ux;

    Uy3 = Uy2 * Uy;
    Uy4 = Uy3 * Uy;

    U02 = U0 * U0;
    UP2 = UP * UP;

    SinhEmR = sinh(Y);
    CoshEmR = cosh(Y);
    SechEmR = 1 / cosh(Y); // Sech(Eta - R)
    TanhEmR = sinh(Y) / cosh(Y);

    Sinh2EmR = sinh(2 * Y);
    Cosh2EmR = cosh(2 * Y);

    CoshEmR2 = pow(CoshEmR,2);
    SinhEmR2 = pow(SinhEmR,2);

    Cs2 = pow(Cs(pow(E,F)),2);
    Cs2m1 = (-1 + Cs2);
    TU0 = Tau*U0;
    TU02 = 2*TU0;
    T2 = 2*Tau;
    Ux2p1 = 1 + Ux2;

}

double PerfectFluidEquations::GetRHSValue(const usi aEquation){

    PRINT_DEBUG_3("<PerfectFluidEquations::GetRHSValue>");

// imported from Mathematica notebook and optimized
// checked against Mathematica hydro output

switch(aEquation){

    case 0: // F

    return (
        (-2*Ux2p1*CoshEmR*(-(DFDx*TU0*Ux) + DFDz*SinhEmR + DFDz*Ux2*SinhEmR) +
             TU02*Uy3*(-(DFDy*CoshEmR*Cs2m1) + DRDy*Cs2*SinhEmR) -
             TU02*Uy*(DFDy*Ux2p1*CoshEmR*Cs2m1 +
                Cs2*((DUxDy + DUyDx)*Ux*CoshEmR - DRDy*SinhEmR - DRDy*Ux2*SinhEmR)) +
             Uy4*(2*DRDz*Cs2 + DFDz*Cs2m1*Sinh2EmR) +
             Cs2*(2*DRDz + DUxDx*TU02*CoshEmR + DUyDy*TU02*Ux2p1*CoshEmR +
                TU02*Ux*(-(DFDx*CoshEmR) + DRDx*SinhEmR) + TU02*Ux3*(-(DFDx*CoshEmR) + DRDx*SinhEmR) + DFDz*Sinh2EmR +
                2*Ux2*(2*DRDz + DFDz*Sinh2EmR) + Ux4*(2*DRDz + DFDz*Sinh2EmR)) +
             2*Uy2*(2*DRDz*Cs2 + DUxDx*TU0*CoshEmR*Cs2 +
                TU0*Ux*(-(DFDx*CoshEmR*Cs2m1) + DRDx*Cs2*SinhEmR) +
                DFDz*Cs2m1*Sinh2EmR + Ux2*
                 (2*DRDz*Cs2 + DFDz*Cs2m1*Sinh2EmR)))/
           (T2*U02*(-(U02*CoshEmR2) + Cs2*(UP2*CoshEmR2 + SinhEmR2)))
                   );

    break;

    case 1: // R

        return (
(-DRDy*T2*pow(Uy,5)*CoshEmR*Cs2m1 -
     2*Uy4*CoshEmR*Cs2m1*(DRDx*Tau*Ux - DRDz*U0*SinhEmR) +
     T2*Uy3*(-(DRDy*CoshEmR*(-2*Ux2p1 + (1 + 2*Ux2)*Cs2)) -
        DFDy*Cs2m1*SinhEmR) - T2*Uy*
      (-(DRDy*pow(1 + Ux2,2)*CoshEmR) + DFDy*Ux2p1*Cs2m1*SinhEmR +
        Ux*Cs2*(DRDy*Ux*Ux2p1*CoshEmR + (DUxDy + DUyDx)*SinhEmR)) +
     Cs2*(-2*DFDz*U0*Ux2 - DRDx*T2*Ux3*Ux2p1*CoshEmR +
        T2*(DUxDx + (DUyDy - DFDx*Ux)*Ux2p1)*SinhEmR + DRDz*U0*Sinh2EmR + 2*DRDz*U0*Ux2*Sinh2EmR +
        DRDz*U0*Ux4*Sinh2EmR) + Ux2p1*
      (DRDx*T2*(Ux + Ux3)*CoshEmR + T2*DFDx*Ux*SinhEmR - DRDz*U0*Ux2*Sinh2EmR + U0*(2*DFDz - DRDz*Sinh2EmR)) +
     2*Uy2*(-DRDx*T2*Ux3*CoshEmR*Cs2m1 +
        Tau*Ux*(-(DRDx*CoshEmR*(-2 + Cs2)) - DFDx*Cs2m1*SinhEmR) +
        DRDz*U0*Ux2*Cs2m1*Sinh2EmR + U0*(DFDz - DRDz*Sinh2EmR) +
        Cs2*(DUxDx*Tau*SinhEmR + U0*(-DFDz + DRDz*Sinh2EmR))))/
   (T2*pow(U02,1.5)*(-(U02*CoshEmR2) + Cs2*(UP2*CoshEmR2 + SinhEmR2)))
           );

    break;

    case 2: // Ux

    return (
(2*U02*CoshEmR*(Tau*(DFDx + DUxDx*Ux + DUxDy*Uy) - DUxDz*U0*SinhEmR) +
     Cs2*(-2*DRDx*Tau*Ux4*SinhEmR - 2*Ux3*(DRDz*U0 + Tau*((DUxDx + DUyDy)*CoshEmR + DRDy*Uy*SinhEmR)) -
        2*(DFDx*Tau + DUxDy*Tau*Uy - DUxDz*U0*SinhEmR)*(Uy2*CoshEmR + SinhEmR*TanhEmR) -
        2*Ux*(DRDz*U0*Uy2 + DUxDx*Tau*(2*Uy2*CoshEmR + Cosh2EmR*SechEmR) +
           Tau*(DUyDy*CoshEmR - DFDy*Uy*SechEmR + DRDy*(Uy + Uy3)*SinhEmR) + U0*(DRDz + DFDz*TanhEmR)) +
        Ux2*(DUxDz*U0*Sinh2EmR + T2*(DUyDx*Uy*CoshEmR - SinhEmR*(DRDx*(1 + Uy2) + DFDx*TanhEmR)))))/
   (T2*sqrt(U02)*(-(U02*CoshEmR2) + Cs2*(UP2*CoshEmR2 + SinhEmR2)))
                   );

    break;

    case 3: // Uy

    return (
(2*U02*CoshEmR*(DUyDx*Tau*Ux + DUyDy*Tau*Uy - DUyDz*U0*SinhEmR) +
     DFDy*T2*(U02*CoshEmR - Cs2*(Ux2*CoshEmR + (1 + Uy2)*SinhEmR*TanhEmR)) -
     Cs2*(2*DRDz*U0*Uy + 2*DRDz*U0*Ux2*Uy + 2*DRDz*U0*Uy3 + DRDx*T2*U02*Ux*Uy*SinhEmR -
        DUyDz*U0*Ux2*Sinh2EmR - DUyDz*U0*Uy2*Sinh2EmR + 2*DFDz*U0*Uy*TanhEmR -
        2*DUyDz*U0*SinhEmR2*TanhEmR + T2*(DUxDx*(Uy + Uy3)*CoshEmR +
           Uy*(-(Ux*(DUxDy*Uy*CoshEmR + DFDx*SechEmR)) + DUyDy*(Uy2*CoshEmR + Cosh2EmR*SechEmR) +
              DRDy*(Uy + Uy3)*SinhEmR + Ux2*(2*DUyDy*CoshEmR + DRDy*Uy*SinhEmR)) +
           DUyDx*Ux*(Ux2*CoshEmR + SinhEmR*TanhEmR))))/
   (T2*sqrt(U02)*(-(U02*CoshEmR2) + Cs2*(UP2*CoshEmR2 + SinhEmR2)))
                   );

    break;

    default:

    return -666;
}
}

inline double PerfectFluidEquations::Cs(const double aX){

    PRINT_DEBUG_3("<PerfectFluidEquations::Cs>");

    return sqrt(mEquationOfState->ST(aX) / mEquationOfState->EpT(aX));

}

inline void PerfectFluidEquations::PrintInfo(void){

    PRINT_DEBUG_3("<PerfectFluidEquations::PrintInfo>");

    PrintRawMessage("-> Perfect Fluid");

}

double PerfectFluidEquations::FormatOutput(double aX, int aN){

    PRINT_DEBUG_3("<PerfectFluidEquations::FormatOutput>");

    switch(aN){

        case 0: return UnitsFactor * exp(aX);

        default: return aX;
    }
}

double PerfectFluidEquations::FormatInput(double aX, int aN){

    PRINT_DEBUG_3("<PerfectFluidEquations::FormatOutput>");

    switch(aN){

        case 0: return log(aX);

        default: return aX;
    }
}
