/** 
*
*   RealisticEquationOfState.cxx
*	
*   Copyright (C) 2017 Radoslaw Ryblewski and Michael Strickland
*
*   This file is part of aHydro3p1.
*
*   aHydro3p1 is free software: you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation, either version 3 of the License, or
*   (at your option) any later version.
*
*   aHydro3p1 is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with aHydro3p1.  If not, see <http://www.gnu.org/licenses/>.
*    
*/

#include "Global.h"

/*!
*
*  \brief Default constructor
*
*  \li aDirectory = sets the location of realistic EOS *.tsv files w.r.t. the *.exe file (example: ./input/eos/eos_directory_name/)
*
*/

RealisticEquationOfState::RealisticEquationOfState(string aDirectory)
: mN(14){

    PRINT_DEBUG_3("<RealisticEquationOfState::RealisticEquationOfState>");

    mNames = new string [mN];

    // read eos.tsv file
    // in the code only arrays No 0,1,2,3,7,11,12,13 are used
    // other are generated for completeness

    mNames[0] = "ET"; mNames[1] = "PT"; mNames[2] = "ST"; mNames[12] = "EpT"; mNames[13] = "SpT";
    mNames[3] = "TE"; mNames[4] = "PE"; mNames[5] = "SE";
    mNames[6] = "EP"; mNames[7] = "TP"; mNames[8] = "SP";
    mNames[9] = "ES"; mNames[10] = "PS"; mNames[11] = "TS";

    Parser *tParser = new Parser(aDirectory + "eos.tsv");

    mVecs = tParser->ReadEOS(mN, mNames);

    // optimization, minimal and maximal values in fm

    Tmin = exp(mVecs[0]->GetXMin());
    Tmax = exp(mVecs[0]->GetXMax());
    Emin = exp(mVecs[3]->GetXMin());
    Emax = exp(mVecs[3]->GetXMax());
    Pmin = exp(mVecs[7]->GetXMin());
    Pmax = exp(mVecs[7]->GetXMax());
    Smin = exp(mVecs[11]->GetXMin());
    Smax = exp(mVecs[11]->GetXMax());
    Epmin = exp(mVecs[12]->Interpolate(log(Tmin)));
    Epmax = exp(mVecs[12]->Interpolate(log(Tmax)));
    Spmin = exp(mVecs[13]->Interpolate(log(Tmin)));
    Spmax = exp(mVecs[13]->Interpolate(log(Tmax)));
    Csmin = sqrt((PT(Tmin * 1.001) - PT(Tmin * 0.999)) / (ET(Tmin * 1.001) - ET(Tmin * 0.999)));
    Csmax = sqrt((PT(Tmax * 1.001) - PT(Tmax * 0.999)) / (ET(Tmax * 1.001) - ET(Tmax * 0.999)));

/// block generating check.tsv file, can be read by ./addons/eos/eos.nb

    // uncomment to check functions of temperature

//    string name = aDirectory + "check.tsv";
//    ofstream o(name.c_str());
//    int arrNo = 1;
//    double Xmin = mVecs[arrNo]->GetXMin();
//    double Xmax = mVecs[arrNo]->GetXMax();
//    double XN = mVecs[arrNo]->GetXPts();
//    double dX = (Xmax - Xmin) / (XN - 1);
// //   for(int i = -500; i < XN+1000; ++i) o << Xmin + dX * i << "\t" << log(ST(exp(Xmin + dX * i))) << endl; // <== for other functions of T
//    for(int i = -500; i < XN+1000; ++i) o << exp(Xmin + dX * i) * hbarc << "\t" << CsT(exp(Xmin + dX * i)) << endl; // <== for CsT
//    o.close();

    // uncomment to check functions of energy density, entropy density, pressure

//    string name = aDirectory + "check.tsv";
//    ofstream o(name.c_str());
//    int arrNo = 11;
//    double Xmin = mVecs[arrNo]->GetXMin();
//    double Xmax = mVecs[arrNo]->GetXMax();
//    double XN = mVecs[arrNo]->GetXPts();
//    double dX = (Xmax - Xmin) / (XN - 1);
//    for(int i = -500; i < XN+1000; ++i) o << Xmin + dX * i << "\t" << log(TS(exp(Xmin + dX * i))) << endl;
//    o.close();

/// block end
}

/*!
*
*  \brief Destructor
*
*/

RealisticEquationOfState::~RealisticEquationOfState(){

    PRINT_DEBUG_3("<RealisticEquationOfState::~RealisticEquationOfState>");

    if(mVecs) for(int i = 0; i < mN; ++i) delete mVecs[i];
    delete[] mVecs;
}

/*!
*
*  \brief Returns energy density in fm^-4
*
*  \li x = temperature in fm^-1
*
*/

double RealisticEquationOfState::ET(double x){

    PRINT_DEBUG_3("<RealisticEquationOfState::ET>");

    if (x < Tmin){
        return Emin;
    }else if(x > Tmax){
        return Emax;
    }else{
        return exp(mVecs[0]->Interpolate(log(x)));
    }
}

/*!
*
*  \brief Returns pressure in fm^-4
*
*  \li x = temperature in fm^-1
*
*/

double RealisticEquationOfState::PT(double x){

    PRINT_DEBUG_3("<RealisticEquationOfState::PT>");

    if (x < Tmin){
        return Pmin;
    }else if(x > Tmax){
        return Pmax;
    }else{
        return exp(mVecs[1]->Interpolate(log(x)));
    }
}

/*!
*
*  \brief Returns entropy density at given temperature in fm^-3
*
*  \li x = temperature in fm^-1
*
*/


double RealisticEquationOfState::ST(double x){ // S(T) = dP/dT

    PRINT_DEBUG_3("<RealisticEquationOfState::ST>");

    if (x < Tmin){
        return Smin;
    }else if(x > Tmax){
        return Smax;
    }else{
        return exp(mVecs[2]->Interpolate(log(x)));
    }

}

/*!
*
*  \brief Returns temperature at given energy density in fm^-1
*
*  \li x = energy density in fm^-4
*
*/

double RealisticEquationOfState::TE(double x){

    PRINT_DEBUG_3("<RealisticEquationOfState::TE>");

    if (x < Emin){
        return Tmin;
    }else if(x > Emax){
        return Tmax;
    }else{
        return exp(mVecs[3]->Interpolate(log(x)));
    }

}

/*!
*
*  \brief Returns temperature at given pressure in fm^-1
*
*  \li x = pressure in fm^-4
*
*/

double RealisticEquationOfState::TP(double x){

    PRINT_DEBUG_3("<RealisticEquationOfState::PressureToTemperature>");

    if (x < Pmin){
        return Tmin;
    }else if(x > Pmax){
        return Tmax;
    }else{
        return exp(mVecs[7]->Interpolate(log(x)));
    }

}

/*!
*
*  \brief Returns temperature at given entropy density in fm^-1
*
*  \li x = entropy density in fm^-3
*
*/

double RealisticEquationOfState::TS(double x){

    PRINT_DEBUG_3("<RealisticEquationOfState::TS>");

    if (x < Smin){
        return Tmin;
    }else if(x > Smax){
        return Tmax;
    }else{
        return exp(mVecs[11]->Interpolate(log(x)));
    }

}

/*!
*
*  \brief Returns speed of sound at given temperature
*
*  \li x = temperature in fm^-1
*
*/

double RealisticEquationOfState::CsT(double x){

    PRINT_DEBUG_3("<RealisticEquationOfState::CsT>");

    if (x < Tmin * 1.001){
        return Csmin;
    }else if(x > Tmax * 0.999){
        return Csmax;
    }else{
        return sqrt((PT(x * 1.001) - PT(x * 0.999)) / (ET(x * 1.001) - ET(x * 0.999)));
    }

}

/*!
*
*  \brief Returns E'(T) = dE/dT in fm^-3
*
*  \li x = temperature in fm^-1
*
*/

double RealisticEquationOfState::EpT(double x){

    PRINT_DEBUG_3("<RealisticEquationOfState::EpT>");

    if (x < Tmin){
        return Epmin;
    }else if(x > Tmax){
        return Epmax;
    }else{
        return exp(mVecs[12]->Interpolate(log(x)));
    }

}
/*!
*
*  \brief Returns S'(T) = dS/dT in fm^-2
*
*  \li x = temperature in fm^-1
*
*/

double RealisticEquationOfState::SpT(double x){

    PRINT_DEBUG_3("<RealisticEquationOfState::SpT>");

    if (x < Tmin){
        return Spmin;
    }else if(x > Tmax){
        return Spmax;
    }else{
        return exp(mVecs[13]->Interpolate(log(x)));
    }

}

/*!
*
*  \brief Prints Info
*
*/

inline void RealisticEquationOfState::PrintInfo(void){

    PrintRawMessage("-> Realistic (Interpolated HG and Lattice QGP)");
    PrintRawMessage("\tTemp. range [GeV]", "(" + d2s(Tmin * hbarc) + "," + d2s(Tmax * hbarc) + ")");
    PrintRawMessage("\tPoints", d2s(mVecs[0]->GetXPts()));
    PrintRawMessage("\tEnergy range [GeV^4]", "(" + d2s(Emin * hbarc4) + "," + d2s(Emax * hbarc4) + ")");
    PrintRawMessage("\tPoints", d2s(mVecs[3]->GetXPts()));
    PrintRawMessage("\tPressure range [GeV^4]", "(" + d2s(Pmin * hbarc4) + "," + d2s(Pmax * hbarc4) + ")");
    PrintRawMessage("\tPoints", d2s(mVecs[6]->GetXPts()));
    PrintRawMessage("\tEntropy range [GeV^3]", "(" + d2s(Smin * hbarc3) + "," + d2s(Smax * hbarc3) + ")");
    PrintRawMessage("\tPoints", d2s(mVecs[9]->GetXPts()));
    PrintRawMessage("\tFor arguments out-of-range the min or max values used.");
}
