
/**
 *
 *   teq.cxx
 *
 *   Copyright (C) 2016 Mubarak Alqahtani, and Michael Strickland
 *
 *   This file is part of aHydro3p1.
 *
 *   aHydro3p1 is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   aHydro3p1 is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with aHydro3p1.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "Global.h"
#include "RealisticEquationOfState.h"
#include "ThermalMass.h"

/*!
 *
 *  \brief Default constructor
 *
 */

RelaxationTime::RelaxationTime() {
    PRINT_DEBUG_3("<RelaxationTime:RelaxationTime>");
}


/*!
 *
 *  \brief Destructor
 *
 */

RelaxationTime::~RelaxationTime(){
    PRINT_DEBUG_3("<RelaxationTime::~RelaxationTime>");
}


//************************************************************************
// the integration
//************************************************************************
double RelaxationTime::integrateteq(functionWrapperteq * fw, double x) {
    
    gsl_integration_workspace *w = gsl_integration_workspace_alloc(NPTS);
    double result, error;
    struct integralParamsteq params = { x };
    
    gsl_function F;
    F.function = fw->function;
    F.params = &params;
    
    gsl_integration_qags(&F, 0, 1, 0, INT_TOLERANCE, NPTS, w, &result, &error);
    
    gsl_integration_workspace_free(w);
    
    return result;
}

//************************************************************************
// modified Struve functions integrands
//************************************************************************
double RelaxationTime::L_0_integrand(double t, void *params) {
    
    struct integralParamsteq *my_params = (struct integralParamsteq *) params;
    double x = my_params->x;
    
    return pow(1-t*t,-0.5)*sinh(x*t);
}

double RelaxationTime::L_1_integrand(double t, void *params) {
    struct integralParamsteq *my_params = (struct integralParamsteq *) params;
    double x = my_params->x;

    
    return pow(1-t*t,0.5)*sinh(x*t);   
}

//************************************************************************
// modified Struve functions
//************************************************************************
double RelaxationTime::L_0(double x) {
    functionWrapperteq fw;
    fw.function = &L_0_integrand;;
    return (2./M_PI)*integrateteq(&fw,x);
}

double RelaxationTime::L_1(double x) {
    functionWrapperteq fw;
    fw.function = &L_1_integrand;
    return (2./M_PI)*x*integrateteq(&fw,x);
}

double RelaxationTime::L_m1(double x) {
   
    return L_1(x)+2./M_PI;
}

//************************************************************************
// kappa function used for teq
//************************************************************************


double RelaxationTime::kappa(double x){
    if (x>10 && x<=16)
	return 23.917820576435723 + x*(-6.587047003709422 + x*(1.0389211240054883 + 
         x*(0.269559582221178 + x*(-0.1519442735330117 + x*(0.03326644324779483 + 
                  x*(-0.004451405752073866 + x*(0.0004018724287337314 + 
                        x*(-0.000025165573393435347 + x*(1.08470988422957e-6 + 
                              x*(-3.0850744201559074e-8 + (5.230643257996886e-10 - 4.014572564960685e-12*x)*x))))))))));
    if (x>16)
	return (-4.761641793638534e16 + x*(2.4561014396426644e16 + x*(-1.3359046811426145e15 + 
           x*(7.687442840133364e13 + x*(-4.696307920310203e12 + 
                 x*(3.0576592145928955e11 + x*(-2.1311387582807064e10 + 
                       x*(1.5982994553222654e9 + x*(-1.2976088448257443e8 + 
                             x*(1.1486306250000002e7 + x*(-1.1182765869140625e6 + 
                                   x*(121057.03125 + x*(-14781.4453125 + x*(2075.625 + x*(-343.125 + x*(67.5 + x*(-15. + 15.*x)))))))))))))))))/Power(x,17);
    else
	return Power(x,3)*(-(1/x) + K(1,x)/K(2,x) + (3*K(3,x))/(Power(x,2)*K(2,x)) - (M_PI*(1 - x*K(0,x)*L_m1(x) - x*K(1,x)*L_0(x)))/(2.*K(2,x)));
}


void RelaxationTime::setup(Hfuncs aHfuncs, ThermalMass aMass, LinearTransportCoefficient aTransportCoeff) {
    mass = aMass;
    etabar = aTransportCoeff;
    hfuncs = aHfuncs;
    RelaxationTime::loadInterpolation();
}

//************************************************************************
//  the realistic teq when m=m(T)
//************************************************************************

// see arXiv:1509.02913 for more information
// We rewrite teq in terms of kappa, to use the expansion defined above! But, it is the same as the one listed in arXiv:1612.07329 
double RelaxationTime::teqEval(double aT){
    double m = mass.mass(aT);
    double e = hfuncs.H3(1,1,1,m/aT)*aT*aT*aT*aT + mass.Beq(aT);
    double p = hfuncs.H3X(1,1,1,m/aT)*aT*aT*aT*aT - mass.Beq(aT);
    return (15.*etabar.Value(aT))*(e+p)/(4*M_PI*Nt*aT*aT*aT*m*m*K(2,m/aT)*kappa(m/aT));
    //return (15.*eb)*(1+e/p)/(kappa(m/aT)*aT);

}

double RelaxationTime::teq(double T){
    double LT = log(T);
    if (LT>=LTMIN && LT<=LTMAX) { return exp(gsl_spline_eval(mSpline, LT, mAcc)); }
    return RelaxationTime::teqEval(T);
}

void RelaxationTime::loadInterpolation() {
    
    double T;
    double dLT = (LTMAX-LTMIN)/(LTPTS-1.);
    
    for (int i=0;i<LTPTS;i++) {
        LT[i] = LTMIN + i*dLT;
        T = exp(LT[i]);
        LTEQ[i] = log(teqEval(T));
    }
    
   mAcc = gsl_interp_accel_alloc();
   mSpline = gsl_spline_alloc(gsl_interp_cspline, LTPTS);
   gsl_spline_init(mSpline, LT, LTEQ, LTPTS);
    
}

void RelaxationTime::freeInterpolation() {
    gsl_spline_free(mSpline);
    gsl_interp_accel_free(mAcc);
}


