/** 
*
*   RungeKuttaStepper.cxx
*	
*   Copyright (C) 2017 Radoslaw Ryblewski and Michael Strickland
*
*   This file is part of aHydro3p1.
*
*   aHydro3p1 is free software: you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation, either version 3 of the License, or
*   (at your option) any later version.
*
*   aHydro3p1 is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with aHydro3p1.  If not, see <http://www.gnu.org/licenses/>.
*    
*/

#include "Global.h"

/*!
 *
 *  \brief Default constructor
 *
 *  \li *aConfigurator = pointer to the configuration
 *  \li *aGrid = pointer to the grid
 *
 */

RungeKuttaStepper::RungeKuttaStepper(Configurator *aConfigurator, Grid *aGrid)
: mVArr(0), mEquations(0), mGrid(aGrid),
mXP(aGrid->GetXP()), mYP(aGrid->GetYP()), mZP(aGrid->GetZP()), mdT(aGrid->GetdT()){

    PRINT_DEBUG_3("RungeKuttaStepper::RungeKuttaStepper");

}

/*!
 *
 *  \brief Destructor
 *
 */

RungeKuttaStepper::~RungeKuttaStepper(){

    PRINT_DEBUG_3("RungeKuttaStepper::~RungeKuttaStepper");

}

/*!
 *
 *  \brief Allocates memory
 *
 */

void    RungeKuttaStepper::AllocateMemory(Equations *aEquations){

    PRINT_DEBUG_3("RungeKuttaStepper::AllocateMemory");

    // initilize with zeros

    try{
    mEquations = aEquations;
    mVArr = new Vector3D** [mEquations->GetNumberOfEquations()];
        for (int i = 0; i < aEquations->GetNumberOfEquations(); i++){
            mVArr[i] = new Vector3D*[mNumberOfVectors];
            for (int j = 0; j < mNumberOfVectors; j++){
            mVArr[i][j] = new Vector3D(
                mEquations->GetName(i),
                mGrid->GetXMin(),mGrid->GetXMax(),mGrid->GetXP(),
                mGrid->GetYMin(),mGrid->GetYMax(),mGrid->GetYP(),
                mGrid->GetZMin(),mGrid->GetZMax(),mGrid->GetZP()
                );
            }
        }
    } catch (std::bad_alloc) {

        exit(ProgramStatus(ERROR_MEMORY_BAD_ALLOCATION, "RungeKuttaStepper::AllocateMemory"));

    }
}

/*!
 *
 *  \brief Initializes vectors
 *
 */

void    RungeKuttaStepper::Initialize(Initializer **aIArr){

    PRINT_DEBUG_3("RungeKuttaStepper::Initialize");

    // initilize with Initializers

        PrintRawMessage("");
        for (int n = 0; n < mEquations->GetNumberOfEquations(); n++){
                Vector3D* tVec = mVArr[n][0];
                PrintRawMessage("Initializing", mEquations->GetName(n));
                for (usi i = 0; i < mXP; i++)
                    for (usi j = 0; j < mYP; j++)
                        for (usi k = 0; k < mZP; k++)
                            (*tVec)(i,j,k)  = mEquations->FormatInput(aIArr[n]->GetValue(i,j,k), n);
        }
}

/*!
 *
 *  \brief Frees memory that is used by the class
 *
 */

void RungeKuttaStepper::FreeMemory(){

    PRINT_DEBUG_3("RungeKuttaStepper::FreeMemory");

    if(mVArr) {
        for (int i = 0; i < mEquations->GetNumberOfEquations(); i++) {
            for (int j = 0; j < mNumberOfVectors; j++){
                delete mVArr[i][j];
            }
            delete[] mVArr[i];
        }
        delete[] mVArr;
    } else{
    PrintMessage("Nothing to Do");
    }
    mVArr = 0;

}

/*!
 *
 *  \brief Return arrays used by the stepper
 *
 */

inline Vector3D *** RungeKuttaStepper::GetArrays(void){

    PRINT_DEBUG_3("RungeKuttaStepper::GetArrays");

    return mVArr;

}

/*!
 *
 *  \brief Make a full Runge-Kutta step
 *
 *  \li aTau = proper time at which to evaluate the RHS of the evolution equations (positive definite real number)
 *
 */

void    RungeKuttaStepper::MakeTimeStep(const double aTau){

    PRINT_DEBUG_3("RungeKuttaStepper::MakeTimeStep");

    // copy data from arrays No.0 to arrays No.1

    CopyArrays(0,1);

    // make one time step dT by doing 4 steps of RK algorithm

    MakeStep(aTau,             0,6,2,1);

    MakeStep(aTau + mdT / 2.0, 3,3,2,1);

    MakeStep(aTau + mdT / 2.0, 3,3,1,1);

    MakeStep(aTau + mdT,       3,6,1,0);

    // swap pointers of arrays No.0 with arrays No.1

    SwapArrays(0,1);
    
}

/*!
 *
 *  \brief Copy arrays
 *
 *  \li aFrom = index of array to copy from (integer between 0 and 3)
 *  \li aTo = index of array to copy to (integer between 0 and 3)
 *
 */

inline void RungeKuttaStepper::CopyArrays(const usi aFrom, const usi aTo){

    PRINT_DEBUG_3("RungeKuttaStepper::CopyArrays");

    const usi mNEq = mEquations->GetNumberOfEquations();
    Vector3D *tVecT(0), *tVecF(0);

    for(usi n = 0; n < mNEq; n++){
    tVecF = mVArr[n][aFrom];
    tVecT = mVArr[n][aTo];
    for (usi i = 0; i < mXP; i++)
        for (usi j = 0; j < mYP; j++)
            for (usi k = 0; k < mZP; k++){

                (*tVecT)(i,j,k) = (*tVecF)(i,j,k);

            }
    }
}

/*!
 *
 *  \brief Swap arrays using a pointer swap
 *
 *  \li aArrA = index of array A (integer between 0 and 3)
 *  \li aArrB = index of array B (integer between 0 and 3)
 *
 */

inline void RungeKuttaStepper::SwapArrays(const usi aArrA, const usi aArrB){

    PRINT_DEBUG_3("RungeKuttaStepper::SwapArrays");

    Vector3D *tVecTemp = 0;
    usi n;
    const usi mNEq = mEquations->GetNumberOfEquations();

    for(n = 0; n < mNEq; n++){

        tVecTemp = mVArr[n][aArrA];
        mVArr[n][aArrA] = mVArr[n][aArrB];
        mVArr[n][aArrB] = tVecTemp;

    }
}

/*!
 *
 *  \brief Make an intermediate Runge-Kutta step
 *
 *  \li aTau = proper time at which to evaluate the RHS of the evolution equations (positive definite real number)
 *  \li aArg = index of the mVArr to use to load the RHS of the evolution equations (integer between 0 and 3)
 *  \li aC1 = scaling factor for first part of RK step (postive definite integer)
 *  \li aC2 = scaling factor for second part of RK step (postive definite integer)
 *  \li aLoad = boolean (0 or 1) which determines which array to load with the result; intermediate steps should use 1
 *          and the final step should use 0
 *
*/

void RungeKuttaStepper::MakeStep(const double aTau, const usi aArg, const usi aC1, const usi aC2, const bool aLoad) {

    PRINT_DEBUG_3("<RungeKuttaStepper::MakeStep>");

	mEquations->LoadRHS(aTau, mVArr, aArg);

    const usi mNEq = mEquations->GetNumberOfEquations();
    Vector3D *tVec0(0), *tVec1(0), *tVec2(0), *tVec3(0);

    for(usi n = 0; n < mNEq; n++){

        switch(static_cast<int>(aLoad)){

            case 1:

                tVec0 = mVArr[n][0];
                tVec1 = mVArr[n][1];
                tVec2 = mVArr[n][2];
                tVec3 = mVArr[n][3];

                for (usi i = 0; i < mXP; i++)
                    for (usi j = 0; j < mYP; j++)
                        for (usi k = 0; k < mZP; k++){

                            (*tVec1)(i,j,k) += mdT * (*tVec2)(i,j,k) / aC1;
                            (*tVec3)(i,j,k) = (*tVec0)(i,j,k) + mdT * (*tVec2)(i,j,k) / aC2;

                        }
                break;

            case 0:

                tVec1 = mVArr[n][1];
                tVec2 = mVArr[n][2];

                for (usi i = 0; i < mXP; i++)
                    for (usi j = 0; j < mYP; j++)
                        for (usi k = 0; k < mZP; k++){

                            (*tVec1)(i,j,k) += mdT * (*tVec2)(i,j,k) / aC1;

                        }
                break;
        }
    }
}

/*!
 *
 *  \brief Returns the number of vector arrays that will be used for the evolution
 *
 */

usi RungeKuttaStepper::GetNumberOfVectors(void){

    PRINT_DEBUG_3("<RungeKuttaStepper::GetNumberOfVectors>");

    return mNumberOfVectors;

}

/*!
 *
 *  \brief Prints Info
 *
 */

void RungeKuttaStepper::PrintInfo(void){

    PRINT_DEBUG_3("<EulerStepper::PrintInfo>");

    PrintRawMessage("-> 4th order Runge-Kutta");

}
