/**
 *
 *   Solver.cxx
 *
 *   Copyright (C) 2017 Radoslaw Ryblewski and Michael Strickland
 *
 *   This file is part of aHydro3p1.
 *
 *   aHydro3p1 is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   aHydro3p1 is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with aHydro3p1.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "Global.h"

Solver::Solver(Configurator *aConfigurator)
: mConfigurator(aConfigurator), mEquationOfState(0), mEquations(0), mStepper(0), mGrid(0), mIArr(0), istest(false){

    PRINT_DEBUG_3("<Solver::Solver>");

    // initialize grid

    mGrid   = new Grid(mConfigurator);

    // print grid info

    PrintMessage("GRID");
    mGrid->PrintInfo();

    // parameters check

    CheckParameters();

    // set options for solver

    SetParameters();

    // set random seed based on current time
    
    mSeed = time(0);
    
    mTerminateQ = false;

}

Solver::~Solver(){

    PRINT_DEBUG_3("<Solver::~Solver>");

    // clean memory

    if(mEquationOfState){
        delete mEquationOfState;
        mEquationOfState = 0;
    }
    if(mEquations){
        delete mEquations;
        mEquations = 0;
    }
    if(mStepper){
        delete mStepper;
        mStepper = 0;
    }
    if(mSmoother){
        delete mSmoother;
        mSmoother = 0;
    }
    if(mGrid){
        delete mGrid;
        mGrid = 0;
    }
    if(mIArr){
        delete [] mIArr;
        mIArr = 0;
    }
    if(tDate){
        delete [] tDate;
        tDate = 0;
    }

    // configurator deleted inside main function

    mConfigurator       = 0;

}

void    Solver::SetParameters(void){

    PRINT_DEBUG_3("<Solver::SetParameters>");

    // create hashed directory name for output data
    // used localtime to convert time_t to struct tm
    // used asctime to convert tm structure to string

    PrintMessage("DIRECTORIES");

    mHydroID = mConfigurator->GetParameter("JobID");
    if (mHydroID.compare("Random")==0) {
        time_t tTimeHash = time(0);
        srand((unsigned)tTimeHash);
        int rint = rand(); // random integer
        ostringstream temp;
        temp<<rint;
        string tHashParam = temp.str();
        Crc32 tHash(tHashParam.c_str(),tHashParam.length());
        const string boo = tHash.GetValueHex();
        mHydroID = boo;
    }


    // check which file types will be used
    // set the directories names
    // for each file type create directory if it does not exist yet

    // set recording frequency

    mRecordingFrequency = mConfigurator->GetInt("RecordingFrequency");
    mFileTSV    = static_cast<bool>(mConfigurator->GetInt("FileTSV"));
    mFileBIN    = static_cast<bool>(mConfigurator->GetInt("FileBIN"));

    // check if it is a test run
    // if the mRecordingFrequency is < 0 then output is exported to "test" directory

    if(mRecordingFrequency < 0) {
        istest = true;
        mRecordingFrequency = -mRecordingFrequency;
        mHydroID = "test";
    }

    if(mRecordingFrequency == 0 || (mFileTSV == 0 && mFileBIN == 0)){
        PrintRawMessage("Caution! Measurables will NOT BE RECORDED!");
        PrintRawMessage("         To overrride set: RecordingFrequency != 0");
        PrintRawMessage("         and (FileTSV != 0 or FileBIN != 0)");
        PrintRawMessage("         No directories will be created.");
        mFileTSV    = 0;
        mFileBIN    = 0;
        mRecordingFrequency = 0;
    }

    if(mFileTSV){

        mTSVDir = mConfigurator->GetParameter("TSVDir");
        MakeDirectory(mTSVDir);
        MakeDirectory(mTSVDir + mHydroID + "/");
        if(istest) ClearDirectory(mTSVDir + mHydroID + "/");
        //CopyFile(sConfigurationFileDirectory + sGeneralConfigurationFileName, mTSVDir + mHydroID + "/" + mHydroID + ".ini");
        ofstream outFile;
        outFile.open((mTSVDir + mHydroID + "/" + mHydroID + ".ini").c_str(),ios::out);
        mConfigurator->OutputParametersToFile(outFile);
        outFile.close();    }

    if(mFileBIN){

        mBINDir = mConfigurator->GetParameter("BINDir");
        MakeDirectory(mBINDir);
        MakeDirectory(mBINDir + mHydroID + "/");
        if(istest) ClearDirectory(mBINDir + mHydroID + "/");
        //CopyFile(sConfigurationFileDirectory + sGeneralConfigurationFileName, mBINDir + mHydroID + "/" + mHydroID + ".ini");
        ofstream outFile;
        outFile.open((mBINDir + mHydroID + "/" + mHydroID + ".ini").c_str(),ios::out);
        mConfigurator->OutputParametersToFile(outFile);
        outFile.close();    }

}

string    Solver::SolveEquations(){

    PRINT_DEBUG_3("<Solver::SolveEquations>");

    // choose EquationOfState, hydrodynamic Equations type and Stepper type

    PrintMessage("EQUATION OF STATE");
    mEquationOfState    = GetEquationOfState();

    PrintMessage("HYDRODYNAMIC EQUATIONS");
    mEquations          = GetEquations();

    // set equation of state

    mEquations->SetEquationOfState(mEquationOfState);

    // set the stepper

    PrintMessage("STEPPER");
    mStepper            = GetStepper();

    // set the initial conditions

    PrintMessage("INITIAL CONDITIONS");
    mIArr               = GetInitializerArr();

    // memory allocation for stepper

    mStepper->AllocateMemory(mEquations);

    // initialization of arrays for stepper

    mStepper->Initialize(mIArr);

    // set the smoother

    PrintMessage("SMOOTHER");
    mSmoother            = GetSmoother();

    // print amount of RAM memory used by arrays which will be used during calculation

    PrintMessage("MEMORY");
    PrintRawMessage("Memory used (RAM)", d2s((mStepper->GetNumberOfVectors() * mEquations->GetNumberOfEquations() * mGrid->GetXP() * mGrid->GetYP() * mGrid->GetZP() * sizeof(double)) / (1024 * 1024)) + " MB.");

    // set initial time of evolution

    tTau = mGrid->GetTMin();

    // initialization of memory usage container

    tMem.tMemoryUsedBIN = 0;
    tMem.tMemoryUsedTSV = 0;

    // initialization of time of calculation

    tTimeBegin = time(0); // set actual time

    PrintMessage("EVOLUTION");

    int i = 0;

    // print monitor header

    PrintMonitorHeader();

    // export I.C. data

    cout << "\t" << i << "\r\t\t" << tTau;
    tMem = ExportMeasurables(tMem, i, tTau);

    // time evolution loop

    for (i = 1; i < mGrid->GetTP(); i++, tTau += mGrid->GetdT()){

        cout << "\t" << i << "\r\t\t" << tTau + mGrid->GetdT() <<"\r\t\t\t" << flush;

        // make a time step of the used stepper

        mStepper->MakeTimeStep(tTau);

        // smooth data using wLAX if it's turned on
        
        if (mSmoother) mSmoother->Smooth();
        
        // enforce Landau matching
       
        mEquations->EnforceLandauMatching(mStepper->GetArrays());
        
        // export data for time step
        
        tMem = ExportMeasurables(tMem, i, tTau + mGrid->GetdT());

        if(mTerminateQ) break;
    }

    // show the time of calculation

    tTimeEnd = time(0);
    tDate = GetTimeDifference(tTimeBegin, tTimeEnd);
    PrintEvolutionOutroMessage("Time Evolution : Done", tDate);

    // free the memory allocated by Stepper

    mStepper->FreeMemory();

    return (mBINDir + mHydroID + "/");
}

void Solver::ExportSnapshot(MemoryUsed &aMem, const double aTau) {

    PRINT_DEBUG_3("<Solver::ExportSnapshot>");

    if(mFileTSV){
        aMem.tMemoryUsedTSV += ExportSnapshotTSV(mStepper->GetArrays(), aTau);
    }
    if(mFileBIN){
        aMem.tMemoryUsedBIN += ExportSnapshotBIN(mStepper->GetArrays(), aTau);
    }
}

double Solver::ExportSnapshotTSV(Vector3D ***aArr, double aTau) {

    PRINT_DEBUG_3("<Solver::ExportSnapshotTSV>");

    double tMemoryUsed = 0;
    for(int i = 0; i < mEquations->GetNumberOfEquations(); ++i) tMemoryUsed += ExportArrayTSV(aArr[i][0], aTau, i);
    return tMemoryUsed;

}

double Solver::ExportSnapshotBIN(Vector3D ***aArr, double aTau) {

    PRINT_DEBUG_3("<Solver::ExportSnapshotBIN>");

    double tMemoryUsed = 0;
    for(int i = 0; i < mEquations->GetNumberOfEquations(); ++i) tMemoryUsed += ExportArrayBIN(aArr[i][0], aTau, i);
    return tMemoryUsed;

}

double Solver::ExportArrayBIN(Vector3D *aArr, double aTau, int aN) {

    PRINT_DEBUG_3("<Solver::ExportArrayBIN>");

    fstream tSnapshotFile;
    string  tSnapshotFileName = mBINDir + mHydroID + "/" + aArr->GetName() + ".bin";

    tSnapshotFile.open(tSnapshotFileName.c_str(), ios::app | ios:: out | ios::binary);

    // these two lines have to be after open() function

    struct stat tFileStatisticsBefore;
    stat(tSnapshotFileName.c_str(),&tFileStatisticsBefore);

    // buffer definition

    const int tSizeOfType = sizeof(float);
    const unsigned long long tBufferSize = mGrid->GetYP() * mGrid->GetZP();
    float tBuffer[tBufferSize];

    for (usi i = 0; i < mGrid->GetXP(); i++){
        for (usi j = 0; j < mGrid->GetYP(); j++)
            for (usi k = 0; k < mGrid->GetZP(); k++){
                tBuffer[j * mGrid->GetZP() + k] = static_cast<float>(mEquations->FormatOutput((*aArr)(i,j,k), aN));
            }
        tSnapshotFile.write(reinterpret_cast<char*>(&tBuffer), tSizeOfType * tBufferSize); // here we can also try cast in an old style c++ fashion
    }

    tSnapshotFile.close();

    struct stat tFileStatisticsAfter;
    stat(tSnapshotFileName.c_str(),&tFileStatisticsAfter);

    return (tFileStatisticsAfter.st_size - tFileStatisticsBefore.st_size);
}

double Solver::ExportArrayTSV(Vector3D *aArr, double aTau, int aN) {

    PRINT_DEBUG_3("<Solver::ExportArrayTSV>");

    fstream tSnapshotFile;
    ostringstream sTau;
    sTau << aTau;           // convert double to string
    string tSnapshotFileName = mTSVDir + mHydroID + "/" + aArr->GetName() + "_" + sTau.str() + ".tsv";

    tSnapshotFile.open(tSnapshotFileName.c_str(), ios::trunc | ios::out);
    tSnapshotFile.precision(6);

    stringstream tBuffer;

    for (usi i = 0; i < mGrid->GetXP(); i++){
        for (usi j = 0; j < mGrid->GetYP(); j++)
            for (usi k = 0; k < mGrid->GetZP(); k++){
                tBuffer << mGrid->GetX(i) << "\t" << mGrid->GetY(j) << "\t" << mGrid->GetZ(k) << "\t" << mEquations->FormatOutput((*aArr)(i,j,k), aN) << "\n";
            }
    }
    tSnapshotFile << tBuffer.rdbuf(); // faster than tBuffer.str();
    tBuffer.str("");

    tSnapshotFile.close();

    struct stat tFileStatistics;
    stat(tSnapshotFileName.c_str(),&tFileStatistics);

    return tFileStatistics.st_size;
}

EquationOfState* Solver::GetEquationOfState(){

    PRINT_DEBUG_3("<Solver::GetEquationOfState>");

    EquationOfState *tEquationOfState = 0;

    switch(mConfigurator->GetInt("EquationOfState")){

        case 1:

            tEquationOfState = new IdealEquationOfState(
                                                        mConfigurator->GetInt("IdealEquationOfStateStatistics"),
                                                        mConfigurator->GetDouble("IdealEquationOfStateDegeneracy"),
                                                        mConfigurator->GetInt("IdealEquationOfStateQuarkFlavors"));

            break;

        case 2:

            tEquationOfState = new RealisticEquationOfState(
                                                            mConfigurator->GetParameter("RealisticEquationOfStateDirectory"));

            break;

        default:

            exit(ProgramStatus(ERROR_OPTION_MISSING,
                               "Solver::GetEquationOfState: Cannot find equation of state for " + mConfigurator->GetParameter("EquationOfState")));
    }

    tEquationOfState->PrintInfo();

    return tEquationOfState;
}

Equations* Solver::GetEquations(){

    PRINT_DEBUG_3("<Solver::GetEquations>");

    Equations *tEquations = 0;

    switch(mConfigurator->GetInt("Equations")){

        case 1:

            tEquations = new PerfectFluidEquations(mConfigurator, mGrid);

            break;

        case 2:

            tEquations = new AnisotropicFluidEquations(mConfigurator, mGrid);

            break;

        case 3:

            tEquations = new AnisotropicFluidEquationsNew(mConfigurator, mGrid);

            break;

        default:

            exit(ProgramStatus(ERROR_OPTION_MISSING,
                               "Solver::GetEquations: Cannot find equations for " + mConfigurator->GetParameter("Equations")));
    }

    tEquations->PrintInfo();
    return tEquations;
}

Stepper* Solver::GetStepper(){

    PRINT_DEBUG_3("<Solver::GetStepper>");

    Stepper *tStepper = 0;

    switch(mConfigurator->GetInt("Stepper")){

        case 0:

            tStepper = new EulerStepper(mConfigurator, mGrid);

            break;

        case 1:

            tStepper = new RungeKuttaStepper(mConfigurator, mGrid);

            break;

        default:

            exit(ProgramStatus(ERROR_OPTION_MISSING,
                               "Solver::GetStepper: Cannot find stepper for " + mConfigurator->GetParameter("Stepper")));
    }

    tStepper->PrintInfo();
    return tStepper;
}

Smoother* Solver::GetSmoother(){

    PRINT_DEBUG_3("<Solver::GetSmoother>");

    Smoother *tSmoother = 0;

    switch(mConfigurator->GetInt("Smoother")){

        case 0:

            PrintRawMessage("-> No Smoother");

            break;

        case 1:

            tSmoother = new WeightedLAXSmoother(mConfigurator, mGrid, mStepper, mEquations);

            break;

        default:

            exit(ProgramStatus(ERROR_OPTION_MISSING,
                               "Solver::GetSmoother: Cannot find smoother for " + mConfigurator->GetParameter("Smoother")));
    }

    if(tSmoother!=0) tSmoother->PrintInfo();
    return tSmoother;
}

Initializer **  Solver::GetInitializerArr(void){

    PRINT_DEBUG_3("<Solver::GetInitializerArr>");

    Initializer **tIArr = 0;
    tIArr   = new Initializer* [mEquations->GetNumberOfEquations()];
    for(int i = 0; i < mEquations->GetNumberOfEquations(); ++i){
        PrintRawMessage("Configuring initializer for", mEquations->GetName(i));
        tIArr[i] = GetInitializer(mEquations->GetName(i));
    }
    return tIArr;
}

Initializer* Solver::GetInitializer(string aName){

    PRINT_DEBUG_3("<Solver::GetInitializer>");

    Initializer *tInitializer = 0;

    if(aName == "ux"){          // initial ux profile for pefect-fluid equations and anisotropic equations

        tInitializer = GetInitialFlowX();

    }else if(aName == "uy"){    // initial uy profile for pefect-fluid equations and anisotropic equations

        tInitializer = GetInitialFlowY();

    }else if(aName == "rap"){   // initial longitudinal rapidity profile for pefect-fluid equations and anisotropic equations

        tInitializer = GetInitialRapidity();

    }else if(aName == "lnt"){   // initial log(T) profile for pefect-fluid equations

        tInitializer = GetInitialSource();

    }else if(aName == "lambda"){// initial \Lambda profile for anisotropic equations

        tInitializer = GetInitialSource();

    }else if(aName == "xi"){    // initial anisotropy profile for anisotropic equations

        tInitializer = GetInitialAnisotropy("InitialAnisotropyValue");

    // following three need to be updated!!

    }else if(aName == "ax"){    // initial anisotropy profile for anisotropic equations

        tInitializer = GetInitialAnisotropy("InitialAX");

    }else if(aName == "ay"){    // initial anisotropy profile for anisotropic equations

        tInitializer = GetInitialAnisotropy("InitialAY");

    }else if(aName == "az"){    // initial anisotropy profile for anisotropic equations

        tInitializer = GetInitialAnisotropy("InitialAZ");

    }else if(aName == "temp"){    // initial temperature profile for anisotropic equations
        
        tInitializer = GetInitialSource();
        
    }else{

        tInitializer = 0;

        exit(ProgramStatus(ERROR_OPTION_MISSING,
                           "Solver::GetInitializer: Cannot find initializer for " + aName));

    }
    tInitializer->PrintInfo();

    return tInitializer;
}

InitialSource* Solver::GetInitialSource(){

    PRINT_DEBUG_3("<Solver::GetInitialSource>");

    InitialSource *tInitialSource;

    switch(mConfigurator->GetInt("InitialSource")){

        case 0:

            tInitialSource = new FlatInitialSource(mConfigurator, mGrid);
            break;

        case 1:

            tInitialSource = new GaussianInitialSource(mConfigurator, mGrid);
            break;

        case 2: case 3:

            tInitialSource = new GlauberInspiredInitialSource(mConfigurator, mEquationOfState, mGrid);
            break;

        case 4:

            tInitialSource = new GaussianBlobsInitialSource(mConfigurator, mGrid);
            break;

        case 5:

            tInitialSource = new MCGlauberInitialSource(mConfigurator, mGrid, mSeed);
            break;

        default:

            exit(ProgramStatus(ERROR_OPTION_MISSING,
                               "Solver::GetInitialSource: Cannot find initializer for " + mConfigurator->GetParameter("InitialSource")));
    }
    return tInitialSource;
}

InitialFlowX* Solver::GetInitialFlowX(){

    PRINT_DEBUG_3("<Solver::GetInitialFlowX>");

    InitialFlowX *tInitialFlowX;

    switch(mConfigurator->GetInt("UxInitialFlow")){

        case 1:

            tInitialFlowX = new NoInitialFlowX(mConfigurator, mGrid);

            break;

        default:

            exit(ProgramStatus(ERROR_OPTION_MISSING,
                               "Solver::GetInitialFlowX: Cannot find initializer for " + mConfigurator->GetParameter("UxInitialFlow")));
    }
    return tInitialFlowX;
}

InitialFlowY* Solver::GetInitialFlowY(){

    PRINT_DEBUG_3("<Solver::GetInitialFlowY>");

    InitialFlowY *tInitialFlowY;

    switch(mConfigurator->GetInt("UyInitialFlow")){

        case 1:

            tInitialFlowY = new NoInitialFlowY(mConfigurator, mGrid);

            break;

        default:

            exit(ProgramStatus(ERROR_OPTION_MISSING,
                               "Solver::GetInitialFlowY: Cannot find initializer for " + mConfigurator->GetParameter("UyInitialFlow")));
    }
    return tInitialFlowY;
}

InitialRapidity* Solver::GetInitialRapidity(){

    PRINT_DEBUG_3("<Solver::GetInitialRapidity>");

    InitialRapidity *tInitialRapidity;

    switch(mConfigurator->GetInt("RInitialFlow")){

        case 1:

            tInitialRapidity = new BjorkenInitialRapidity(mConfigurator, mGrid);

            break;

        default:

            exit(ProgramStatus(ERROR_OPTION_MISSING,
                               "Solver::GetInitialRapidity: Cannot find initializer for " + mConfigurator->GetParameter("RInitialFlow")));
    }
    return tInitialRapidity;
}

InitialAnisotropy* Solver::GetInitialAnisotropy(string aKey){

    PRINT_DEBUG_3("<Solver::GetInitialAnisotropy>");

    InitialAnisotropy *tInitialAnisotropy;

    switch(mConfigurator->GetInt("InitialAnisotropy")){

        case 1:

            tInitialAnisotropy = new FlatInitialAnisotropy(mConfigurator, mGrid, aKey);

            break;

        default:

            exit(ProgramStatus(ERROR_OPTION_MISSING,
                               "Solver::GetInitialAnisotropy: Cannot find initializer for " + mConfigurator->GetParameter("InitialAnisotropy")));
    }
    return tInitialAnisotropy;
}

double* Solver::GetTimeDifference(time_t aTimeStart, time_t aTimeStop, double aMult){

    PRINT_DEBUG_3("<Solver::GetTimeDifference>");

    double tTimeDifference = aMult * difftime(aTimeStop, aTimeStart);
    const int tSize = 4;
    double *tDate = new double [tSize];
    double tMultiplier[tSize - 1] = {60 * 60 * 24, 60 * 60, 60};
    for(int i = 0; i < tSize - 1; i++) {
        tDate[i] = floor(tTimeDifference / tMultiplier[i]);
        tTimeDifference = fmod(tTimeDifference, tMultiplier[i]);
    }
    tDate[tSize - 1] = tTimeDifference;

    return tDate;
}

void Solver::CheckParameters(void){

    PRINT_DEBUG_3("<Solver::CheckParameters>");



}

MemoryUsed Solver::ExportMeasurables(MemoryUsed tMem, int i, double time){

    PRINT_DEBUG_3("<Solver::ExportMeasurables>");

    if((mRecordingFrequency) && (!(i % mRecordingFrequency)) ){

        cout << " \r\t\t\t export... " << flush;

        ExportSnapshot(tMem, time);

        cout << " \r\t\t\t" <<  (tMem.tMemoryUsedBIN + tMem.tMemoryUsedTSV) / (1024 * 1024) << " MB.";

     
            if(mEquations->GetNumberOfEquations() == 8){
            if(tTau != mGrid->GetTMin()){
            mTerminateQ = true;

            for (i = 0; i < mGrid->GetXP(); i++)
                for (usi j = 0; j < mGrid->GetYP(); j++)
                    for (usi k = 0; k < mGrid->GetZP(); k++)
                      if(mConfigurator->GetDouble("FinalTemperature")/hbarc < mEquations->GetTeffpt(i,j,k)) mTerminateQ = false;


            if(mTerminateQ) {
                PrintRawMessage("");
            PrintRawMessage("FinalTemperature reached in the entire (x-y-eta)-volume. ");
            PrintRawMessage("Terminating evolution.");
                PrintRawMessage("");
            }
            }
            }

    }

    if(mEquations->GetNumberOfEquations() == 4){
        cout << "\r\t\t\t\t\t" << exp(((mStepper->GetArrays())[0][0])->Interpolate(0,0,0)) * hbarc << flush;
    }
    if(mEquations->GetNumberOfEquations() == 5){
        cout << "\r\t\t\t\t\t" << ((mStepper->GetArrays())[0][0])->Interpolate(0,0,0) * hbarc << flush;
        cout << "\r\t\t\t\t\t\t\t" << ((mStepper->GetArrays())[4][0])->Interpolate(0,0,0) << flush;
    }
    if(mEquations->GetNumberOfEquations() == 8){
        cout << "\r\t\t\t\t\t" << ((mStepper->GetArrays())[6][0])->Interpolate(0,0,0) * hbarc << flush;
        cout << "\r\t\t\t\t\t\t\t" << ((mStepper->GetArrays())[2][0])->Interpolate(0,0,0) << flush;
        cout << "\r\t\t\t\t\t\t\t\t\t" << ((mStepper->GetArrays())[7][0])->Interpolate(0,0,0) * hbarc << flush;
        //if(tTau != mGrid->GetTMin()){ cout << "\r\t\t\t\t\t\t\t\t\t\t\t" << mEquations->GetTeff(0,0,0) * hbarc << flush; }
        //cout << "\r\t\t\t\t\t\t\t\t\t\t\t\t\t" << ((mStepper->GetArrays())[0][0])->Interpolate(0,0,0) << flush;
        
    }

    cout << endl;


    return tMem;
}

void Solver::PrintMonitorHeader(void){

    PRINT_DEBUG_3("<Solver::PrintMonitorHeader>");

    cout << "\tSTEP" << "\r\t\tTAU [fm]"  << "\r\t\t\t  HD[MB]" << "\r\t\t\t\t\tLambda [GeV]" << flush;
    if(mEquations->GetNumberOfEquations() == 5) cout << "\r\t\t\t\t\t\t\txi" << flush;
    if(mEquations->GetNumberOfEquations() == 8) cout << "\r\t\t\t\t\t\t\talpha_z" <<  "\r\t\t\t\t\t\t\t\t\tTeff [GeV]" << flush;
    cout << endl;
}
