/**
 *
 *   ThermalMass.cxx
 *
 *   Copyright (C) 2016 Mubarak Alqahtani and Michael Strickland
 *
 *   This file is part of aHydro3p1.
 *
 *   aHydro3p1 is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   aHydro3p1 is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with aHydro3p1.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "Global.h"
#include "ThermalMass.h"

/*!
 *
 *  \brief Default constructor
 *
 */

ThermalMass::ThermalMass() {
    PRINT_DEBUG_3("<ThermalMass:ThermalMass>");
}


/*!
 *
 *  \brief Destructor
 *
 */

ThermalMass::~ThermalMass(){
    PRINT_DEBUG_3("<ThermalMass::~ThermalMass>");

}

//************************************************************************

void ThermalMass::readTfnc() {
    ifstream inputFile;
    char fname[255];
    sprintf(fname,"%sTfuncData.tsv",mEOSDir.c_str());
    string line;
    inputFile.open(fname);
    getline(inputFile, line); // read the header and discard since we don't need it
    getline(inputFile, line); // read number lines
    istringstream ss(line);
    ss >> nTfnc;
    TfncX = new double[nTfnc];
    TfncY = new double[nTfnc];
    int i = 0;
    while (getline(inputFile, line))
    {
        istringstream ss(line);
        ss >> TfncX[i] >> TfncY[i];
        i++;
    }
    inputFile.close();
}

void ThermalMass::readMassData() {
    ifstream inputFile;
    char fname[255];
    sprintf(fname,"%smassData.tsv",mEOSDir.c_str());
    string line;
    inputFile.open(fname);
    getline(inputFile, line); // read the header and discard since we don't need it
    getline(inputFile, line); // read number lines
    istringstream ss(line);
    ss >> npts;
    T = new double[npts];
    M = new double[npts];
    DM = new double[npts];
    BEQ = new double[npts];
    int i = 0;
    while (getline(inputFile, line))
    {
        istringstream ss(line);
        ss >> T[i] >> M[i] >> DM[i] >> BEQ[i];
        i++;
    }
    inputFile.close();
    
    // load mins and maxes
    
    mTmin = T[0];
    mTmax = T[npts-1];
    
    mMmin = M[0];
    mMmax = M[npts-1];
    
    mDMmin = DM[0];
    mDMmax = DM[npts-1];
    
    mBmin = BEQ[0];
    mBmax = BEQ[npts-1];
    
}

void ThermalMass::loadInterpolations(string aEOSDir) {
    
    //PrintRawMessage("==> Initializing mass, Beq, and Temperature interpolations\n");
    
    mEOSDir = aEOSDir; // set EOSDir for local use
    
    readMassData(); // read mass data from file
    
    accm = gsl_interp_accel_alloc();
    splinem = gsl_spline_alloc(gsl_interp_cspline, npts);
    gsl_spline_init(splinem, T, M, npts);
    
    accdm = gsl_interp_accel_alloc();
    splinedm = gsl_spline_alloc(gsl_interp_cspline, npts);
    gsl_spline_init(splinedm, T, DM, npts);
    
    accbeq = gsl_interp_accel_alloc();
    splinebeq = gsl_spline_alloc(gsl_interp_cspline, npts);
    gsl_spline_init(splinebeq, T, BEQ, npts);
    
    readTfnc(); // loads TfncX, TfncY, and nTfnc
    
    accTfnc = gsl_interp_accel_alloc();
    splineTfnc = gsl_spline_alloc(gsl_interp_cspline, nTfnc);
    gsl_spline_init(splineTfnc, TfncX, TfncY, nTfnc);
    
}

void ThermalMass::freeInterpolations() {
    gsl_spline_free(splinem);
    gsl_interp_accel_free(accm);
    gsl_spline_free(splinedm);
    gsl_interp_accel_free(accdm);
    gsl_spline_free(splinebeq);
    gsl_interp_accel_free(accbeq);
    gsl_spline_free(splineTfnc);
    gsl_interp_accel_free(accTfnc);
}


double ThermalMass::mass(double T) {
    if (T < mTmin) {
        //PrintRawMessage("OutOfBounds Tmin -> Mmin",mMmin);
        return mMmin;
    }
    else if (T > mTmax) {
        //PrintRawMessage("OutOfBounds Tmax -> Mmax",mMmax);
        return mMmax;
    }
    else return gsl_spline_eval(splinem, T, accm);
}

double ThermalMass::dTmass(double T) {
    if (T < mTmin) {
        //PrintRawMessage("OutOfBounds Tmin -> DMmax");
        return 0;
    }
    else if (T > mTmax) {
        //PrintRawMessage("OutOfBounds Tmax -> DMmax");
        return 0;
    }
    else return gsl_spline_eval(splinedm, T, accdm);
}

double ThermalMass::Beq(double T) {
    if (T < mTmin) {
        //PrintRawMessage("OutOfBounds Tmin -> Bmin");
        return mBmin;
    }
    else if (T > mTmax) {
        //PrintRawMessage("OutOfBounds Tmax -> Bmax");
        return mBmax;
    }
    else return gsl_spline_eval(splinebeq, T, accbeq);
}
