/**
 *
 *   TriangulationFreezer.cxx
 *
 *   Copyright (C) 2017 Radoslaw Ryblewski and Michael Strickland
 *
 *   This file is part of aHydro3p1.
 *
 *   aHydro3p1 is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   aHydro3p1 is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with aHydro3p1.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "Global.h"

/*!
 *
 *  \brief Default constructor
 *
 *  \li aConfigurator = pointer to configuration
 *  \li aBINDirectory = directory containing the binary file to perform freeze-out extraction on
 *
 */

TriangulationFreezer::TriangulationFreezer(Configurator *aConfigurator, string aBINDir)
: mConfigurator(aConfigurator), mBINDir(aBINDir), mGrid(0), mField(0), mNumberOfEquations(0){
    
    PRINT_DEBUG_3("<TriangulationFreezer::TriangulationFreezer>");
    
    mTSVDir  = mConfigurator->GetParameter("TSVDir");
    mXMLDir  = mConfigurator->GetParameter("XMLDir");
    mHydroID = GetFilename(GetDirectory(mBINDir));
    
    //options for freeze-out extraction
    
    TFO = mConfigurator->GetDouble("TFO") / hbarc; // GeV -> fm
    
    // switch between different hydrodynamic equations types
    
    mEquationsChoice  = mConfigurator->GetInt("Equations");
    
    switch(mEquationsChoice){
            
        case 1: // perfect fluid
            
            mNumberOfEquations = 4;
            mNames   = new string [mNumberOfEquations];
            mNamesFO = new string [mNumberOfEquations];
            mUnitsFO = new string [mNumberOfEquations];
            
            mNames[0] = "lnt";
            mNames[1] = "rap";
            mNames[2] = "ux";
            mNames[3] = "uy";
            
            mNamesFO[0] = "Distance";
            mNamesFO[1] = "FluidRap";
            mNamesFO[2] = "FluidUx";
            mNamesFO[3] = "FluidUy";
            
            mUnitsFO[0] = "GeV^-1";
            mUnitsFO[1] = "1";
            mUnitsFO[2] = "1";
            mUnitsFO[3] = "1";
            
            break;
            
        case 2: // anisotropic fluid
            
            mNumberOfEquations = 5;
            mNames   = new string [mNumberOfEquations];
            mNamesFO = new string [mNumberOfEquations];
            mUnitsFO = new string [mNumberOfEquations];
            
            mNames[0] = "lambda";
            mNames[1] = "rap";
            mNames[2] = "ux";
            mNames[3] = "uy";
            mNames[4] = "xi";
            
            mNamesFO[0] = "Distance";
            mNamesFO[1] = "FluidRap";
            mNamesFO[2] = "FluidUx";
            mNamesFO[3] = "FluidUy";
            mNamesFO[4] = "FluidXi";
            
            mUnitsFO[0] = "GeV^-1";
            mUnitsFO[1] = "1";
            mUnitsFO[2] = "1";
            mUnitsFO[3] = "1";
            mUnitsFO[4] = "1";
            
            break;
            
        case 3: // anisotropic fluid - new
            
            mNumberOfEquations = 9; // 8 + additional temp table for THERMINATOR
            mNames   = new string [mNumberOfEquations];
            mNamesFO = new string [mNumberOfEquations];
            mUnitsFO = new string [mNumberOfEquations];
            
            mNames[0] = "temp";
            mNames[1] = "ax";
            mNames[2] = "ay";
            mNames[3] = "az";
            mNames[4] = "ux";
            mNames[5] = "uy";
            mNames[6] = "rap";
            mNames[7] = "lambda";
            mNames[8] = "temp";
            
            mNamesFO[0] = "FreezeOutVar";
            mNamesFO[1] = "FluidAx";
            mNamesFO[2] = "FluidAy";
            mNamesFO[3] = "FluidAz";
            mNamesFO[4] = "FluidUx";
            mNamesFO[5] = "FluidUy";
            mNamesFO[6] = "FluidRap";
            mNamesFO[7] = "FluidLambda";
            mNamesFO[8] = "FluidTeff";
            
            mUnitsFO[0] = "1/fm";
            mUnitsFO[1] = "1";
            mUnitsFO[2] = "1";
            mUnitsFO[3] = "1";
            mUnitsFO[4] = "1";
            mUnitsFO[5] = "1";
            mUnitsFO[6] = "1";
            mUnitsFO[7] = "1/fm";
            mUnitsFO[8] = "1/fm";
            
            break;
            
        default:
            
            exit(ProgramStatus(ERROR_OPTION_MISSING,
                               "TriangulationFreezer::TriangulationFreezer: Cannot find Equations for " + d2s(mEquationsChoice)));
    }
    
    // read .bin file containing "temperature" like quantity into mField, be carefull about units
    Parser *tParser = 0;
    tParser = new Parser(mBINDir + mNames[0] + ".bin");
    mField = tParser->ReadBIN(mConfigurator);
    delete tParser;
    
    // ranges of hydro grid
    
    tmin = mField->GetTMin();
    tmax = mField->GetTMax();
    xmax = mField->GetXMax();
    ymax = mField->GetYMax();
    emax = mField->GetZMax();
    
    deltat = (mField->GetTMax() - mField->GetTMin()) / (mField->GetTPts() - 1);
    deltax = 2 * mField->GetXMax() / (mField->GetXPts() - 1);
    deltay = 2 * mField->GetYMax() / (mField->GetYPts() - 1);
    deltae = 2 * mField->GetZMax() / (mField->GetZPts() - 1);
    
    mEOS = GetEquationOfState();
    meFO = mEOS->ET(TFO);
    
    PrintInfo();
    
}

/*!
 *
 *  \brief Default destructor
 *
 */

TriangulationFreezer::~TriangulationFreezer(){
    
    PRINT_DEBUG_3("<TriangulationFreezer::~TriangulationFreezer>");
    
    // free memory
    
    if(mField){
        delete mField;
        mField = 0;
    }
    
    if(mNames){
        delete [] mNames;
        mNames = 0;
    }
    
    if(mNamesFO){
        delete [] mNamesFO;
        mNamesFO = 0;
    }
    
    if(mUnitsFO){
        delete [] mUnitsFO;
        mUnitsFO = 0;
    }
    
}

void TriangulationFreezer::ExtractFOHS(void){
    
    FindCells();
    
    MakeDirectory(mXMLDir);
    SetDescriptions();
    ExportFOHSXML(mXMLDir + mHydroID + "_FOHS.xml");
    
    PrintRawMessage("");
    PrintMessage("Done");
    
}


EquationOfState* TriangulationFreezer::GetEquationOfState(){
    
    EquationOfState *tEquationOfState = 0;
    
    switch(mConfigurator->GetInt("EquationOfState")){
            
        case 1:
            
            tEquationOfState = new IdealEquationOfState(
                                                        mConfigurator->GetInt("IdealEquationOfStateStatistics"),
                                                        mConfigurator->GetDouble("IdealEquationOfStateDegeneracy"),
                                                        mConfigurator->GetInt("IdealEquationOfStateQuarkFlavors"));
            
            break;
            
        case 2:
            
            tEquationOfState = new RealisticEquationOfState(
                                                            mConfigurator->GetParameter("RealisticEquationOfStateDirectory"));
            
            break;
            
        default:
            
            exit(ProgramStatus(ERROR_OPTION_MISSING,
                               "TriangulationFreezer::GetEquationOfState: Cannot find equation of state for " + mConfigurator->GetParameter("EquationOfState")));
    }
    
    tEquationOfState->PrintInfo();
    
    return tEquationOfState;
}

/*!
 *
 *  \brief Prints Info
 *
 */

void TriangulationFreezer::PrintInfo(void){
    
    PrintRawMessage("");
    PrintRawMessage("Freeze-out method", "CORNELIUS 1.3");
    PrintRawMessage("");
    PrintRawMessage("Field domain:");
    PrintRawMessage("   tau", "(" + d2s(tmin ) + "," + d2s(tmax) + ")");
    PrintRawMessage("   x",   "(" + d2s(-xmax) + "," + d2s(xmax) + ")");
    PrintRawMessage("   y",   "(" + d2s(-ymax) + "," + d2s(ymax) + ")");
    PrintRawMessage("   eta", "(" + d2s(-emax) + "," + d2s(emax) + ")");
    PrintRawMessage("Field grid:");
    PrintRawMessage("   dtau", deltat);
    PrintRawMessage("   dx",   deltax);
    PrintRawMessage("   dy",   deltay);
    PrintRawMessage("   deta", deltae);
    PrintRawMessage("");
    
}

void TriangulationFreezer::FindCells(void){
    
    
    // finds all cells with the freeze-out surface using the condition that at least one cell corner is above and one is below isovalue
    
    const double iso = TFO;
    
    const double i1max = mField->GetTPts() - 1;
    const double i2max = mField->GetXPts() - 1;
    const double i3max = mField->GetYPts() - 1;
    const double i4max = mField->GetZPts() - 1;
    
    HyperCube *tHCube = new HyperCube;
    
    // initialize Cornelius
    
    int dim = 4;
    double *dx = new double[dim];
    dx[0] = deltat; // tau
    dx[1] = deltax;
    dx[2] = deltay;
    dx[3] = deltae; // eta
    Cornelius tCornelius;
    tCornelius.init(dim,iso,dx);
    
    unsigned long int NEfound = 0;
    
    double ttau;
    double tx;
    double ty;
    double teta;
    double tau = 0;
    
    quadruple centroid;
    quadruple normalvector;
    
    // go over all cells in the table, (i1,i2,i3,i4) <=> (tau,x,y,eta)
    
    for (int i1 = 0; i1 < i1max; ++i1)
        for (int i2 = 0; i2 < i2max; ++i2)
            for (int i3 = 0; i3 < i3max; ++i3)
                for (int i4 = 0; i4 < i4max; ++i4) {
                    
                    // load the hypercube from the freezeout field
                    tHCube->SetHyperCube(mField, i1,i2,i3,i4);
                    
                    // check for surface(s) within the hypercube
                    tCornelius.find_surface_4d(tHCube->GetHyperCube());
                    
                    // if surface(s) found
                    if (tCornelius.get_Nelements()>0) {
                        
                        NEfound += tCornelius.get_Nelements();
                        
                        // compute absolute position of the hypercube being considered
                        ttau    =  tmin + i1 * deltat;
                        tx      = -xmax + i2 * deltax;
                        ty      = -ymax + i3 * deltay;
                        teta    = -emax + i4 * deltae;
                        
                        // output data to file
                        for (int i = 0; i < tCornelius.get_Nelements(); ++i) {
                            
                            // compute tau
                            tau = ttau + tCornelius.get_centroid_elem(i,0);
                            // load centroid
                            centroid.t = tau;
                            centroid.x = tx + tCornelius.get_centroid_elem(i,1);
                            centroid.y = ty + tCornelius.get_centroid_elem(i,2);
                            centroid.z = teta + tCornelius.get_centroid_elem(i,3);
                            mCentroids.push_back(centroid);
                            // load normalvector
                            normalvector.t = tau * tCornelius.get_normal_elem(i,0);
                            normalvector.x = tau * tCornelius.get_normal_elem(i,1);
                            normalvector.y = tau * tCornelius.get_normal_elem(i,2);
                            normalvector.z = tau * tCornelius.get_normal_elem(i,3);
                            mNormalVectors.push_back(normalvector);
                        }
                        
                    }
                    
                }
    
    
    PrintRawMessage("Number of studied hypercubes", i1max * i2max * i3max * i4max);
    PrintRawMessage("Number of surface elements found", NEfound);
    
    if(tHCube) {delete tHCube; tHCube = 0;}
    
}

void TriangulationFreezer::SetDescriptions(void){
    
    // set descriptions of parameters
    
    mEquations = "UNKNOWN";
    switch(mConfigurator->GetInt("Equations")){
            
        case 1:
            mEquations = "3+1D Perfect Fluid";
            break;
            
        case 2:
            mEquations = "3+1D Anisotropic Fluid (LO)";
            break;
            
        case 3:
            mEquations = "3+1D Anisotropic Fluid (LO) New";
            break;
            
        default:
            PrintRawMessage("Warning: Unknown option for", "<Equations>.");
    }
    
    mEquationOfState = "UNKNOWN";
    switch(mConfigurator->GetInt("EquationOfState")){
            
        case 1:
            mEquationOfState = "Ideal";
            break;
            
        case 2:
            mEquationOfState = "Realistic";
            break;
            
        default:
            PrintRawMessage("Warning: Unknown option for", "<EquationOfState>.");
    }
    
    mDeviceType = "UNKNOWN";
    //    switch(mConfigurator->GetInt("Device")){
    
    //        case 1:
    //            mDeviceType = "RHIC";
    //            break;
    //
    //        case 2:
    //            mDeviceType = "LHC";
    //            break;
    //
    //        default:
    //            PrintRawMessage("Warning: Unknown option for", "<Device>.");
    //    }
    
    mCollisionalSystem = "UNKNOWN";
    //    switch(mConfigurator->GetInt("CollisionalSystem")){
    //
    //        case 1:
    //            mCollisionalSystem = "AuAu";
    //            break;
    //
    //        case 2:
    //            mCollisionalSystem = "PbPb";
    //            break;
    //
    //        default:
    //            PrintRawMessage("Warning: Unknown option for", "<CollisionalSystem>.");
    //    }
    
    mCMin = "0";
    mCMax = "0";
    
    mInitialSource = "UNKNOWN";
    switch(mConfigurator->GetInt("InitialSource")){
            
        case 0:
            mInitialSource = "Flat";
            break;
            
        case 1:
            mInitialSource = "3D Gauss";
            break;
            
        case 2:
            mInitialSource = "Glauber Inspired - Shifted source";
            break;
            
        case 3:
            mInitialSource = "Glauber Inspired - Tilted source";
            break;
            
        case 4:
            mInitialSource = "Sum of Random 3D Gaussian Bumps";
            break;
            
        case 5:
            mInitialSource = "Monte Carlo Glauber";
            break;
            
        default:
            PrintRawMessage("Warning: Unknown option for", "<InitialSource>.");
    }
    
    mDensityScaling = "UNKNOWN";
    switch(mConfigurator->GetInt("DensityOfSourcesScaling")){
            
        case 1:
            mDensityScaling = "Entropy density";
            break;
            
        case 2:
            mDensityScaling = "Energy density";
            break;
            
        default:
            PrintRawMessage("Warning: Unknown option for", "<DensityOfSourcesScaling>.");
    }
    
    mDensityNormalization = "UNKNOWN";
    switch(mConfigurator->GetInt("DensityOfSourcesNormalization")){
            
        case 1:
            mDensityNormalization = "Normalize to central";
            break;
            
        case 2:
            mDensityNormalization = "Normalize to non-central";
            break;
            
        default:
            PrintRawMessage("Warning: Unknown option for", "<DensityOfSourcesNormalization>.");
    }
    
    mUxInitialFlow = "UNKNOWN";
    switch(mConfigurator->GetInt("UxInitialFlow")){
            
        case 1:
            mUxInitialFlow= "No flow";
            break;
            
        default:
            PrintRawMessage("Warning: Unknown option for", "<UxInitialFlow>.");
    }
    
    mUyInitialFlow = "UNKNOWN";
    switch(mConfigurator->GetInt("UyInitialFlow")){
            
        case 1:
            mUyInitialFlow = "No flow";
            break;
            
        default:
            PrintRawMessage("Warning: Unknown option for", "<UyInitialFlow>.");
    }
    
    mRInitialFlow = "UNKNOWN";
    switch(mConfigurator->GetInt("RInitialFlow")){
            
        case 1:
            mRInitialFlow = "Bjorken";
            break;
            
        default:
            PrintRawMessage("Warning: Unknown option for", "<RInitialFlow>.");
    }
    
    mIdealStatistics = "UNKNOWN";
    switch(mConfigurator->GetInt("IdealEquationOfStateStatistics")){
            
        case 1:
            mIdealStatistics = "Boltzmann";
            break;
            
        case 2:
            mIdealStatistics = "BE";
            break;
            
        case 3:
            mIdealStatistics = "FD";
            break;
            
        default:
            PrintRawMessage("Warning: Unknown option for", "<IdealEquationOfStateStatistics>.");
    }
}

void TriangulationFreezer::ExportFOHSXML(string aFileName){
    
    ofstream tFile;
    ostringstream oss;
    
    PrintRawMessage("Exporting freeze-out hypersurface to " + aFileName + " file.");
    
    tFile.open(aFileName.c_str(), ios::trunc | ios::out);
    
    // export the XML tags to the .xml file
    
    // general information about the run
    oss << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" << endl;
    oss << "<HYPERSURFACE filename=\"" << mHydroID << "\" version=\"1.0\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xsi:noNamespaceSchemaLocation=\"http://therminator2.ifj.edu.pl/hypersurface.xsd\">" << endl;
    oss << GetPARAMETERS("\t") << endl;
    oss << GetDESCRIPTION("\t") << endl;
    oss << "<!-- Included variable names : ";
    for (int n = 1; n < mNumberOfEquations;++n) oss << mNamesFO[n] << " ";
    oss << " -->" << endl;
    tFile << oss.str(); oss.str("");
    
    // output centroids
    oss << "<VECTOR name=\"centroids\" unit=\"1\" count=\"" << mNormalVectors.size() << "\">" << endl;
    for (int i=0; i < mCentroids.size(); i++) {
        quadruple centroid = mCentroids.at(i);
        oss << centroid.t << "\t";
        oss << centroid.x << "\t";
        oss << centroid.y << "\t";
        oss << centroid.z << "\n";
    }
    oss << "</VECTOR>" << endl;
    
    // output normal vectors
    oss << "<VECTOR name=\"normalvectors\" unit=\"1\" count=\"" << mNormalVectors.size() << "\">" << endl;
    for (int i=0; i < mNormalVectors.size(); i++) {
        quadruple normalvector = mNormalVectors.at(i);
        oss << normalvector.t << "\t";
        oss << normalvector.x << "\t";
        oss << normalvector.y << "\t";
        oss << normalvector.z << "\n";
    }
    oss << "</VECTOR>" << endl;
    tFile << oss.str(); oss.str("");
    
    // output all variables interpolated to the centroid
    for (int n = 1; n < mNumberOfEquations; n++) ExportVariable("\t",tFile,n);
    
    oss << "</HYPERSURFACE>" << endl;
    tFile << oss.str(); oss.str("");
    tFile.close();
    
}

void TriangulationFreezer::ExportVariable(string Tab, ofstream &aFile, int n){
    
    PrintRawMessage("----------------------------------");
    
    PrintRawMessage("Exporting " + mNamesFO[n]);
    
    // read the .bin file containing the variable ??? Check units ???
    Parser *tParser = 0;
    tParser = new Parser(mBINDir + mNames[n] + ".bin");
    Vector4D *myVariable = 0;
    myVariable = tParser->ReadBIN(mConfigurator);
    delete tParser;
    
    ostringstream oss;
    oss << "<VECTOR name=\"" << mNamesFO[n] << "\" unit=\"" << mUnitsFO[n] << "\" count=\"" << mNormalVectors.size()<< "\">" << endl;
    aFile << oss.str(); oss.str("");
    
    for (int i=0; i < mCentroids.size(); i++) {
        quadruple centroid = mCentroids.at(i);
        oss << myVariable->Interpolate(centroid.t,centroid.x,centroid.y,centroid.z);
        if (i<mCentroids.size()-1) oss << "\t";
        else oss << endl;
    }
    aFile << oss.str(); oss.str("");
    
    oss << "</VECTOR>" << endl;
    aFile << oss.str(); oss.str("");
    
}

string TriangulationFreezer::GetPARAMETERS(string Tab){
    
    ostringstream oss;
    
    // export tags
    
    oss << "<PARAMETERS>" << endl;
    oss << Tab << GetTAG("PARAMETER","Tau_i","fm", mConfigurator->GetDouble("TauMin")) << endl;
    oss << Tab << GetTAG("PARAMETER","Lambda", "fm", mConfigurator->GetDouble("Lambda")) << endl;
    oss << Tab << GetTAG("PARAMETER","Temperature", "MeV", mConfigurator->GetDouble("TFO") * 1000) << endl;
    oss << Tab << GetTAG("PARAMETER","Mu_B", "MeV", mConfigurator->GetDouble("Mu_B")) << endl;
    oss << Tab << GetTAG("PARAMETER","Mu_I", "MeV", mConfigurator->GetDouble("Mu_I")) << endl;
    oss << Tab << GetTAG("PARAMETER","Mu_S", "MeV", mConfigurator->GetDouble("Mu_S")) << endl;
    oss << Tab << GetTAG("PARAMETER","Mu_C", "MeV", mConfigurator->GetDouble("Mu_C")) << endl;
    oss << Tab << GetTAG("PARAMETER","device", "1", mDeviceType) << endl;
    oss << Tab << GetTAG("PARAMETER","colliding_system", "1", mCollisionalSystem) << endl;
    oss << Tab << GetTAG("PARAMETER","colliding_energy", "GeV", mConfigurator->GetDouble("CollisionEnergy")) << endl;
    oss << Tab << GetTAG("PARAMETER","centrality_min", "%", mCMin) << endl;
    oss << Tab << GetTAG("PARAMETER","centrality_max", "%", mCMax) << endl;
    oss << Tab << GetTAG("PARAMETER","impact_parameter", "fm", mConfigurator->GetDouble("ImpactParameter")) << endl;
    oss << Tab << GetTAG("PARAMETER","temperature_at_center", "MeV", mConfigurator->GetDouble("InitialTemperature") * 1000) << endl;
    oss << "</PARAMETERS>";
    
    return oss.str();
}

string TriangulationFreezer::GetDESCRIPTION(string Tab){
    
    ostringstream oss;
    
    oss << "<DESCRIPTION>" << endl;
    oss << Tab << "<ITEM name=\"" << "initial_nuclear_profile" << "\">" << endl;
    oss << Tab << Tab << GetTAG("DETAIL","type","1", mInitialSource) << endl;
    
    switch(mConfigurator->GetInt("InitialSource")){
            
        case 0:
            
            //oss << Tab << Tab << GetTAG("DETAIL","T0","MeV", mConfigurator->GetDouble("InitialTemperature") * 1000) << endl;
            oss << Tab << "</ITEM>" << endl;
            break;
            
        case 1:
            
            oss << Tab << Tab << GetTAG("DETAIL","sigma_x","fm", mConfigurator->GetDouble("SigmaX")) << endl;
            oss << Tab << Tab << GetTAG("DETAIL","sigma_y","fm", mConfigurator->GetDouble("SigmaY")) << endl;
            oss << Tab << Tab << GetTAG("DETAIL","sigma_eta","1", mConfigurator->GetDouble("SigmaZ")) << endl;
            //oss << Tab << Tab << GetTAG("DETAIL","T0","MeV", mConfigurator->GetDouble("InitialTemperature") * 1000) << endl;
            oss << Tab << "</ITEM>" << endl;
            break;
            
        case 2: case 3:
            
            oss << Tab << Tab << GetTAG("DETAIL","impact_parameter", "fm", mConfigurator->GetDouble("ImpactParameter")) << endl;
            oss << Tab << Tab << GetTAG("DETAIL","shift_in_rapidity", "1", mConfigurator->GetDouble("ShiftInRapidity")) << endl;
            oss << Tab << Tab << GetTAG("DETAIL","plateau_half_width", "1", mConfigurator->GetDouble("PlateauHalfWidth")) << endl;
            oss << Tab << Tab << GetTAG("DETAIL","gauss_half_width", "1", mConfigurator->GetDouble("GaussianHalfWidth")) << endl;
            //oss << Tab << Tab << GetTAG("DETAIL","T0", "MeV", mConfigurator->GetDouble("InitialTemperature") * 1000) << endl;
            oss << Tab << "</ITEM>" << endl;
            
            oss << Tab << "<ITEM name=\"" << "eos_to_initial_condition" << "\">" << endl;
            oss << Tab << mDensityScaling << endl;
            oss << Tab << "</ITEM>" << endl;
            
            oss << Tab << "<ITEM name=\"" << "normalization" << "\">" << endl;
            oss << Tab << mDensityNormalization << endl;
            oss << Tab << "</ITEM>" << endl;
            
            oss << Tab << "<ITEM name=\"" << "glauber_optical" << "\">" << endl;
            oss << Tab << Tab << GetTAG("DETAIL","InelasticCrossSection","mb", mConfigurator->GetDouble("InelasticCrossSection")) << endl;
            oss << Tab << Tab << GetTAG("DETAIL","MixingFactor", "1", mConfigurator->GetDouble("MixingFactor")) << endl;
            oss << Tab << Tab << GetTAG("DETAIL","NumberOfNucleonsA", "1", mConfigurator->GetDouble("NumberOfNucleonsA")) << endl;
            oss << Tab << Tab << GetTAG("DETAIL","NumberOfNucleonsB", "1", mConfigurator->GetDouble("NumberOfNucleonsB")) << endl;
            oss << Tab << Tab << GetTAG("DETAIL","SurfaceDiffusenessA", "fm", mConfigurator->GetDouble("SurfaceDiffusenessA")) << endl;
            oss << Tab << Tab << GetTAG("DETAIL","SurfaceDiffusenessB", "fm", mConfigurator->GetDouble("SurfaceDiffusenessB")) << endl;
            oss << Tab << Tab << GetTAG("DETAIL","NuclearSaturationDensityA", "1", mConfigurator->GetDouble("NuclearSaturationDensityA")) << endl;
            oss << Tab << Tab << GetTAG("DETAIL","NuclearSaturationDensityB", "1", mConfigurator->GetDouble("NuclearSaturationDensityB")) << endl;
            oss << Tab << "</ITEM>" << endl;
            
            break;
            
        case 5:
            
            oss << Tab << Tab << GetTAG("DETAIL","impact_parameter", "fm", mConfigurator->GetDouble("ImpactParameter")) << endl;
            //oss << Tab << Tab << GetTAG("DETAIL","shift_in_rapidity", "1", mConfigurator->GetDouble("ShiftInRapidity")) << endl;
            //oss << Tab << Tab << GetTAG("DETAIL","plateau_half_width", "1", mConfigurator->GetDouble("PlateauHalfWidth")) << endl;
            //oss << Tab << Tab << GetTAG("DETAIL","gauss_half_width", "1", mConfigurator->GetDouble("GaussianHalfWidth")) << endl;
            //oss << Tab << Tab << GetTAG("DETAIL","T0", "MeV", mConfigurator->GetDouble("InitialTemperature") * 1000) << endl;
            oss << Tab << "</ITEM>" << endl;
            
            break;
            
    }
    
    oss << Tab << "<ITEM name=\"" << "initial_flow" << "\">" << endl;
    oss << Tab << Tab << GetTAG("DETAIL","x-direction","1", mUxInitialFlow) << endl;
    oss << Tab << Tab << GetTAG("DETAIL","y-direction","1", mUyInitialFlow) << endl;
    oss << Tab << Tab << GetTAG("DETAIL","eta-direction","1", mRInitialFlow) << endl;
    oss << Tab << "</ITEM>" << endl;
    
    oss << Tab << "<ITEM name=\"" << "equation_of_state" << "\">" << endl;
    oss << Tab << Tab << GetTAG("DETAIL","type","1", mEquationOfState) << endl;
    if(mConfigurator->GetInt("EquationOfState") == 1){
        oss << Tab << Tab << GetTAG("DETAIL","statistics","1", mIdealStatistics) << endl;
        oss << Tab << Tab << GetTAG("DETAIL","degeneracy_factor","1", mConfigurator->GetDouble("IdealEquationOfStateDegeneracy")) << endl;
        oss << Tab << Tab << GetTAG("DETAIL","quark_flavors","1", mConfigurator->GetInt("IdealEquationOfStateQuarkFlavors")) << endl;
    }
    if(mConfigurator->GetInt("EquationOfState") == 2){
        oss << Tab << Tab << GetTAG("DETAIL","T ht TC","1", "QGP_from_lattice_QCD") << endl;
        oss << Tab << Tab << GetTAG("DETAIL","T eq TC","1", "crossover") << endl;
        oss << Tab << Tab << GetTAG("DETAIL","T lt TC","1", "Hadron_gas") << endl;
    }
    oss << Tab << "</ITEM>" << endl;
    
    oss << Tab << "<ITEM name=\"" << "hydrodynamic_equations" << "\">" << endl;
    oss << Tab << Tab << GetTAG("DETAIL","type","1",mEquations) << endl;
    oss << Tab << Tab << GetTAG("DETAIL","tau_min","fm", tmin) << endl;
    oss << Tab << Tab << GetTAG("DETAIL","tau_max","fm", tmax) << endl;
    oss << Tab << Tab << GetTAG("DETAIL","tau_step_size","1", deltat) << endl;
    oss << Tab << Tab << GetTAG("DETAIL","x_min","fm", -xmax) << endl;
    oss << Tab << Tab << GetTAG("DETAIL","x_max","fm",  xmax) << endl;
    oss << Tab << Tab << GetTAG("DETAIL","x_spacing","1", deltax) << endl;
    oss << Tab << Tab << GetTAG("DETAIL","y_min","fm", -ymax) << endl;
    oss << Tab << Tab << GetTAG("DETAIL","y_max","fm",  ymax) << endl;
    oss << Tab << Tab << GetTAG("DETAIL","y_spacing","1", deltay) << endl;
    oss << Tab << Tab << GetTAG("DETAIL","eta_min","fm", -emax) << endl;
    oss << Tab << Tab << GetTAG("DETAIL","eta_max","fm",  emax) << endl;
    oss << Tab << Tab << GetTAG("DETAIL","eta_spacing","1", deltae) << endl;
    oss << Tab << "</ITEM>" << endl;
    
    oss << "</DESCRIPTION>";
    
    return oss.str();
}

string TriangulationFreezer::GetTAG(string aTagName, string aDetailName, string aUnit, double aValue){
    
    ostringstream oss;
    oss << "<" << aTagName << " name=\"" << aDetailName << "\" unit=\"" << aUnit << "\">" << aValue << "</" << aTagName << ">";
    return oss.str();
    
}

string TriangulationFreezer::GetTAG(string aTagName, string aDetailName, string aUnit, int aValue){
    
    ostringstream oss;
    oss << "<" << aTagName << " name=\"" << aDetailName << "\" unit=\"" << aUnit << "\">" << aValue << "</" << aTagName << ">";
    return oss.str();
    
}

string TriangulationFreezer::GetTAG(string aTagName, string aDetailName, string aUnit, string aValue){
    
    ostringstream oss;
    oss << "<" << aTagName << " name=\"" << aDetailName << "\" unit=\"" << aUnit << "\">" << aValue << "</" << aTagName << ">";
    return oss.str();
    
}

