/** 
*
*   Vector1D.cxx
*	
*   Copyright (C) 2017 Radoslaw Ryblewski and Michael Strickland
*
*   This file is part of aHydro3p1.
*
*   aHydro3p1 is free software: you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation, either version 3 of the License, or
*   (at your option) any later version.
*
*   aHydro3p1 is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with aHydro3p1.  If not, see <http://www.gnu.org/licenses/>.
*    
*/

#include "Global.h"

// STATIC VARIABLES

double Vector1D::mXin;
double Vector1D::mXi0;
double Vector1D::mXip;
double Vector1D::mX0;
double Vector1D::mXd;


/*!
*
*  \brief Default constructor
*
*/

Vector1D::Vector1D()
: mVecName(""), mVec(0), mXmin(0.0), mXmax(1.0), mXpts(1){

    PRINT_DEBUG_3("<Vector1D::Vector1D>");

    mVec    = new double[1];
    mVec[0]	= 0.0;
}

/*!
*
*  \brief Default constructor
*
*  \li aName = aname of the vector
*  \li aXmin = minimal value on the structured grid
*  \li aXmax = maximal value on the structured grid
*  \li aXpts = number of points on the structured grid
*
*/

Vector1D::Vector1D(string aName, double aXmin, double aXmax, int aXpts)
: mVecName(aName), mXmin(aXmin), mXmax(aXmax), mXpts(aXpts){

    PRINT_DEBUG_3("<Vector1D::Vector1D>");

    if(mXpts <= 0) mXpts = 1;

    mDi = (mXpts - 1) / (mXmax - mXmin);

    mVec = new double [mXpts];
    for (int i=0; i < mXpts; i++) {
        mVec[i] = 0.0;
    }
}

/*!
*
*  \brief Default copying constructor
*
*  \li aVector = original vector which will be copied (the content is not copied)
*
*/

Vector1D::Vector1D(const Vector1D& aVector){

    PRINT_DEBUG_3("<Vector1D::Vector1D>");

    mXmin = aVector.mXmin;
    mXmax = aVector.mXmax;
    mXpts = aVector.mXpts;
    mDi = aVector.mDi;

    mVecName = "";
    mVec = new double [mXpts];
    for (int i=0; i<mXpts; i++) {
        mVec[i] = 0.0;
    }
}

/*!
*
*  \brief Default destructor
*
*  \li aVector = original vector which will be copied (the content is not copied)
*
*/

Vector1D::~Vector1D(){

    PRINT_DEBUG_3("<Vector1D::~Vector1D>");

    if(mVec){
        delete[] mVec;
    }
}

/*!
*
*  \brief Function returning value of the vector for given argument using linear interpolation (weighted average: f(x) = (1-x)*f(0) + x*f(1))
*
*  \li aX = argument of the vector
*
*/

double Vector1D::Interpolate(double aX){

    PRINT_DEBUG_3("<Vector1D::~Vector1D>");

    mti = (aX - mXmin) * mDi;
    mi = static_cast<int>(mti);
    mti -= mi;

    return mVec[mi] * (1-mti) + mVec[mi+1] * mti;
}

/*!
*
*  \brief Function calculating derivative of the vector in X direction
*
*  \li aName = name of the vector containing derivative
*
*/

Vector1D* Vector1D::DerivativeX(const char* aName){

    PRINT_DEBUG_3("<Vector1D::~DerivativeX>");

    int ti;
    Vector1D* tVec;

    mXd = 1/ mDi; //(mXmax - mXmin) / (mXpts - 1);
    tVec = new Vector1D(*this);
    tVec->mVecName = aName;
    for (int i=0; i<mXpts; i++) {
        ti = InitDerivative(i, mXmin, mXmax, mXpts);
        (*tVec)(i) = Derivative(mVec[ti-1], mVec[ti], mVec[ti+1]);
    }
    return tVec;
}

inline int Vector1D::InitDerivative(int aIdx, double aAMin, double aAMax, int aAPts){

    PRINT_DEBUG_3("<Vector1D::~InitDerivative>");

    if(aIdx == 0) {
        mXin = aAMin;
        mXi0 = aAMin + mXd;
        mXip = aAMin + mXd * 2.0;
        mX0  = aAMin;
        return aIdx + 1;
    } else if (aIdx == aAPts - 1) {
        mXin = aAMax - mXd * 2.0;
        mXi0 = aAMax - mXd;
        mXip = aAMax;
        mX0  = aAMax;
        return aIdx - 1;
    } else {
        mXin = aAMin + mXd * (aIdx - 1);
        mXi0 = aAMin + mXd *  aIdx;
        mXip = aAMin + mXd * (aIdx + 1);
        mX0  = mXi0;
        return aIdx;
    }

}

inline double Vector1D::Derivative(double aFin, double aFi0, double aFip){

    PRINT_DEBUG_3("<Vector1D::~DerivativeX>");

    return (
        mXin * (mXin - 2.0 * mX0) * (aFip - aFi0) +
        mXi0 * (mXi0 - 2.0 * mX0) * (aFin - aFip) +
        mXip * (mXip - 2.0 * mX0) * (aFi0 - aFin)
    ) / (2.0 * mXd * mXd * mXd);

}

