/** 
*
*   WeightedLAXSmoother.cxx
*	
*   Copyright (C) 2017 Radoslaw Ryblewski and Michael Strickland
*
*   This file is part of aHydro3p1.
*
*   aHydro3p1 is free software: you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation, either version 3 of the License, or
*   (at your option) any later version.
*
*   aHydro3p1 is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with aHydro3p1.  If not, see <http://www.gnu.org/licenses/>.
*    
*/

#include "Global.h"

/*!
 *
 *  \brief Default constructor
 *
 *  \li *aConfigurator = pointer to the configuration
 *  \li *aGrid = pointer to the grid
 *  \li *aStepper = pointer to the stepper
 *  \li *aEquations = pointer to the equations
 *
 */

WeightedLAXSmoother::WeightedLAXSmoother(Configurator *aConfigurator, Grid *aGrid, Stepper *aStepper, Equations *aEquations)
: mVArr(aStepper->GetArrays()), mNEq(aEquations->GetNumberOfEquations()), mXP(aGrid->GetXP()), mYP(aGrid->GetYP()), mZP(aGrid->GetZP()), mLAXfrac(aConfigurator->GetDouble("LAXfrac"))
{

    PRINT_DEBUG_3("WeightedLAXSmoother::WeightedLAXSmoother");

}

/*!
 *
 *  \brief Destructor
 *
 */

WeightedLAXSmoother::~WeightedLAXSmoother(){

    PRINT_DEBUG_3("WeightedLAXSmoother::~WeightedLAXSmoother");
}

/*!
 *
 *  \brief Perform weighted LAX smoothing
 *
 */

void    WeightedLAXSmoother::Smooth(){

    PRINT_DEBUG_3("WeightedLAXSmoother::Smooth");

    usi im,ip,jm,jp,km,kp;

    // copy data from arrays No.0 to arrays No.1

    CopyArrays(0,1);

    for(usi n = 0; n < mNEq; n++){
        for (usi i = 0; i < mXP; i++) {
            im = (i!=0) ? i-1 : i;
            ip = (i<mXP-1) ? i+1 : i;
            for (usi j = 0; j < mYP; j++) {
                jm = (j!=0) ? j-1 : j;
                jp = (j<mYP-1) ? j+1 : j;
                for (usi k = 0; k < mZP; k++) {
                    km = (k!=0) ? k-1 : k;
                    kp = (k<mZP-1) ? k+1 : k;
                    // 3d mode
                    (*mVArr[n][0])(i,j,k) = (1-mLAXfrac)*(*mVArr[n][1])(i,j,k)
                            + mLAXfrac*(   (*mVArr[n][1])(im,j,k) + (*mVArr[n][1])(ip,j,k)
                                         + (*mVArr[n][1])(i,jm,k) + (*mVArr[n][1])(i,jp,k)
                                         + (*mVArr[n][1])(i,j,km) + (*mVArr[n][1])(i,j,kp)   )/6;
                }
            }
        }
    }

}

/*!
 *
 *  \brief Copy arrays
 *
 *  \li aFrom = index of array to copy from (integer between 0 and 3)
 *  \li aTo = index of array to copy to (integer between 0 and 3)
 *
 */

inline void WeightedLAXSmoother::CopyArrays(const usi aFrom, const usi aTo){

    PRINT_DEBUG_3("WeightedLAXSmoother::CopyArrays");

    Vector3D *tVecT(0), *tVecF(0);

    for(usi n = 0; n < mNEq; n++){
    tVecF = mVArr[n][aFrom];
    tVecT = mVArr[n][aTo];
    for (usi i = 0; i < mXP; i++)
        for (usi j = 0; j < mYP; j++)
            for (usi k = 0; k < mZP; k++){

                (*tVecT)(i,j,k) = (*tVecF)(i,j,k);

            }
    }
}

/*!
 *
 *  \brief Print information
 *
 */

void WeightedLAXSmoother::PrintInfo(void){

    PRINT_DEBUG_3("<WeightedLAXSmoother::PrintInfo>");

    PrintRawMessage("-> Weighted LAX Smoother");
    PrintRawMessage("\tWeight factor", mLAXfrac);
}

